/*
  Copyright 2009,2014
  John Culp, All Rights Reserved
*/

#ifndef ASSEMBLER_INCLUDED
#define ASSEMBLER_INCLUDED

#include <iostream>

#include "Line.hh"
#include "AssemFile.hh"
#include "AssemError.hh"

typedef unsigned long MN; //Mnemonic
const MN MN_NOP = 0x00;
const MN MN_LUI = 0x01;
const MN MN_LLI = 0x02;
const MN MN_ADD = 0x03;
const MN MN_SUB = 0x04;
const MN MN_ADDL = 0x05;
const MN MN_AND = 0x06;
const MN MN_OR  = 0x07;
const MN MN_XOR = 0x08;
const MN MN_NOT = 0x09;
const MN MN_BSET = 0x0a;
const MN MN_BCLR = 0x0b;
const MN MN_RSL = 0x0c;
const MN MN_RSR = 0x0d;
const MN MN_MUL = 0x10;
const MN MN_CMP = 0x30;
const MN MN_SRI = 0x40;
const MN MN_SRR = 0x41;
const MN MN_LRI = 0x42;
const MN MN_LRR = 0x43;
const MN MN_SEQZ = 0x50;
const MN MN_SNEQZ = 0x51;
const MN MN_SBSET = 0x52;
const MN MN_SBCLR = 0x53;
const MN MN_JI = 0x60;
const MN MN_JR = 0x61;
const MN MN_HLT = 0x70;

//offsets of various parts of the instruction
const unsigned I_CMD_O = 25;
const unsigned I_SRCA_O = 22;
const unsigned I_SRCB_O = 19;
const unsigned I_DEST_O = 16;

/* hack. should probably not inherit */
class Assembler : public AssemFile
{
public:
  Assembler(void);
  ~Assembler(void);

  bool DoAssem(std::ostream &rfout);

protected:
  Line debugLine; //copy for debugging (hack).
  long LookupAddress(const std::string &rLabel) const;

  //assembler Mnemonic
  //value from 0 to 127 that is upper 7 bits if instruction word. returns big num if unknown
  unsigned long TxtToCmd(const std::string &rTxt) const;
  unsigned long TxtToReg(const std::string &rTxt) const; //returns 0 to 7 inclusive for $A through $H
  //unsigned long TxtToULong(const std::string &rTxt) const;
  //long TxtToLong(const std::string &rTxt) const;
  long long TxtToValueLL(const std::string &rTxt) const; //handles hex, fp, and int
  //uses above handles truncation to 16bits
  unsigned long TxtToImm(const std::string &rTxt) const;
  //returns a 32 bit value
  unsigned long TxtToValue(const std::string &rTxt) const;
  //unsigned long ParseTarget(const std::string &rArg) const;

  void AssembleInsnMacroLine(const Line &rLine, std::ostream &rfout) const;
  void AssembleInsnLine(const Line &rLine, std::ostream &rfout) const;

  void HandleLUI(const Line &rLine, unsigned long & rinsn) const;
  void HandleLLI(const Line &rLine, unsigned long & rinsn) const;
  void HandleADD(const Line &rLine, unsigned long & rinsn) const;
  void HandleSUB(const Line &rLine, unsigned long & rinsn) const;
  void HandleADDL(const Line &rLine, unsigned long & rinsn) const;
  void HandleAND(const Line &rLine, unsigned long & rinsn) const;
  void HandleOR(const Line &rLine, unsigned long & rinsn) const;
  void HandleXOR(const Line &rLine, unsigned long & rinsn) const;
  void HandleNOT(const Line &rLine, unsigned long & rinsn) const;
  void HandleBSET(const Line &rLine, unsigned long & rinsn) const;
  void HandleBCLR(const Line &rLine, unsigned long & rinsn) const;
  void HandleRSL(const Line &rLine, unsigned long & rinsn) const;
  void HandleRSR(const Line &rLine, unsigned long & rinsn) const;
  void HandleMUL(const Line &rLine, unsigned long & rinsn) const;
  void HandleCMP(const Line &rLine, unsigned long & rinsn) const;
  void HandleSRI(const Line &rLine, unsigned long & rinsn) const;
  void HandleSRR(const Line &rLine, unsigned long & rinsn) const;
  void HandleLRI(const Line &rLine, unsigned long & rinsn) const;
  void HandleLRR(const Line &rLine, unsigned long & rinsn) const;
  void HandleSEQZ(const Line &rLine, unsigned long & rinsn) const;
  void HandleSNEQZ(const Line &rLine, unsigned long & rinsn) const;
  void HandleSBSET(const Line &rLine, unsigned long & rinsn) const;
  void HandleSBCLR(const Line &rLine, unsigned long & rinsn) const;
  void HandleJI(const Line &rLine, unsigned long & rinsn) const;
  void HandleJR(const Line &rLine, unsigned long & rinsn) const;
  void HandleHLT(const Line &rLine, unsigned long & rinsn) const;
};

#endif //ASSEMBLER_INCLUDED
