/*
  Copyright 2009,2014
  John Culp, All Rights Reserved
*/

#include <ctype.h>
#include <iostream>
#include <fstream>
#include <list>
#include <string>

#include "AssemFile.hh"
#include "Assembler.hh"

using namespace std;

void PrintUsage(void)
{
  cerr << "usage: assem <-f> in_file (.asm) out_file (.txt) <optional-list-file-name>" << endl;
  cerr << "-f forces the include of the floating point library" << endl;
}

int main(int argc, char**argv)
{
  if(argc != 3 && argc != 4 && argc != 5)
    {
      PrintUsage();
      return -1;
    };

  bool include_libfp = false;
  int param_offset = 0;
  if( string(argv[1]) == "-f")
    {
      include_libfp = true;
      param_offset = 1;
    };

  string inFileName(argv[1+param_offset]);
  string outAsmFileName(argv[2+param_offset]);

  string outListFileName;
  if(!include_libfp){
    if( argc == 4 ) outListFileName = argv[3];
  }else{
    if( argc == 5 ) outListFileName = argv[4];
  }

  cout << "including libFP" << endl;

  ifstream fin(inFileName.c_str() );
  if( !fin )
    {
      cerr << "cannot open file: " << inFileName << endl;
      return -2;
    };

  ofstream fout( outAsmFileName.c_str() );
  if( !fout )
    {
      cerr << "cannot open output file: " << outAsmFileName << endl;
      return -3;
    };

  //AssemFile assemFile;
  Assembler assemFile;
  if( !assemFile.LoadFile(fin) )
    {
      cerr << "failed to load file" << endl;
      return -3;
    };

  fin.close();
  if(include_libfp)
    {
      /* yeah. this is going to make people real happy. */
      /* hack, hack, hack */
      fin.open( "/home/john/tvc-all/assem/libfp.asm");
      if( !fin )
	{
	  cerr << "cannot open libfp file: " << inFileName << endl;
	  return -33;
	};

      if( !assemFile.LoadFile(fin) )
	{
	  cerr << "failed to load libfp file" << endl;
	  return -34;
	};
    };

  if( !assemFile.ParseFile() )
    {
      return -4;
    };

  //process the file.L
  if( !assemFile.DoAssem(fout) )
    {
      cerr << "assembly failed" << endl;
      return -5;
    };

  cerr << "done assembly" << endl;

  fin.close();
  fout.close();

  if( !outListFileName.empty() )
    {
      ofstream listOut( outListFileName.c_str() );
      if( !listOut )
	{
	  cerr << "cannot open list output file: " << outListFileName << endl;
	  return -6;
	};
      assemFile.OutputList(listOut);
      listOut.close();
      cerr << "done list output" << endl;
    }

  return 0;
}
