	;;Copyright 2010,2014
	;;John Culp, All Rights Reserved 

	;; arguments are in A and B. store results at address pointed to by the SP, but don't increment SP
fp_mul
	FNSETUP
	;; first test if either argument is zero.
	BCLR %A %C 31
	SNEQZ %C
	JI fp_mul_result_zero
	BCLR %B %C 31
	SNEQZ %C
	JI fp_mul_result_zero
	;; set E to be mask of exponent
	LI %E 7F800000h
	AND %A %E %C
	RSR %C %C 23 ; reg C is signed exponent of A
	ADDL %C %C -127  ; remove bias
	AND %B %E %D
	RSR %D %D 23 ; reg D is signed exponent of B
	ADDL %D %D -127  ; remove  bias
	ADD %C %D %F ;store exponent in F
	;; load E with the significand mask
	LI %E 007FFFFFh
	AND %A %E %C 		;%C holds significand of %A
	BSET %C %C 23  		;added hidden bit to C
	AND %B %E %D
	BSET %D %D 23		;reg d holds significand of B with hidden bit
	;; right shift both significands as we cannot use all significant bits
	RSR %C %C 12
	RSR %D %D 12
	MUL %C %D %C
	;; see if hidden bit is set in significand
	SBCLR %C 23
	JI fp_mul_shift_exponent
	;; we must left shift significand becase sig bit is not set
	;; borrow one from the exponent to compensate
	RSL %C %C 1
	ADDL %F %F -1
fp_mul_shift_exponent
	;; clear the hidden bit
	BCLR %C %C 23
	;; hack. I'm off by a factor of two on output? why?
	ADDL %F %F 1 		;hack why is this needed? algorithmic misunderstanding here.
	;; add the bias back to the exponent
	ADDL %F %F 127
	;; ok we know know the significand and exponent are correct,
	;; they just need to be in the correct places
	;; first left shift the exponent
	RSL %F %F 23
	;; now set the significand in reg F
	OR %F %C %F
	;; ok finally compute the sign bit
	XOR %A %B %C		;high bit is the result
	SBSET %C 31
	JI fp_mul_done
	BSET %F %F 31
	JI fp_mul_done
fp_mul_result_zero
	LI %F 0
fp_mul_done
	;; note that unmodified arguments are still in register A and B
	;; but we are going to restore them anyway.
	SRR %F %H 0
	RETURN


fp_add
	FNSETUP
fp_add_assem_entry 		;for subtract
	;; if argA is zero return argB
	BCLR %A %C 31 
	SEQZ %C
	JI fp_add_test_b
	ADDL %B %F 0 		;move B to F
	JI fp_add_done
	;; if argB is zero return argA
fp_add_test_b
	BCLR %B %C 31
	SEQZ %C
	JI fp_add_start_add
	ADDL %A %F 0		;move A to F
	JI fp_add_done
fp_add_start_add
	;; set C to exponent of A and D to exponent of B
	LI %G 7F800000h 	; set G to be mask of exponent
	AND %A %G %C
	RSR %C %C 23 ; 
	ADDL %C %C -127  ; remove bias, reg C is now signed exponent of A
	AND %B %G %D
	RSR %D %D 23 ; 
	ADDL %D %D -127  ; remove bias, reg D is signed exponent of B 
	;; put the values related to the largest exponent (between A & B) in %A and %C
	CMP %C %D %G
	SBSET %G 4
	JI fp_add_exchange_done
	;; exchange a and b
	OR %B %B %G
	OR %A %A %B
	OR %G %G %A
	;; exchange c and d
	OR %D %D %G
	OR %C %C %D
	OR %G %G %C
fp_add_exchange_done
	SUB %C %D %E 		;actual exponent difference
	;; if difference between exponents is too large, then there
	;; is no point adding the numbers, and we should just return the
	;; argument with the largest exponent
	LI %F 24
	CMP %E %F %F
	SBSET %F 2
	JI fp_add_do_add
	;;we are done, the exponent difference is too much. return the %A value
	OR %A %A %F
	JI fp_add_done
fp_add_do_add
	;; at this point A & B are raw values C & D are masked exponents and E is the signed exponent difference
	;; thinking about the sign of +/-A + +/-B is too much.
	;; so make sure A is positive, by 'multiplying' -1 to each value
	;; to preserve sign modifications, we push a value to the stack so
	;; we remember what we have done (and to preserve registers)
	;; value to push is in %G
	LI %F 0
	LI %G 0
	SBSET %A 31
	JI fp_add_sign_exchange_finish
	BSET %F %F 31
	XOR %A %F %A
	XOR %B %F %B
	LLI %G 1
fp_add_sign_exchange_finish
	PUSH %G			;push our sign modification state
	;; our result will use the exponent from the %A value (it is bigger)
	OR %C %C %G 		;store the proper signed exponent in G
	;;
	;; right now A & B are raw values C & D are masked exponents (possibly A,C and B,D are exchanged)
	;; we know A has a positive value, though B might be negative
	;; G is the signed exponent of A (the value with the largest exponent)
	;; E is the signed exponent difference between A dn B (and must be positive)
	;; 
	;; make %F the mask of the signficand
	LI %F 007FFFFFh
	;; we have the exponent figured out, so lets reuse C and D
	;; this is the last use of raw arguments
	AND %A %F %C
	AND %B %F %D
	;; set the hidden bits
	BSET %C %C 23
	BSET %D %D 23
	;; right shift the B value for difference between exponents
fp_add_shift_start
	SNEQZ %E
	JI fp_add_shift_done
	RSR %D %D 1
	ADDL %E %E -1
	JI fp_add_shift_start
fp_add_shift_done
	;; 
	;; right now A and B hold the possibly exchanged initial arguments
	;; C and D are the significands of A and B. D has been right shifted so that
	;; it effetivly has the same exponent as C;
	;; register G is the signed exponent of A,B
	;; 
	;; perform an add or subtract significands depending upon sign of %B
	SBCLR %B 31
	JI fp_add_do_subtract
	;; doing an add
	ADD %C %D %C
	;; see if one past hidden bit is set, if so, right shift
	;; and increment exponent
	SBSET %C 24
	JI fp_add_add_subtract_done
	RSR %C %C 1
	ADDL %G %G 1
	JI fp_add_add_subtract_done
fp_add_do_subtract			;90
	;; sign extend our negative value
	NOT %D %D
	ADDL %D %D 1
	;; add
	ADD %C %D %C
	;; if our result is negative, we must make it positive, and note that we are reversing the sign
	SBSET %C 31
	JI fp_add_sub_shift_start
	PULL %D
	ADDL %D %D 1
	PUSH %D
	;; reverse sign again.
	NOT %C %C
	ADDL %C %C 1
fp_add_sub_shift_start
	;; we must left shift until a 1 appears in the hidden significant bit
	;; unless the significand is zero, in which case our result is zero
	SNEQZ %C  ;; sanity check so don't loop zero register	
	JI fp_add_add_subtract_done
fp_add_sub_shift_loop_start
 	;sanity check for C was done above
	SBCLR %C 23
	JI fp_add_sub_shift_done
	RSL %C %C 1
	ADDL %G %G -1
	JI fp_add_sub_shift_loop_start
fp_add_sub_shift_done
fp_add_add_subtract_done
	;; all teh sign bs will now be in D and our stack is all fixed up.
	;; this must be done now before the possible branch to done.
	PULL %D 
	;; we will build the return value in register F
	LI %F 0
	;; if the significand is zero then just return a zero.
	SNEQZ %C
	JI fp_add_done
	;; LI %F 0 ;this was done above.
	;; first the exponent
	ADDL %G %F 127
	RSL %F %F 23
	;; clear the hidden bit in the significand
	BCLR %C %C 23
	OR %C %F %F
	;; make sign adjustments as required
	LI %A 1
	CMP %A %D %A
	;; sign change can be 0 1 or 2, only if 1 do we flip sign of result
	SBSET %A 0		;test for == 1
	JI fp_add_done
	;; flip output sign bit
	LI %A 80000000h
	XOR %F %A %F
fp_add_done
	SRR %F %H 0
	RETURN

fp_sub
	FNSETUP
	;; negate the second argument and call add
	SBCLR %B 31
	JI fp_sub_make_pos
	BSET %B %B 31 		; if sign was clear, we must set it
	JI fp_sub_do_add
fp_sub_make_pos
	BCLR %B %B 31 		; clear sign bit
fp_sub_do_add
	JI fp_add_assem_entry	;no reason for nested call...(JI instead of call)
	HLT 			;not reachable


fp_div
	FNSETUP
	;; 
	;; subtract the two exponents and push the signed result to the stack
	;; 
	BCLR %A %C 31 		;clear sign bit and store arguments in C & D
	;; first quickly check if A is zero. if so the result is zero.
	SEQZ %C
	JI fp_div_a_not_zero
	LI %F 0
	JI fp_div_done
fp_div_a_not_zero
	BCLR %B %D 31
	RSR %C %C 23		;RSR away the significand
	RSR %D %D 23
	ADDL %C %C -127		;remove the bias
	ADDL %D %D -127
	SUB %C %D %F		;do the subtract
	PUSH %F
	;;
	;; now setup for the division. get the significands's in C&D with
	;; hidden bit set. clear E and set the hidden bit, which we
	;; will use as a counter, and to identify the relevant bit for division
	LI %E 007FFFFFh
	AND %A %E %C
	AND %B %E %D
	BSET %C %C 23
	BSET %D %D 23
	LI %E 00000000h
	BSET %E %E 23
	;; 
	;; do the division. build result in F
	;;
	LI %F 00000000h
fp_div_loop
	CMP %C %D %G		;g holds comparison bitmask
	SBCLR %G 0 		;test for ==
	JI fp_div_bit_set
	SBCLR %G 1		; test for >
	JI fp_div_bit_set
	JI fp_div_shift_and_loop	;not >=. move to next bit
fp_div_bit_set
	OR %E %F %F
	SUB %C %D %C
fp_div_shift_and_loop
	RSR %D %D 1
	RSR %E %E 1
	SEQZ %E
	JI fp_div_loop
	;;
	;; done with division. figure out the sign of the result.
	;; we will no longer need the origional arguments, so
	;; use A to hold results of XOR. bit 31 of result is correct
	;; sign bit
	;;
	XOR %A %B %A
	;;
	;; division result needs a bit set at the hidden bit. to
	;; make this happen (if it is not set) shift result left and subtract
	;; one from the exponent until the hidden bit is set.
	;; make sure that the value is not zero before we start shifting.
	;; as we don't need an infinite loop.
	;;
	PULL %B 		;get computed exponent in B
	SNEQZ %F
	JI fp_div_shift_done
fp_div_shift_start
	SBCLR %F 23
	JI fp_div_shift_done
	RSL %F %F 1
	ADDL %B %B -1
	JI fp_div_shift_start
fp_div_shift_done
	;; 
	;; build the completed result
	;; B holds the exponent %F holds the significand A has high bit sign
	;;
	BCLR %F %F 23 		;clear hidden bit in significand
	ADDL %B %B 127		;add the bias
	RSL %B %B 23		;left shift exponent into proper bits
	OR %F %B %F		;combine exponent and significand
	;; set or clear sign bit
	SBSET %A 31
	JI fp_div_clear_sign
	;; set the sign bit
	BSET %F %F 31
	JI fp_div_done
fp_div_clear_sign
	BCLR %F %F 31
fp_div_done
	SRR %F %H 0
	RETURN


fp_cast_to_int
	FNSETUP
	LI %B 80000000h		; arg is in %A
	AND %A %B %E				 ;%E is non-zero if negative bit is set
	NOT %B %B				 ;invert the negative bit mask for all other bits
	AND %A %B %A				 ;clear the sign bit if it was ever set
	SEQZ %A
	JI fp_cti_end_zero_test
	LI %A 0
	SRR %A %H 0		;save to the address of the stack pointer. don't increment
fp_cti_end_zero_test
	;; going to keep the significand in a and the exponent in b
	;; right now both are in %A
	LI %C 7F800000h
	AND %A %C %C		;selet the exp bits
	RSR %C %B 23 		;shift exp bits to lsb and save in %B
	;; actually the exponent could be negative. test if this is the case and set all the high bits
	;; to be sure the exp is treated as negative in subsesquent opps.
	;;; 	SBSET %B 7
	;;; 	JI end_fix_negative
	;;; 	LI %C FFFFFF00h
	;;; 	OR %C %B %B
	;;; end_fix_negative
	LI %C 7FFFFFh
	AND %A %C %A		;get significand in %A
	BSET %A %A 23		;set the hidden significant bit
	;; subtract "hidden"/implict 23 bits in exponent and the 127 offset
	ADDL %B %B -150
	LI %D 0
	SBSET %B 31
	JI fp_cti_exp_is_pos
fp_cti_exp_is_negative
	ADDL %B %C 0		;copy exp to %C
fp_cti_start_shift_right
	SNEQZ %C 		;relys on approacing from correct direction
	JI fp_cti_done_shift
	RSR %A %A 1		;right shift the significand
	ADDL %C %C 1		;increment 'ii'
	JI fp_cti_start_shift_right
fp_cti_end_shift_right
	JI fp_cti_done_shift 		;not ever reached...
fp_cti_exp_is_pos
	LI %C 0
fp_cti_start_shift_left
	CMP %C %B %D
	SBSET %D 4
	JI fp_cti_end_shift_left
	RSL %A %A 1
	ADDL %C %C 1
	JI fp_cti_start_shift_left
fp_cti_end_shift_left
fp_cti_done_shift
	SNEQZ %E
	JI fp_cti_end_negate
	LI %B 0
	SUB %B %A %A
fp_cti_end_negate
	;; put result (%A) in correct mem locaion
	SRR %A %H 0
	RETURN

fp_cast_to_float //arg in %A
	FNSETUP
	SEQZ %A
	JI fp_ctf_handle_zero_end
	;; return a zero value
	SRR %A %H 0
	RETURN
fp_ctf_handle_zero_end
	LI %E 0 		;store results in %E
	LI %B 0
	CMP %A %B %D
	SBSET %D 4
	JI fp_ctf_end_negate
	SUB %B %A %A		;%A = 0-%A i.e. negate
	BSET %E %E 31		;set the negative bit in the output.
fp_ctf_end_negate
	LI %B 0 		;%B holds the exp value
	LI %C FFFFFFh		; comparison value
fp_ctf_inc_exp_loop_start
	CMP %A %C %D
	SBSET %D 1
	JI fp_ctf_inc_exp_loop_end
	RSR %A %A 1
	ADDL %B %B 1
	JI fp_ctf_inc_exp_loop_start
fp_ctf_inc_exp_loop_end
	LI %C 800000h
fp_ctf_dec_exp_loop_start
	CMP %A %C %D
	SBSET %D 3
	JI fp_ctf_dec_exp_loop_end
	RSL %A %A 1
	ADDL %B %B -1
	JI fp_ctf_dec_exp_loop_start
fp_ctf_dec_exp_loop_end
	;; have sign bit aleady set in %E our results
	;; move the exp values to the result
	ADDL %B %B 150
	RSL %B %B 23
	OR %B %E %E
	LI %B 7FFFFFh 		;significand mask
	AND %B %A %A
	OR %A %E %E		;move significand bits to result
	SRR %E %H 0
	RETURN

fp_is_pos
	FNSETUP	                ;this is wrong. should access via %F
	LRR %H %A -1		;load A with function param
	SBCLR %A 31
	JI fp_is_pos_neg 	;prevent assembler name collision
	LI %A 1
	JI fp_is_pos_end
fp_is_pos_neg
	LI %A 0
fp_is_pos_end
	SRR %A %H 0 		; store result
	RETURN

fp_is_neg
	FNSETUP			;this is wrong. should access via %F
	LRR %H %A -1		;load A with function param
	SBCLR %A 31
	JI fp_is_neg_neg 	;prevent assembler name collision
	LI %A 0
	JI fp_is_neg_end
fp_is_neg_neg
	LI %A 1	
fp_is_neg_end
	SRR %A %H 0 		; store result
	RETURN
