/*
  Copyright 2010, 2014
  John Culp, All Rights Reserved
*/

#include <iostream>
#include <assert.h>
#include <typeinfo>
#include <sstream>

#include "Compiler.hh"

using namespace std;

Compiler::Compiler(Program &rTProg) :
  rProg(rTProg), og("\t")
{}

Compiler::~Compiler(void)
{}

std::string Compiler::OG(void)
{
  return og;
}

void Compiler::DoIt(void)
{
  ExecState execState;

  rProg.Exec(execState);
  std::list<MemOpp> initMemOpps( execState.moList );
  
  list<FunctionDef>::iterator cPlace = rProg.fdList.begin();
  while(cPlace != rProg.fdList.end())
    {
      FunctionDef &rFD = *cPlace;
      rProg.PrintTree(rFD.name + "_tvcc-1.graph");

      cout << "inserting compiler temps?!" << endl;
      InsertCompilerTemps(rFD);

      rFD.idGen.GetID();
      rFD.idGen.GetCTID();

      rFD.AssignUniqueIDs( rFD.idGen );

      /*
      VarStack globalOrig(rFD.name);
      globalOrig.PushContext("global");
      rProg.FillGlobalsFromMBList(globalOrig);

      VarStack thisStack(globalOrig);
      rFD.PushVarTypes(thisStack);
      thisStack.PopContext();
      */
      VarStack globalOrig();

      VarStack thisStack(rFD.name);
      thisStack.PushContext("global");
      rProg.FillGlobalsFromMBList(thisStack);
      thisStack.PushContext( string("fun_") + rFD.name );
      rFD.FillLocalsFromParameters(thisStack);
      //cout << "jblah" << endl;
      //thisStack.Print(cout);
      rFD.PushVarTypes(thisStack);
      //this is not really important
      thisStack.PopContext(); //function
      thisStack.PopContext(); //global


      rProg.PrintTree(rFD.name + "_tvcc-2.graph");

      ConvertTreeToStraightLine(rFD.name, rFD);

      rProg.PrintTree(rFD.name + "_tvcc-3.graph");

      //build a mem block for each function of it's compiler temps and 
      //user defined variables
      
      //MemBlock function_params;
      BuildMemBlocks(rFD, thisStack);

      cPlace++;
    };
}

void Compiler::InsertCompilerTemps(FullOppList &rFD)
{
  //only certain types of nodes are sensible top level nodes
  std::list<Opp*>::iterator cPlace = rFD.opps.begin();
  while(cPlace != rFD.opps.end() )
    {
      Opp* pOpp = *cPlace;
      assert(pOpp);

      InsertCompilerTemps_Chain(*pOpp);

      cPlace++;
    };
}

void Compiler::InsertCompilerTemps_Chain(Opp &rOpp)
{
  if( CodeVarDec *pCVD = dynamic_cast<CodeVarDec*>(&rOpp) )
    {
      //cout << "found code var dec" << endl;
      if(pCVD->pValueSrc)
	{
	  if( IE_Call *pCall = dynamic_cast<IE_Call*>(pCVD->pValueSrc) )
	    {
	      IE_CompTemp *pCompTemp = new IE_CompTemp(pCall);
	      pCVD->UpdateSrc(pCompTemp);
	      //skip past the comp temp and insert more temps on that chain
	      InsertCompilerTemps_Chain(*(pCompTemp->pSrc));
	    }else
	    InsertCompilerTemps_Chain(*pCVD->pValueSrc);
	};
      return;
    };

  if( CodeVarWrite *pCVW = dynamic_cast<CodeVarWrite*>(&rOpp) )
    {
      if( IE_Call *pCall = dynamic_cast<IE_Call*>(pCVW->pValueSrc) )
	{
	  IE_CompTemp *pCompTemp = new IE_CompTemp(pCall);
	  pCVW->UpdateSrc( pCompTemp );
	  //skip past the comp temp and insert more temps on that chain
	  InsertCompilerTemps_Chain(*(pCompTemp->pSrc));
	}else{
	InsertCompilerTemps_Chain(*pCVW->pValueSrc);
      };
      return;
    };

  if( OppReturn *pRet = dynamic_cast<OppReturn*>(&rOpp) )
    {
      if( !pRet->RetNull() )
	InsertCompilerTemps(pRet->oppList);
      return;
    };

  if( CodeVarWriteOffset *pCVWO = dynamic_cast<CodeVarWriteOffset*>(&rOpp) )
    {
      //cout << "found code var write offset!" << endl;
      if( IE_Call *pCall = dynamic_cast<IE_Call*>(pCVWO->pValueSrc) )
	{
	  IE_CompTemp *pCompTemp = new IE_CompTemp(pCall);
	  pCVWO->UpdateValueSrc( pCompTemp );
	  //skip past the comp temp and insert more temps on that chain
	  InsertCompilerTemps_Chain(*(pCompTemp->pSrc));
	}else{
	InsertCompilerTemps_Chain(*pCVWO->pValueSrc);
      };

      //cout << "found code var write offset!" << endl;
      if( IE_Call *pCall = dynamic_cast<IE_Call*>(pCVWO->pOffsetSrc) )
	{
	  IE_CompTemp *pCompTemp = new IE_CompTemp(pCall);
	  pCVWO->UpdateOffsetSrc( pCompTemp );
	  //skip past the comp temp and insert more temps on that chain
	  InsertCompilerTemps_Chain(*(pCompTemp->pSrc));
	}else{
	InsertCompilerTemps_Chain(*pCVWO->pOffsetSrc);
      };
      return;
    };

  if( OppFor *pFor = dynamic_cast<OppFor*>(&rOpp) )
    {
      //cout << "found a for" << endl;
      InsertCompilerTemps(pFor->initList);
      InsertCompilerTemps(pFor->testList);
      InsertCompilerTemps(pFor->incList);
      InsertCompilerTemps(pFor->oppList);
      return;
    };

  if( OppWhile *pWhile = dynamic_cast<OppWhile*>(&rOpp) )
    {
      //cout << "found a while" << endl;
      InsertCompilerTemps(pWhile->testList);
      InsertCompilerTemps(pWhile->oppList);
      return;
    };

  if( OppIf *pIf = dynamic_cast<OppIf*>(&rOpp) )
    {
      //cout << "found an if" << endl;
      InsertCompilerTemps(pIf->testList);
      InsertCompilerTemps(pIf->oppList);
      InsertCompilerTemps(pIf->elseList);
      return;
    };

  if( OppCall *pCall = dynamic_cast<OppCall*>(&rOpp) )
    {
      std::list<CallArg>::iterator cPlace = pCall->exprList.exprList.begin();
      while(cPlace != pCall->exprList.exprList.end() )
	{
	  CallArg &rCA = *cPlace;
	  InsertCompilerTemps(rCA);
	  cPlace++;
	};
      return;
    };

  string name;
  rOpp.GetGName(name);

  cout << "!!!! UNHANDLED OPP !!!! name: " << name << endl;
}

void Compiler::InsertCompilerTemps_Chain(IE &rIE)
{
  cout << "following chain" << endl;

  if( IE_Binary *pOB = dynamic_cast<IE_Binary*>(&rIE) )
    {
      cout << "found binary opp!!" << endl;

      //insert a compiler var as nodes for pA and pB, provided these
      //are not IE_Imm.  RegVar would be ok, but not at this stage.
      if( VerifyBinaryArg( *(pOB->pA) ) )
	{
	  cout << "inserting compiler temp for a" << endl;
	  pOB->UpdateA( new IE_CompTemp( (pOB->pA)) );
	  //ok now pA is a CompTemp; skip past it on that branch.
	  InsertCompilerTemps_Chain( *(dynamic_cast<IE_CompTemp&>( *(pOB->pA) ).pSrc) );
	};

      if( VerifyBinaryArg( *(pOB->pB) ) )
	{
	  cout << "inserting compiler temp for b" << endl;
	  pOB->UpdateB( new IE_CompTemp( (pOB->pB)) );
	  //ok now pB is a CompTemp; skip past it on that branch.
	  InsertCompilerTemps_Chain( *(dynamic_cast<IE_CompTemp&>( *(pOB->pB) ).pSrc) );
	};

      return;
    };

  if( IE_Unary *pUnary = dynamic_cast<IE_Unary*>(&rIE) )
    {
      if( VerifyBinaryArg( *(pUnary->pSrc) ) )
	{
	  pUnary->UpdateSrc( new IE_CompTemp(pUnary->pSrc) );
	  InsertCompilerTemps_Chain( *(dynamic_cast<IE_CompTemp&>( *(pUnary->pSrc) ).pSrc) );
	};
      return;
    };

  if( IE_Call *pCall = dynamic_cast<IE_Call*>(&rIE) )
    {
      std::list<CallArg>::iterator cPlace = pCall->exprList.exprList.begin();
      while(cPlace != pCall->exprList.exprList.end() )
	{
	  CallArg &rCA = *cPlace;
	  InsertCompilerTemps(rCA);
	  cPlace++;
	};
      return;
    };

  // if( IE_Bin_Not *pNot = dynamic_cast<IE_Bin_Not*>(&rIE) )
  //   {
  //     if( VerifyBinaryArg( *(pNot->pSrc) ) )
  // 	{
  // 	  pNot->UpdateSrc( new IE_CompTemp(pNot->pSrc) );
  // 	  InsertCompilerTemps_Chain( *(dynamic_cast<IE_CompTemp&>( *(pNot->pSrc) ).pSrc) );
  // 	};
  //     return;
  //   };
}

bool Compiler::VerifyBinaryArg(IE &rOpp)
{
  if( dynamic_cast<IE_Binary*>(&rOpp) ) return true;
  if( dynamic_cast<IE_Unary*>(&rOpp) ) return true;

  if( dynamic_cast<IE_Call*>(&rOpp) ) return true;
  /*
  if( dynamic_cast<IE_Shift*>(&rOpp) ) return true;
  if( dynamic_cast<IE_Bin_Not*>(&rOpp) ) return true;
  if( dynamic_cast<IE_Cast*>(&rOpp) ) return true;
  //hack? should I return true for all unary opps?
  */

  if( dynamic_cast<IE_Imm*>(&rOpp) ) return false;

  return false;
}

void Compiler::ConvertTreeToStraightLine(const std::string &rFnName, FullOppList &rFOL)
{
  std::list<Opp*>::iterator oppPlace = rFOL.opps.begin();
  CTTSL_Start(rFnName, rFOL, oppPlace);
}

void Compiler::CTTSL_Start(const std::string &rFnName, FullOppList &rFOL, std::list<Opp*>::iterator &oppPlace)
{
  //only certain types of nodes are sensible top level nodes
  std::list<Opp*>::iterator cPlace = oppPlace;
  while(cPlace != rFOL.opps.end() )
    {
      Opp* pOpp = *cPlace;
      assert(pOpp);

      std::list<Opp*>::iterator origPlace = cPlace;
      CTTSL_Handle_Opp(rFnName, rFOL, pOpp, cPlace);

      if( cPlace == origPlace) //nothing was inserted
	cPlace++;
      //else assume CTTSL_Handle_Opp reset cPlace to where it needed to be.
    };
}

void Compiler::CTTSL_Handle_Opp(const std::string &rFnName, FullOppList &rFOL, Opp *pOpp, std::list<Opp*>::iterator &oppPlace)
{
  assert(pOpp);
  if( CodeVarDec *pCVD = dynamic_cast<CodeVarDec*>(pOpp) )
    {
      cout << "found code var dec" << endl;
      IE *pChild = pCVD->pValueSrc;
      if( pChild ) //we steal the value src
	{
	  cout << "inserting write!" << endl;
	  pCVD->pValueSrc = 0;

	  CodeVarWrite *pCVW = new CodeVarWrite(pCVD->name, *pChild);

	  std::list<Opp*>::iterator insertPlace = oppPlace;
	  insertPlace++;
	  rFOL.opps.insert(insertPlace, pCVW);

	  pOpp = pCVW; //*oppPlace; //just to stay synced.
	  //delete pChild; //hack. wtf. can't delete this. fubar someplace.
	  //pChild = 0;
	}; //could just fall down to process the code var write; buts lets keep things simple
      return;
    };

  if( CodeVarWrite *pCVW = dynamic_cast<CodeVarWrite*>(pOpp) )
    {
      assert( pOpp == *oppPlace );
      cout << "found CodeVarWrite" << endl;
      IE *pChild = pCVW->pValueSrc;

      if( IE_Binary *pOB = dynamic_cast<IE_Binary*>(pChild) )
	{
	  CTSL_Binary(rFOL, oppPlace, pOB, pOB->pA, 1); //note oppPlace updates here..
	  CTSL_Binary(rFOL, oppPlace, pOB, pOB->pB, 2);//and here.
	  return;
	}

      if( IE_Unary *pUnary = dynamic_cast<IE_Unary*>(pChild) )
	{
	  CTSL_Unary(rFOL, oppPlace, pUnary);
	  return;
	};

      if( IE_CompTemp *pCT = dynamic_cast<IE_CompTemp*>(pChild) )
	{
	  CompTempStore *pCTS = new CompTempStore(pCT->id);
	  pCTS->pSrc = pCT->pSrc;
	  
	  IE_CompTempLoad *pCTL = new IE_CompTempLoad(pCT->id, pCT->GetType(), REG_A); //bs register
	  pCVW->pValueSrc = pCTL;

	  //reset starting point so we sweep though our old source tree and CTSL
	  oppPlace = rFOL.opps.insert(oppPlace, pCTS);
	  //oppPlace--; //hack were not starting at right place.

	  pCT->pSrc = 0;
	  delete pCT;
	  pCT = 0;
	  return;
	}

      if( dynamic_cast<IE_Imm*>(pChild) )
	{
	  cout << "ignoring oppImm child of CodeVarWrite" << endl;
	  return;
	};

      if( dynamic_cast<IE_CodeVarRead*>(pChild) ){
	cout << "ignoring var read child of CodeVarWrite" << endl;
	return;
      };

      if( dynamic_cast<IE_CodeVarReadOffset*>(pChild) ){
	cout << "ignoring var read child offset of CodeVarWrite" << endl;
	return;
      };

      if( dynamic_cast<IE_CompTempLoad*>(pChild) ){
	cout << "ignoring CompTempLoad child of CodeVarWrite" << endl;
	return;
      };

      string name;
      pChild->GetGName(name);

      cout << "uknown child of CodeVarWrite was: " << name << endl;
      cout << "location: " << pChild->loc << endl;
      throw("unknown child of code var write");
      return;
    }

  /* this is an absolutly disgusting cut and paste from above, _AND_ a rediculously disgusting
     cut and paste inside of it. */
  if( CodeVarWriteOffset *pCVWO = dynamic_cast<CodeVarWriteOffset*>(pOpp) )
    {
      assert( pOpp == *oppPlace );
      cout << "found CodeVarWriteOffset" << endl;
      IE *pChild = pCVWO->pValueSrc;
      assert(pChild);

      bool value_handled = false;
      if( IE_Binary *pOB = dynamic_cast<IE_Binary*>(pChild) )
	{
	  CTSL_Binary(rFOL, oppPlace, pOB, pOB->pA, 1); //note oppPlace updates here..
	  CTSL_Binary(rFOL, oppPlace, pOB, pOB->pB, 2);//and here.
	  value_handled = true;
	}

      if(!value_handled)
	if( IE_Unary *pUnary = dynamic_cast<IE_Unary*>(pChild) )
	  {
	    CTSL_Unary(rFOL, oppPlace, pUnary);
	    value_handled = true;
	  };

      if(!value_handled)
	if( IE_CompTemp *pCT = dynamic_cast<IE_CompTemp*>(pChild) )
	  {
	    CompTempStore *pCTS = new CompTempStore(pCT->id);
	    pCTS->pSrc = pCT->pSrc;
	    
	    IE_CompTempLoad *pCTL = new IE_CompTempLoad(pCT->id, pCT->GetType(), REG_A); //bs register
	    pCVWO->pValueSrc = pCTL;
	    
	    //reset starting point so we sweep though our old source tree and CTSL
	    oppPlace = rFOL.opps.insert(oppPlace, pCTS);
	    //oppPlace--; //hack were not starting at right place.
	    
	    pCT->pSrc = 0;
	    delete pCT;
	    pCT = 0;
	    value_handled = true;
	  }

     if(!value_handled)
      if( dynamic_cast<IE_Imm*>(pChild) )
	{
	  cout << "ignoring oppImm child of CodeVarWrite" << endl;
	  value_handled = true;
	};

      if(!value_handled)
     if( dynamic_cast<IE_CodeVarRead*>(pChild) ){
	cout << "ignoring var read child of CodeVarWrite" << endl;
	value_handled = true;
      };

      if(!value_handled)
     if( dynamic_cast<IE_CodeVarReadOffset*>(pChild) ){
	cout << "ignoring var read child offset of CodeVarWrite" << endl;
	value_handled = true;
      };

     if(!value_handled)
      if( dynamic_cast<IE_CompTempLoad*>(pChild) ){
	cout << "ignoring CompTempLoad child of CodeVarWrite" << endl;
	value_handled = true;
      };

      if(!value_handled)
	{
	  string name;
	  pChild->GetGName(name);
	  cout << "uknown child of CodeVarWriteOffset value was: " << name << endl;
	  cout << "location: " << pChild->loc << endl;
	  throw("unknown child of code var write offset value");
	};

      pChild = pCVWO->pOffsetSrc;
      assert(pChild);

      bool offset_handled = false;
      if( IE_Binary *pOB = dynamic_cast<IE_Binary*>(pChild) )
	{
	  CTSL_Binary(rFOL, oppPlace, pOB, pOB->pA, 1); //note oppPlace updates here..
	  CTSL_Binary(rFOL, oppPlace, pOB, pOB->pB, 2);//and here.
	  offset_handled = true;
	}

     if(!offset_handled)
      if( IE_Unary *pUnary = dynamic_cast<IE_Unary*>(pChild) )
	{
	  CTSL_Unary(rFOL, oppPlace, pUnary);
	  offset_handled = true;
	};

     if(!offset_handled)
      if( IE_CompTemp *pCT = dynamic_cast<IE_CompTemp*>(pChild) )
	{
	  CompTempStore *pCTS = new CompTempStore(pCT->id);
	  pCTS->pSrc = pCT->pSrc;
	  
	  IE_CompTempLoad *pCTL = new IE_CompTempLoad(pCT->id, pCT->GetType(), REG_A); //bs register
	  pCVWO->pOffsetSrc = pCTL;

	  //reset starting point so we sweep though our old source tree and CTSL
	  oppPlace = rFOL.opps.insert(oppPlace, pCTS);
	  //oppPlace--; //hack were not starting at right place.

	  pCT->pSrc = 0;
	  delete pCT;
	  pCT = 0;
	  offset_handled = true;
	}

     if(!offset_handled)
      if( dynamic_cast<IE_Imm*>(pChild) )
	{
	  cout << "ignoring oppImm child of CodeVarWrite" << endl;
	  offset_handled = true;
	};

     if(!offset_handled)
      if( dynamic_cast<IE_CodeVarRead*>(pChild) ){
	cout << "ignoring var read child of CodeVarWrite" << endl;
	offset_handled = true;
      };

     if(!offset_handled)
      if( dynamic_cast<IE_CodeVarReadOffset*>(pChild) ){
	cout << "ignoring var read child offset of CodeVarWrite" << endl;
	offset_handled = true;
      };

     if(!offset_handled)
      if( dynamic_cast<IE_CompTempLoad*>(pChild) ){
	cout << "ignoring CompTempLoad child of CodeVarWrite" << endl;
	offset_handled = true;
      };

      if(!offset_handled)
	{
	  string name;
	  pChild->GetGName(name);
	  cout << "uknown child of CodeVarWriteOffset offset was: " << name << endl;
	  cout << "location: " << pChild->loc << endl;
	  throw("unknown child of code var write offset offset");
	};
      return;
    }

  if( CompTempStore *pCTS = dynamic_cast<CompTempStore*>(pOpp) )
    {
      cout << "found CompTempStore!" << endl;
      IE *pChild = pCTS->pSrc;
      if( IE_Binary *pOB = dynamic_cast<IE_Binary*>(pChild) ){
	CTSL_Binary(rFOL, oppPlace, pOB, pOB->pA, 1);
	CTSL_Binary(rFOL, oppPlace, pOB, pOB->pB, 2);
	return;
      };

      if( IE_Unary *pUnary = dynamic_cast<IE_Unary*>(pChild) ){
	CTSL_Unary(rFOL, oppPlace, pUnary);
	return;
      };

      if( IE_Call *pCall = dynamic_cast<IE_Call*>(pCTS->pSrc) )
	{
	  bool did_insert = false;
	  std::list<Opp*>::iterator topPlace = oppPlace;
	  topPlace--;
	  std::list<CallArg>::iterator cPlace = pCall->exprList.exprList.begin();
	  while(cPlace != pCall->exprList.exprList.end() )
	    {
	      CallArg &rCA = *cPlace;
	      if( !rCA.opps.empty() )
		{
		  did_insert = true;
		  rFOL.opps.insert(oppPlace, rCA.opps.begin(), rCA.opps.end());
		  rCA.opps.clear();
		};
	      cPlace++;
	    };
	  if(did_insert)
	    oppPlace = topPlace;
	  return;
	};

      if( dynamic_cast<IE_CodeVarRead*>(pChild) ){
	cout << "ignoring var read child of CompTempStore" << endl;
	return;
      };

      string name;
      pChild->GetGName(name);
      cout << "uknown child of CompTempStore was: " << name << " loc: " << pChild->loc << endl;
      throw("unknown child of Comp Temp Store");
      return;
    };

  if( OppCall *pCall = dynamic_cast<OppCall*>(pOpp) )
    {
      bool did_insert = false;
      std::list<Opp*>::iterator topPlace = oppPlace;
      topPlace--;
      std::list<CallArg>::iterator cPlace =  pCall->exprList.exprList.begin();
      while(cPlace != pCall->exprList.exprList.end() )
	{
	  CallArg &rCA = *cPlace;
	  if( !rCA.opps.empty() )
	    {
	      did_insert = true; 
	      rFOL.opps.insert(oppPlace, rCA.opps.begin(), rCA.opps.end());
	      rCA.opps.clear();
	    };
	  cPlace++;
	};
      if(did_insert)
	oppPlace = topPlace;
      return;
    };

  if( OppWhile *pOW = dynamic_cast<OppWhile*>(pOpp) )
    {
      cout << "found OppWhile!" << endl;
      ostringstream wt_name, we_name;
      wt_name << rFnName << "_while_top_" << pOW->oppID;
      we_name << rFnName << "_while_end_" << pOW->oppID;

      std::list<Opp*>::iterator topPlace = rFOL.opps.insert(oppPlace, new Target(wt_name.str()) );

      //insert the test code.
      rFOL.opps.insert(oppPlace, pOW->testList.opps.begin(), pOW->testList.opps.end());
      pOW->testList.opps.clear(); //clear program in oppWhile they are not deleted by its distructor

      //insert opps to test
      ostringstream testVar;
      testVar << rFnName << "_while_" << pOW->oppID << "_test";
      rFOL.opps.insert(oppPlace, new OppRegLoadFromVar(REG_D, testVar.str()) );
      rFOL.opps.insert(oppPlace, new SkipIfZero(REG_D, true) ); //invert the logic
      rFOL.opps.insert(oppPlace, new Jump(we_name.str()) );

      //insert the body code.
      rFOL.opps.insert(oppPlace, pOW->oppList.opps.begin(), pOW->oppList.opps.end());
      //hokay. now _clear_ the list of the while program so they are not deleted by its distructor
      pOW->oppList.opps.clear();

      //insert the end marker.
      rFOL.opps.insert(oppPlace, new Jump(wt_name.str()) );
      rFOL.opps.insert(oppPlace, new Target(we_name.str()) );

      //finally erase the while and delete it
      assert(pOW == *oppPlace);
      oppPlace = rFOL.opps.erase(oppPlace);
      oppPlace = topPlace;
      delete pOW;
      pOW = 0;
      return;
    };

  if( OppFor *pFor = dynamic_cast<OppFor*>(pOpp) )
    {
      cout << "found OppFor!" << endl;
      ostringstream for_loop_top, for_loop_end;
      for_loop_top << rFnName << "_for_top_" << pFor->oppID;
      for_loop_end << rFnName << "_for_end_" << pFor->oppID;

      std::list<Opp*>::iterator topPlace = oppPlace;
      topPlace--;

      //insert the init code.
      rFOL.opps.insert(oppPlace, pFor->initList.opps.begin(), pFor->initList.opps.end());
      pFor->initList.opps.clear(); //clear program in oppWhile they are not deleted by its distructor

      rFOL.opps.insert(oppPlace, new Target(for_loop_top.str()) );

      //insert test code
      rFOL.opps.insert(oppPlace, pFor->testList.opps.begin(), pFor->testList.opps.end());
      pFor->testList.opps.clear(); //clear program in oppWhile they are not deleted by its distructor

      //insert opps to test
      ostringstream testVar;
      testVar << rFnName << "_for_" << pFor->oppID << "_test";
      rFOL.opps.insert(oppPlace, new OppRegLoadFromVar(REG_D, testVar.str()) );
      rFOL.opps.insert(oppPlace, new SkipIfZero(REG_D, true) ); //invert the logic
      rFOL.opps.insert(oppPlace, new Jump(for_loop_end.str()) );

      //insert the body code.
      rFOL.opps.insert(oppPlace, pFor->oppList.opps.begin(), pFor->oppList.opps.end());
      //hokay. now _clear_ the list of the while program so they are not deleted by its distructor
      pFor->oppList.opps.clear();

      //insert the increment code
      rFOL.opps.insert(oppPlace, pFor->incList.opps.begin(), pFor->incList.opps.end());
      pFor->incList.opps.clear(); //clear program in oppWhile they are not deleted by its distructor

      //insert the end marker.
      rFOL.opps.insert(oppPlace, new Jump(for_loop_top.str()) );
      rFOL.opps.insert(oppPlace, new Target(for_loop_end.str()) );

      //finally erase the if and delete it
      assert(pFor == *oppPlace);
      oppPlace = rFOL.opps.erase(oppPlace);
      oppPlace = topPlace;
      delete pFor;
      pFor = 0;
      return;
    };

  if( OppIf *pIf = dynamic_cast<OppIf*>(pOpp) )
    {
      cout << "found OppIf!" << endl;
      ostringstream if_else_name, if_end_name;
      if_else_name << rFnName << "_if_else_" << pIf->oppID;
      if_end_name << rFnName << "_if_end_" << pIf->oppID;

      std::list<Opp*>::iterator topPlace = oppPlace;
      topPlace--;

      //insert the test code.
      rFOL.opps.insert(oppPlace, pIf->testList.opps.begin(), pIf->testList.opps.end());
      pIf->testList.opps.clear(); //clear program in oppWhile they are not deleted by its distructor

      //insert opps to test
      ostringstream testVar;
      testVar << rFnName << "_if_" << pIf->oppID << "_test";
      rFOL.opps.insert(oppPlace, new OppRegLoadFromVar(REG_D, testVar.str()) );
      rFOL.opps.insert(oppPlace, new SkipIfZero(REG_D, true) ); //invert the logic
      rFOL.opps.insert(oppPlace, new Jump(if_else_name.str()) );

      //insert the body code.
      rFOL.opps.insert(oppPlace, pIf->oppList.opps.begin(), pIf->oppList.opps.end());
      //hokay. now _clear_ the list of the while program so they are not deleted by its distructor
      pIf->oppList.opps.clear();

      rFOL.opps.insert(oppPlace, new Jump(if_end_name.str()) );

      //insert the else marker
      rFOL.opps.insert(oppPlace, new Target(if_else_name.str()) );
      //insert the else code (if any)
      rFOL.opps.insert(oppPlace, pIf->elseList.opps.begin(), pIf->elseList.opps.end());
      pIf->elseList.opps.clear();
      //insert the end marker.
      rFOL.opps.insert(oppPlace, new Target(if_end_name.str()) );

      //finally erase the if and delete it
      assert(pIf == *oppPlace);
      oppPlace = rFOL.opps.erase(oppPlace);
      oppPlace = topPlace;
      delete pIf;
      pIf = 0;
      return;
    };


  if( OppReturn *pRet = dynamic_cast<OppReturn*>(pOpp) )
    {
      if( !pRet->RetNull() )
	{
	  if( !pRet->oppList.opps.empty() )
	    {
	      std::list<Opp*>::iterator topPlace = oppPlace;
	      bool topOfList = false;
	      if(topPlace == rFOL.opps.begin() )
		topOfList = true;
	      else
		topPlace--;
	      rFOL.opps.insert(oppPlace, pRet->oppList.opps.begin(), pRet->oppList.opps.end());
	      pRet->oppList.opps.clear();
	      if(topOfList)
		oppPlace = rFOL.opps.begin();
	      else
		oppPlace = topPlace;
	    };
	};
      return;
    };

  string name;
  pOpp->GetGName(name);
  cout << "!!!! UNHANDLED OPP !!!! name: " << name << endl;
  //throw("convert tree to sl: unknown opp");
}

void Compiler::CTSL_Binary(FullOppList &rFOL, std::list<Opp*>::iterator &rOppPlace, IE_Binary *pOB, IE *pBranch, int branch)
{
  if( IE_Imm *pImm = dynamic_cast<IE_Imm*>(pBranch) ){
    //fill in the register id on the IE_Imm
    //I _could_ add a seperate immwrite instruction right now to the OppList, but why; we will
    //just insert the proper instruction at code generation time before the binary opp.
    if(branch == 1)
      pImm->reg = REG_A;
    else
      pImm->reg = REG_B;
    return;
  };

  if( IE_CompTemp *pCT = dynamic_cast<IE_CompTemp*>(pBranch) ){
    //hokay. must split the comp temp into two parts.
    REG_N dest;
    if( branch == 1)
      dest = REG_A;
    else
      dest = REG_B;

    //the entire tree of instructions that happens attached to this comp temp branch must be executed before
    //this binary instruction. remove that sequence of instructions from here. In its place add a stub CompTempLoad that
    //reads the proper 
    IE_CompTempLoad *pLoad = new IE_CompTempLoad(pCT->id, pCT->GetType(), dest);
    //note we are not stomping on a pointer because we sill hold pCT above.
    if( branch == 1 )
      pOB->pA = pLoad;
    else
      pOB->pB = pLoad;

    CompTempStore *pCTS = new CompTempStore(pCT->id);
    pCTS->pSrc = pCT->pSrc;

    pCT->pSrc = 0; //so we don't delete the rest of the branch, only this opp!
    delete pBranch;
    pBranch = 0;

    rOppPlace = rFOL.opps.insert(rOppPlace, pCTS);

    IE *pChild = pCTS->pSrc;
    assert(pChild);
    if( IE_Binary *pOB = dynamic_cast<IE_Binary*>(pChild) ){
      CTSL_Binary(rFOL, rOppPlace, pOB, pOB->pA, 1);
      CTSL_Binary(rFOL, rOppPlace, pOB, pOB->pB, 2);
    }else{
      if( IE_Unary *pU = dynamic_cast<IE_Unary*>(pChild) ){
	CTSL_Unary(rFOL, rOppPlace, pU);
      }else{
	bool used = false;
	/*
	if( IE_CodeVarRead *pCVR = dynamic_cast<IE_CodeVarRead*>(pChild) ){
	  cerr << "ignoring unexpected code var read" << endl;
	  used = true;
	};
	*/

	if( IE_Call *pCall = dynamic_cast<IE_Call*>(pChild) ){
	  used = true;
	};

	if(!used){
	  string name;
	  pChild->GetGName(name);
	  cout << "unknown child of CTS: " << name << endl;
	  throw("unknown child of CTS");
	};
      };
    };
    return;
  };
  
  if( IE_CompTempLoad *pCTL = dynamic_cast<IE_CompTempLoad*>(pBranch) ){
    cerr << "ignoring unexpected cmp temp load" << endl;
    return;
  };

  if( IE_CodeVarRead *pCVR = dynamic_cast<IE_CodeVarRead*>(pBranch) ){
    cerr << "ignoring unexpected code var read" << endl;
    return;
  };

  if( IE_CodeVarAddress *pCVA = dynamic_cast<IE_CodeVarAddress*>(pBranch) ){
    cerr << "ignoring unexpected code var address" << endl;
    return;
  };

  string name;
  pBranch->GetGName(name);
  cout << "unknown branch: " << name << " loc: " << pBranch->loc << endl;
  throw("unknown branch of binary opp");
}

void Compiler::CTSL_Unary(FullOppList &rFOL, std::list<Opp*>::iterator &rOppPlace, IE_Unary *pIEU)
{
  assert(pIEU);
  IE *pSrc = pIEU->pSrc;
  assert(pSrc);

  if( IE_Imm *pImm = dynamic_cast<IE_Imm*>(pSrc) ){
    return; //source of this unary instruction is an immediate. do nothing.
  };

  if( IE_CompTempLoad *pCTL = dynamic_cast<IE_CompTempLoad*>(pSrc) ){
    //cerr << "Compiler::CTSL_Unary ignoring unexpected cmp temp load" << endl;
    return;
  };

  if( IE_CodeVarRead *pCVR = dynamic_cast<IE_CodeVarRead*>(pSrc) ){
    //cerr << "Compiler::CTSL_Unary ignoring unexpected code var read" << endl;
    return;
  };

  if( IE_CodeVarAddress *pCVA = dynamic_cast<IE_CodeVarAddress*>(pSrc) ){
    //cerr << "Compiler::CTSL_Unary ignoring unexpected code var address" << endl;
    return;
  };

  if( IE_CompTemp *pCT = dynamic_cast<IE_CompTemp*>(pSrc) ){
    //hokay. must split the comp temp into two parts.
    //the entire tree of instructions that happens attached to this comp temp branch must be executed before
    //this binary instruction. remove that sequence of instructions from here. In its place add a stub CompTempLoad that
    //reads the proper 
    IE_CompTempLoad *pLoad = new IE_CompTempLoad(pCT->id, pCT->GetType(), REG_A);
    //note we are not stomping on a pointer because we sill hold pCT above.
    pIEU->UpdateSrc(pLoad);

    CompTempStore *pCTS = new CompTempStore(pCT->id);
    pCTS->pSrc = pCT->pSrc;

    pCT->pSrc = 0; //so we don't delete the rest of the branch, only this opp!
    delete pSrc;
    pSrc = 0;
    pCT = 0;

    rOppPlace = rFOL.opps.insert(rOppPlace, pCTS);

    IE *pChild = pCTS->pSrc;
    assert(pChild);
    if( IE_Binary *pOB = dynamic_cast<IE_Binary*>(pChild) ){
      CTSL_Binary(rFOL, rOppPlace, pOB, pOB->pA, 1);
      CTSL_Binary(rFOL, rOppPlace, pOB, pOB->pB, 2);
    }else{
      if( IE_Unary *pU = dynamic_cast<IE_Unary*>(pChild) ){
	CTSL_Unary(rFOL, rOppPlace, pU);
      }else{
	bool used = false;
	if( IE_CodeVarRead *pCVR = dynamic_cast<IE_CodeVarRead*>(pChild) ){
	  cerr << "ignoring unexpected code var read" << endl;
	  used = true;
	};
	
	//if(!used)
	//	{
	string name;
	pChild->GetGName(name);
	cout << "unary unknown child of CTS: " << name << endl;
	throw("unary unknown child of CTS");
	//};
      };
    };
    return;
  };


  string name;
  pSrc->GetGName(name);
  cout << "unknown source: " << name << endl;
  throw("unknown source of unary opp");  
}

void Compiler::BuildMemBlocks(FunctionDef &rFD, VarStack &rVS)
{
  //sweep through program adding variable declerations.
  std::list<Opp*>::iterator cPlace = rFD.opps.begin();
  while(cPlace != rFD.opps.end() )
    {
      Opp* pOpp = *cPlace;
      assert(pOpp);

      if( CodeVarDec *pCVD = dynamic_cast<CodeVarDec*>(pOpp) )
	{
	  cout << "added var var dec" << endl;
	  rFD.declared_vars.decList.AddDecl( ParamDecl(pCVD->var_type, pCVD->name) );
	};

      cPlace++;
    };  

  std::list<unsigned>::iterator dPlace = rVS.ct_list.begin();
  while( dPlace != rVS.ct_list.end() )
    {
      unsigned ct = *dPlace;
      ostringstream name;
      name << rFD.name << "_ct_" << ct;

      rFD.compiler_temps.decList.AddDecl( ParamDecl(P_UINT, name.str() ) );
      dPlace++;
    };  

  std::list<string>::iterator ePlace = rVS.ct_special_list.begin();
  while( ePlace != rVS.ct_special_list.end() )
    {
      rFD.compiler_special_vars.decList.AddDecl( ParamDecl(P_UINT, *ePlace) );
      ePlace++;
    };  
}

void Compiler::EmitCode(FunctionDef &rFD, std::ostream &rOut)
{
  rFD.FillFunctionParams();

  rOut << rFD.name << "\n";
  rOut << ".SEGMENT stack_frame -1\n";
  rFD.function_params.offset = -1; //space for return value
  rFD.declared_vars.offset = rFD.function_params.offset + -1*rFD.function_params.Size();
  rFD.compiler_temps.offset = rFD.declared_vars.offset + -1*rFD.declared_vars.Size();
  rFD.compiler_special_vars.offset = rFD.compiler_temps.offset + -1*rFD.compiler_temps.Size();

  rFD.function_params.EmitStackCode(rOut);
  rFD.declared_vars.EmitStackCode(rOut);
  rFD.compiler_temps.EmitStackCode(rOut);
  rFD.compiler_special_vars.EmitStackCode(rOut);
  rOut << ".SEGMENT_END\n";

  int full_size = 0;
  full_size += 1; //space for return value.
  full_size += rFD.function_params.Size();
  full_size += rFD.declared_vars.Size();
  full_size += rFD.compiler_temps.Size();
  full_size += rFD.compiler_special_vars.Size();

  //copy the stack pointer to reg F.
  rOut << og << "FNSETUP" << "\n";
  rOut << og << "ADDL " << RegToStr(REG_SP) << " " << RegToStr(REG_F) << " 0\n";
  //make enough space on stack for all function's variables
  rOut << og << "LI " << RegToStr(REG_A) << " " << -full_size << '\n';
  rOut << og << "ADD " << RegToStr(REG_A) << " " << RegToStr(REG_SP) << " " << RegToStr(REG_SP) << "\n";

  std::list<Opp*>::iterator cPlace = rFD.opps.begin();
  while(cPlace != rFD.opps.end() )
    {
      Opp* pOpp = *cPlace;
      assert(pOpp);

      EmitCode(rFD, rOut, pOpp);

      cPlace++;
    };

  //restore stack pointer to saved copy (our frame pointer)
  rOut << og << "ADDL " << RegToStr(REG_F) << " " << RegToStr(REG_SP) << " 0\n";
  //just before the return overwrite our frame pointer with the return result as
  //we will not use it again.
  //?or should we overwrite our parameter variables just off the end of the
  //stack as it existed when we were called?
  rOut << og << "RETURN\n";

  rOut << "\n\n";
}

void Compiler::EmitCode(FunctionDef &rFD, std::ostream &rOut, Opp*pOpp)
{
  if( CodeVarDec *pCVD = dynamic_cast<CodeVarDec*>(pOpp) )
    {
      cout << "found code var dec" << endl;
      IE *pChild = pCVD->pValueSrc;
      if(pChild)
	{
	  cerr << "unexpected child of variable dec" << endl;
	};
      return;
    };

  if( CodeVarWrite *pCVW = dynamic_cast<CodeVarWrite*>(pOpp) )
    {
      cout << "found CodeVarWrite" << endl;
      IE *pChild = pCVW->pValueSrc;

      if( IE_Binary *pOB = dynamic_cast<IE_Binary*>(pChild) )
	EmitBinary(rFD, rOut, pOB, REG_C);
      else{
	if( IE_Unary *pUnary = dynamic_cast<IE_Unary*>(pChild) )
	  EmitUnary(rFD, rOut, pUnary, REG_C);
	else
	  EmitFillReg(rFD, rOut, pChild, REG_C);
      };

      bool isGlobal = pCVW->isGlobal;
      ostringstream ctprefix;
      ctprefix << rFD.name << "_ct_";
      if(pCVW->name.substr(0,ctprefix.str().length()) == ctprefix.str() )
	{
	  if(isGlobal)
	    {
	      cout << "HAAAAAACK! global_fixup" << endl;
	      cout << "compiler bug? no way." << endl;
	      isGlobal = false;
	    };
	};

      if(isGlobal)
	rOut << og << "SRI " << RegToStr(REG_C) << " " << pCVW->name << '\n';
      else
	rOut << og << "SRR " << RegToStr(REG_C) << " " << RegToStr(REG_F) << " " << pCVW->name << '\n';
      return;
    };

  if( CodeVarWriteOffset *pCVWO = dynamic_cast<CodeVarWriteOffset*>(pOpp) )
    {
      cout << "found CodeVarWriteOffset" << endl;

      IE *pOffset = pCVWO->pOffsetSrc;
      if( IE_Binary *pOB = dynamic_cast<IE_Binary*>(pOffset) )
	EmitBinary(rFD, rOut, pOB, REG_D);
      else{
	if( IE_Unary *pUnary = dynamic_cast<IE_Unary*>(pOffset) )
	  EmitUnary(rFD, rOut, pUnary, REG_D);
	else
	  EmitFillReg(rFD, rOut, pOffset, REG_D);
      };

      rOut << og << "PUSH " << RegToStr(REG_D) << '\n';

      IE *pChild = pCVWO->pValueSrc;
      if( IE_Binary *pOB = dynamic_cast<IE_Binary*>(pChild) )
	EmitBinary(rFD, rOut, pOB, REG_C);
      else{
	if( IE_Unary *pUnary = dynamic_cast<IE_Unary*>(pChild) )
	  EmitUnary(rFD, rOut, pUnary, REG_C);
	else
	  EmitFillReg(rFD, rOut, pChild, REG_C);
      };

      rOut << og << "PULL " << RegToStr(REG_D) << '\n';
      if(pCVWO->isGlobal)
	rOut << og << "SRR " << RegToStr(REG_C) << " " << RegToStr(REG_D) << " " << pCVWO->name << '\n';
      else{
	rOut << og << "ADD " << RegToStr(REG_F) << " " << RegToStr(REG_D) << " " << RegToStr(REG_D) << '\n';
	rOut << og << "SRR " << RegToStr(REG_D) << " " << RegToStr(REG_F) << " " << pCVWO->name << '\n';
      };
      return;
    };

  if( CompTempStore *pCTS = dynamic_cast<CompTempStore*>(pOpp) )
    {
      cout << "found CompTempStore!" << endl;
      IE *pChild = pCTS->pSrc;
      assert(pChild);
      if( IE_Binary *pOB = dynamic_cast<IE_Binary*>(pChild) )
	EmitBinary(rFD, rOut, pOB, REG_C);
      else{
	if( IE_Unary *pUnary = dynamic_cast<IE_Unary*>(pChild) )
	  EmitUnary(rFD, rOut, pUnary, REG_C);
	else
	  EmitFillReg(rFD, rOut, pChild, REG_C);
      };

      //rOut << og << "SRI " << RegToStr(REG_C) << " " << rFD.name << "_ct_" << pCTS->ct_id << '\n';
      rOut << og << "SRR " << RegToStr(REG_C) << " " << RegToStr(REG_F) << " " << rFD.name << "_ct_" << pCTS->ct_id << '\n';
      return;
    };

  if( OppRegLoadFromVar *pLR = dynamic_cast<OppRegLoadFromVar*>(pOpp) )
    {
      //rOut << og << "LRI " << RegToStr(pLR->reg) << " " << pLR->varName << '\n';
      rOut << og << "LRR " << RegToStr(REG_F) << " " << RegToStr(pLR->reg) << " " << pLR->varName << '\n';
      return;
    };

  if( SkipIfZero *pSkip = dynamic_cast<SkipIfZero*>(pOpp) )
    {
      if( pSkip->inv_logic)
	rOut << og << "SNEQZ " << RegToStr(pSkip->reg) << '\n';
      else
	rOut << og << "SEQZ " << RegToStr(pSkip->reg) << '\n';

      return;
    };

  if( Jump *pJmp = dynamic_cast<Jump*>(pOpp) )
    {
      rOut << og << "JI " << pJmp->target << endl;
      return;
    };

  if( OppCall *pCall = dynamic_cast<OppCall*>(pOpp) )
    {
      int paramOffset = -8;
      std::list<CallArg>::iterator cPlace =  pCall->exprList.exprList.begin();
      while( cPlace != pCall->exprList.exprList.end() )
	{
	  CallArg &rCA = *cPlace;
	  //basically cut and paste from comp temp load
	  rOut << og << "LRR " << RegToStr(REG_F) << " " << RegToStr(REG_A) << " "
	       << rFD.name << "_ct_" << rCA.ctID << '\n';
	  rOut << og << "SRR " << RegToStr(REG_A) << " " << RegToStr(REG_SP) << " "
	       << paramOffset << '\n';
	  paramOffset--;
	  cPlace++;
	};

      rOut << og << "CALL " << pCall->functionName << "\n";
      return;
    };

  if( Target *pTarget = dynamic_cast<Target*>(pOpp) )
    {
      rOut << pTarget->name << "\n";
      return;
    };

  if( OppReturn *pRet = dynamic_cast<OppReturn*>(pOpp) )
    {
      if( !pRet->RetNull() )
	{
	  //pull the compiler temp and save it at the frame pointer.
	  rOut << og << "LRR " << RegToStr(REG_F) << " " << RegToStr(REG_E) << " "
	       << rFD.name << "_ct_" << pRet->ctID << '\n';
	  rOut << og << "SRR " << RegToStr(REG_E) << " " << RegToStr(REG_F)
	       << " 0\n";
	};
      //restore stack pointer to saved copy (our frame pointer)
      rOut << og << "ADDL " << RegToStr(REG_F) << " " << RegToStr(REG_SP) << " 0\n";
      rOut << og << "RETURN" << '\n';
      return;
    };

  string name;
  pOpp->GetGName(name);
  cout << "!!!! UNHANDLED OPP !!!! name: " << name << endl;
  throw("output code: unknown opp");
}

void Compiler::EmitFillReg(FunctionDef &rFD, std::ostream &rOut, IE *pIE, REG_N dest)
{
  if( IE_Imm *pImm = dynamic_cast<IE_Imm*>(pIE) )
    {
      //terrible hack. my eyes are watering.
      rOut << og << "LI " << RegToStr(dest) << " " << (0xFFFFFFFF&(pImm->value)) << '\n';
      //rOut << og << "LI " << RegToStr(dest) << " " << pImm->value << '\n';
      return;
    };
  
  if(IE_CodeVarAddress *pCVR = dynamic_cast<IE_CodeVarAddress*>(pIE) )
    {
      //rOut.flush();
      //assert(false);
      rOut << og << "LI " << RegToStr(dest) << " " << pCVR->name << '\n';
      return;
    }

  if( IE_CodeVarRead *pCVR = dynamic_cast<IE_CodeVarRead*>(pIE) )
    {
      if(pCVR->isGlobal )
	rOut << og << "LRI " << RegToStr(dest) << " " << pCVR->name << '\n';
      else
	rOut << og << "LRR " << RegToStr(REG_F) << " " << RegToStr(dest) << " " << pCVR->name << '\n';
      return;
    };

  if( IE_CompTempLoad *pCTL = dynamic_cast<IE_CompTempLoad*>(pIE) ){
    //rOut << og << "LRI " << RegToStr(dest) << " " << rFD.name << "_ct_" << pCTL->ct_id << '\n';
    rOut << og << "LRR " << RegToStr(REG_F) << " " << RegToStr(dest) << " " << rFD.name << "_ct_" << pCTL->ct_id << '\n';
    return;
  };
  
  if( IE_CodeVarReadOffset *pCVRO = dynamic_cast<IE_CodeVarReadOffset*>(pIE) ){
    IE *pOffset = pCVRO->pOffsetSrc;
    if( IE_Binary *pOB = dynamic_cast<IE_Binary*>(pOffset) )
      EmitBinary(rFD, rOut, pOB, REG_D);
    else{
      if( IE_Unary *pUnary = dynamic_cast<IE_Unary*>(pOffset) )
	EmitUnary(rFD, rOut, pUnary, REG_D);
      else
	EmitFillReg(rFD, rOut, pOffset, REG_D);
    };
    
    if(pCVRO->isGlobal)
      rOut << og << "LRR " << RegToStr(REG_D) << " " << RegToStr(dest) << " " << pCVRO->name << '\n';
    else{
      rOut << og << "ADD " << RegToStr(REG_F) << " " << RegToStr(REG_D) << " " << RegToStr(REG_D) << '\n';
      rOut << og << "LRR " << RegToStr(REG_D) << " " << RegToStr(dest) << " " << pCVRO->name << '\n';
    };
    return;
  };


  if( IE_Call *pCall = dynamic_cast<IE_Call*>(pIE) ){
    list<CallArg>::iterator cPlace = pCall->exprList.exprList.begin();
    int cOffset = 0;
    cOffset -= 7; //skip space for registers
    cOffset -= 1; //skip space for var return
    while(cPlace != pCall->exprList.exprList.end())
      {
	CallArg &rCA = *cPlace;
	//we know all the variables we are trying to load are compiler temps.
	rOut << og << "LRR " << RegToStr(REG_F) << " " << RegToStr(REG_E)
	     << " " << rFD.name << "_ct_" << rCA.ctID << '\n';
	rOut << og << "SRR " << RegToStr(REG_E) << " " << RegToStr(REG_SP) << " " << cOffset << '\n';
	cOffset--;
	cPlace++;
      };
    rOut << og << "CALL " << pCall->functionName << '\n';
    //a return would have been issued to transfer control back here. our registers are restored
    //the end of the calle's stack when it began execution should be offset -7 wrt
    //the current stack.
    rOut << og << "LRR " << RegToStr(REG_SP) << " " << RegToStr(dest) << " -7" << '\n';
    return;
  };

  if( IE_CompTemp *pCT = dynamic_cast<IE_CompTemp*>(pIE) )
    {
      string name;
      pCT->pSrc->GetGName(name);
      cerr << "yeah totaly fucked. name: " << name << pCT->pSrc->loc << endl;
    };

  string name;
  pIE->GetGName(name);
  cout << "uknown reg fill instruction: " << name << " loc: " << pIE->loc << endl;
  throw("unknown reg fill instruction");
}

void Compiler::EmitBinary(FunctionDef &rFD, std::ostream &rOut, IE_Binary *pOB, REG_N dest)
{
  P_TYPE at = pOB->pA->GetType();
  P_TYPE bt = pOB->pB->GetType();

  if( at == P_ERROR || bt == P_ERROR)
    {
      cerr << "WARN: binary opp with error type input!" << endl;
      cerr << "location about: " << pOB->loc << endl;
    }else
    if(at == bt)
      {
	cout << "GOOD!!!  binary opp with good inputs" << endl;
      };

  if( at != bt )
    {
      cerr << "WARN: binary opp with conflicing input types!" << endl;
      cerr << "this probably won't work. location about: " << pOB->loc << endl;
    }

  EmitFillReg(rFD, rOut, pOB->pA, REG_A);
  EmitFillReg(rFD, rOut, pOB->pB, REG_B);

  if( IE_CMP *pCmp = dynamic_cast<IE_CMP*>(pOB) )
    {
      if( at == P_FLOAT && bt == P_FLOAT)
	{
	  cerr << "floating point compare not implemented location about: " << pOB->loc << endl;
	  //technically == would work, but that should not be done on fp values anyway.
	  assert(false);
	};
      enum P_TYPE{P_UINT, P_SINT, P_FLOAT, P_ERROR};

      unsigned do_us_cmp = true;
      if( (at == P_UINT && bt == P_SINT) || (at == P_SINT && bt == P_UINT) )
	{
	  cerr << "ERROR! Compare between different types. doing unsigned comparison! location about: " << pOB->loc << endl;
	}else
	if( at == P_SINT && bt == P_SINT){
	  do_us_cmp = false;
	};

      rOut << og << "CMP " << RegToStr(REG_A) << " " << RegToStr(REG_B) << " " << RegToStr(dest) << '\n';
      unsigned bits_val = 0;
      switch(pCmp->cType)
	{
	case IE_CMP::EQ:
	  bits_val = 1;
	  rOut << og << "LI " << RegToStr(REG_A) << " " << bits_val << '\n';
	  rOut << og << "AND " << RegToStr(REG_A) << " " << RegToStr(dest) << " " << RegToStr(dest) << '\n';
	  return;
	case IE_CMP::NEQ:
	  bits_val = 1;
	  rOut << og << "LI " << RegToStr(REG_A) << " " << bits_val << '\n';
	  rOut << og << "XOR " << RegToStr(REG_A) << " " << RegToStr(dest) << " " << RegToStr(dest) << '\n';
	  return;
	case IE_CMP::GT:
	  bits_val = 2;
	  if(!do_us_cmp) bits_val = 4;
	  rOut << og << "LI " << RegToStr(REG_A) << " " << bits_val << '\n';
	  rOut << og << "AND " << RegToStr(REG_A) << " " << RegToStr(dest) << " " << RegToStr(dest) << '\n';
	  return;
	case IE_CMP::GTE:
	  bits_val = 3;
	  if(!do_us_cmp) bits_val = 5;
	  rOut << og << "LI " << RegToStr(REG_A) << " " << bits_val << '\n';
	  rOut << og << "AND " << RegToStr(REG_A) << " " << RegToStr(dest) << " " << RegToStr(dest) << '\n';
	  return;
	case IE_CMP::LT:
	  bits_val = 8;
	  if(!do_us_cmp) bits_val = 16;
	  rOut << og << "LI " << RegToStr(REG_A) << " " << bits_val << '\n';
	  rOut << og << "AND " << RegToStr(REG_A) << " " << RegToStr(dest) << " " << RegToStr(dest) << '\n';
	  return;
	case IE_CMP::LTE:
	  bits_val = 9;
	  if(!do_us_cmp) bits_val = 17;
	  rOut << og << "LI " << RegToStr(REG_A) << " " << bits_val << '\n';
	  rOut << og << "AND " << RegToStr(REG_A) << " " << RegToStr(dest) << " " << RegToStr(dest) << '\n';
	  return;
	};
      return;
    };

  if( IE_Logic *pLogic = dynamic_cast<IE_Logic*>(pOB) )
    {
      ostringstream ang;
      ang << rFD.name << "_" << pLogic->oppID;
      std::string and_false_label, and_true_label, and_done_label;
      and_true_label = string("and_") + ang.str() + "_true";
      and_false_label = string("and_") + ang.str() + "_false";
      and_done_label = string("and_") + ang.str() + "_done";
      switch(pLogic->cType)
	{
	case IE_Logic::LOGIC_AND:
	  cout << "logic and: " <<  pLogic->oppID << endl;
	  rOut << og << "SNEQZ " << RegToStr(REG_A) << '\n';
	  rOut << og << "JI " << and_false_label << '\n';
	  rOut << og << "SNEQZ " << RegToStr(REG_B) << '\n';
	  rOut << og << "JI " << and_false_label << '\n';
	  rOut << and_true_label << '\n';
	  rOut << og << "LI " << RegToStr(dest) << " 1" << '\n';
	  rOut << og << "JI " << and_done_label << '\n';
	  rOut << and_false_label  << '\n';
	  rOut << og << "LI " << RegToStr(dest) << " 0" << '\n';
          rOut << and_done_label << '\n';
	  break;
	case IE_Logic::LOGIC_OR:
	  rOut << og << "OR " << RegToStr(REG_A) << " " << RegToStr(REG_B) << " " << RegToStr(dest) << '\n'; //if any bit is set in either one result will be non-zero
	};
      return;
    };
  
  //handle the default case:
  if(at == P_FLOAT && bt == P_FLOAT)
    {
      if( dynamic_cast<IE_Mul*>(pOB) )
	{
	  rOut << og << "CALL fp_mul" << '\n';
	  rOut << og << "LRR %H " << RegToStr(dest) << " -7" << '\n';
	  return;
	};

      if( dynamic_cast<IE_Div*>(pOB) )
	{
	  rOut << og << "CALL fp_div" << '\n';
	  rOut << og << "LRR %H " << RegToStr(dest) << " -7" << '\n';
	  return;
	};

      if( dynamic_cast<IE_Add*>(pOB) )
	{
	  rOut << og << "CALL fp_add" << '\n';
	  rOut << og << "LRR %H " << RegToStr(dest) << " -7" << '\n';
	  return;
	};

      if( dynamic_cast<IE_Sub*>(pOB) )
	{
	  rOut << og << "CALL fp_sub" << '\n';
	  rOut << og << "LRR %H " << RegToStr(dest) << " -7" << '\n';
	  return;
	};

      throw("unknown binary op with fp args");
    }else
    rOut << og << pOB->GetANM() << " " << RegToStr(REG_A) << " " << RegToStr(REG_B) << " " << RegToStr(dest) << '\n';
}

void Compiler::EmitUnary(FunctionDef &rFD, std::ostream &rOut, IE_Unary *pUnary, REG_N dest)
{
  P_TYPE it = pUnary->pSrc->GetType();

  if( it == P_ERROR )
    {
      cerr << "WARN: unary opp with error type input!" << endl;
      cerr << "location about: " << pUnary->loc << endl;
    }

  EmitFillReg(rFD, rOut, pUnary->pSrc, REG_A);

  if( IE_Shift *pShift = dynamic_cast<IE_Shift*>(pUnary) )
    {
      if(it != P_UINT && it != P_SINT)
	{
	  cerr << "WARN: shifting a non-integer type! " << pUnary->loc << endl; 
	  throw("can't handle non-integer unary input type");
	};
      rOut << og << pShift->GetANM() << " " << RegToStr(REG_A) << " " << RegToStr(dest) << " " << pShift->offset << '\n';
      return;
    };

  if( IE_Bin_Not *pNot = dynamic_cast<IE_Bin_Not*>(pUnary) )
    {
      if(it != P_UINT && it != P_SINT)
	{
	  cerr << "WARN: shifting a non-integer type! " << pUnary->loc << endl; 
	  throw("can't handle non-integer unary input type");
	};
      rOut << og << "NOT " << RegToStr(REG_A) << " " << RegToStr(dest) << '\n';
      return;
    };

  if( IE_Cast *pCast = dynamic_cast<IE_Cast*>(pUnary) )
    {
      P_TYPE ot = pUnary->GetType();
      if( (it == P_UINT || it == P_SINT ) && (ot == P_UINT || ot == P_SINT ) ) //integer to integer. just copy
	{
	  rOut << og << "ADDL " << RegToStr(REG_A) << " " << RegToStr(dest) << " 0" << '\n';
	}else{
	if( it == P_FLOAT && (ot == P_SINT || ot == P_UINT) ) //int to float
	  {
	    //float to int cast
	    //throw("can't handle float to int cast");
	    rOut << og << "CALL fp_cast_to_int" << '\n';
	    rOut << og << "LRR %H " << RegToStr(dest) << " -7" << '\n';
	  }else
	  if( (it == P_UINT || it == P_SINT ) && ot == P_FLOAT )
	    {
	      //int to float cast
	      //throw("can't handle int to float cast");
	      rOut << og << "CALL fp_cast_to_float" << '\n';
	      rOut << og << "LRR %H " << RegToStr(dest) << " -7" << '\n';
	    }else
	    {
	      cerr << "FATAL: unknown cast! " << it << " " << ot << endl;
	      throw("unknown cast");
	    };
      };
      return;
    };

  throw("unhandled unary instruction emit");
} 


std::string Compiler::RegToStr(REG_N reg)
{
  switch(reg)
    {
    case REG_A:
      return string("%A");
    case REG_B:
      return string("%B");
    case REG_C:
      return string("%C");
    case REG_D:
      return string("%D");
    case REG_E:
      return string("%E");
    case REG_F:
      return string("%F");
    case REG_SP:
      return string("%H");
    default:
      return string("%err");
    };
}
