/*
  Copyright 2010, 2014
  John Culp, All Rights Reserved
*/

#ifndef COMPILER_INCLUDED
#define COMPILER_INCLUDED

#include <iostream>

#include "Program.hh"

//a container for the transforms that can be applied to the program
//for now the transforms must be applied in order.

class Compiler
{
public:
  Compiler(Program &rTProg);
  ~Compiler(void);

  std::string OG(void);

  void DoIt(void);

  void AssignIDs(void);

  //first tranvere through program and insert compiler temps between all computational opps.
  void InsertCompilerTemps(FullOppList &rFD);

  //for every binary opp, skip the load to reg instruction for each argument
  //and (recursivly) convert each branch to a straight line. then emit the
  //two register load instructions, perform and perform the opp
  void ConvertTreeToStraightLine(const std::string &rFnName, FullOppList &rFOL);

  //emit code. at this stage the program should be a list of very short
  //trees like store->binaryopp->arga->loadfromreg argb->loadfromreg
  //we need to recognize the patterns and emit to assembly language
  //only simple patterns need to be recognized; unrecognized patterns
  //are compiler errors.
  void EmitCode(FunctionDef &rFD, std::ostream &rOut);
  
protected:

  void InsertCompilerTemps_Chain(Opp &rOpp);
  void InsertCompilerTemps_Chain(IE &rIE);
  bool VerifyBinaryArg(IE &rOpp);

  void CTTSL_Start(const std::string &rFnName, FullOppList &rFOL, std::list<Opp*>::iterator &oppPlace);
  void CTTSL_Handle_Opp(const std::string &rFnName, FullOppList &rFOL, Opp* pOpp, std::list<Opp*>::iterator &oppPlace);
  void CTSL_Binary(FullOppList &rFOL, std::list<Opp*>::iterator &rOppPlace, IE_Binary *pOB,
		   IE *pBranch, int branch);
  void CTSL_Unary(FullOppList &rFOL, std::list<Opp*>::iterator &rOppPlace, IE_Unary *pIEU);

  void BuildMemBlocks(FunctionDef &rFD, VarStack &rVS); //MemBlock &rDefMB, MemBlock &rCTMB,
  //		      MemBlock &rSpecialMB);

  void EmitCode(FunctionDef &rFD, std::ostream &rOut, Opp*pOpp);
  void EmitBinary(FunctionDef &rFD, std::ostream &rOut, IE_Binary *pOB, REG_N dest);
  void EmitUnary(FunctionDef &rFD, std::ostream &rOut, IE_Unary *pUnary, REG_N dest);
  void EmitFillReg(FunctionDef &rFD, std::ostream &rOut, IE *pIE, REG_N dest);

  std::string RegToStr(REG_N);

  Program &rProg;

  std::string og; //output gap.
};


#endif //COMPILER_INCLUDED
