/*
  Copyright 2010, 2014
  John Culp, All Rights Reserved
*/

#include <assert.h>
#include <iostream>
#include <algorithm>

#include "ExecState.hh"

using namespace std;

MemOpp::MemOpp(MO_TYPE tMOType, unsigned long tAddr, DTYPE tValue) :
  moType(tMOType), addr(tAddr), value(tValue)
{}

MemOpp::MemOpp(const MemOpp &rhs) :
  moType(rhs.moType), addr(rhs.addr), value(rhs.value)
{}

MemOpp::~MemOpp(void)
{}

const MemOpp& MemOpp::operator=(const MemOpp &rhs)
{
  moType  = rhs.moType;
  addr = rhs.addr;
  value = rhs.value;
  
  return *this;
}

bool MemOpp::operator==(const MemOpp &rhs) const
{
  return moType == rhs.moType && addr == rhs.addr &&  value == rhs.value;
}

bool MemOpp::operator!=(const MemOpp &rhs) const
{
  return !( (*this) == rhs);
}

void MemOpp::Describe(ostream &rOut) const
{
  rOut << "addr: " << addr;

  if(moType == MO_STORE)
    rOut << " STORE: ";
  else
    rOut << " LOAD: ";

  rOut << value;
}

// *********************

bool CompareMOLists(const std::list<MemOpp> &rOne,
		    const std::list<MemOpp> &rTwo)
{
  list<MemOpp>::const_iterator cPlace = rOne.begin();
  list<MemOpp>::const_iterator dPlace = rTwo.begin();
  while( cPlace != rOne.end() && dPlace != rTwo.end() )
    {
      if( *cPlace != *dPlace )
	{
	  cerr << "Memory opps differ! " << endl;
	  (*cPlace).Describe(cerr);
	  (*dPlace).Describe(cerr);
	  return false;
	};

      cPlace++;
      dPlace++;
    };

  if( cPlace != rOne.end() || dPlace != rTwo.end() )
    {
      cerr << "memory opp lists not the same length" << endl;
      return false;
    };

  return true;
}

// *********************

ExecState::ExecState(void) :
  recordMO(false), ct_vars(65535), memory(65535), stack(65535)
{
  Reset(); //just for kicks and giggles
}

ExecState::~ExecState(void)
{}

void ExecState::Reset(void)
{
  ClearMOList();

  fill( memory.begin(), memory.end(), 0 );
  fill( stack.begin(), stack.end(), 0 );
}

DTYPE ExecState::Load(unsigned long address)
{
  assert( address < memory.size() );
  DTYPE value = memory[address];

  if( recordMO )
    moList.push_back( MemOpp(MemOpp::MO_LOAD, address, value) );

  return value;
}

void ExecState::Store(unsigned long address, DTYPE value)
{
  assert( address < memory.size() );

  cout << "storing value: " << value << " at address: " << address << endl;

  if( recordMO )
    moList.push_back( MemOpp(MemOpp::MO_STORE, address, value) );

  memory[address] = value;
}

void ExecState::StoreCT(unsigned long address, DTYPE value)
{
  assert( address < ct_vars.size() );

  cout << "storing stack value: " << value << " at address: " << address << endl;
  ct_vars[address] = value;
}

DTYPE ExecState::LoadCT(unsigned long address)
{
  assert( address < ct_vars.size() );
  DTYPE value = ct_vars[address];

  return value;
}

void ExecState::ClearMOList(void)
{
  moList.clear();
}

DTYPE ExecState::StackLoad(unsigned long address)
{
  assert( address < stack.size() );
  DTYPE value = stack[address];

  return value;
}

void ExecState::StackStore(unsigned long address, DTYPE value)
{
  assert( address < stack.size() );

  cout << "storing stack value: " << value << " at address: " << address << endl;
  stack[address] = value;
}
