/*
  Copyright 2010, 2014
  John Culp, All Rights Reserved
*/

#ifndef EXECSTATE_INCLUDED
#define EXECSTATE_INCLUDED

#include <iostream>
#include <vector>
#include <list>

//everything is represented as this, all datatypes are either
//32 bit signed integers, 32 bit unsigned integers, or 32 bit fp values
//of course this isn't actually 32 bits on x64.
//hack. hack. hack. fix later. too late to touch now.
typedef unsigned long DTYPE;

//not the best place to define this. hardware defined registers
//procedures return values in register F. see processor documentation.
enum REG_N{REG_A, REG_B, REG_C, REG_D, REG_E, REG_F, REG_AT, REG_SP, REG_INVALID};

class MemOpp //a record for testing transformation
{
public:
  enum MO_TYPE{MO_LOAD, MO_STORE};
  MemOpp(MO_TYPE tMOType, unsigned long tAddr, DTYPE tValue);
  MemOpp(const MemOpp &rhs);
  ~MemOpp(void);

  const MemOpp& operator=(const MemOpp &rhs);

  bool operator==(const MemOpp &rhs) const;
  bool operator!=(const MemOpp &rhs) const;
  
  void Describe(std::ostream &rOut) const;

  MO_TYPE moType;
  unsigned long addr;
  DTYPE value;
};

bool CompareMOLists(const std::list<MemOpp> &rOne,
		    const std::list<MemOpp> &rTwo);

class ExecState
{
public:
  ExecState(void);
  virtual ~ExecState(void);

  void Reset(void); //Clears the MOlist too

  bool recordMO;
  DTYPE Load(unsigned long address);
  void Store(unsigned long address, DTYPE value);

  void StoreCT(unsigned long address, DTYPE value);
  DTYPE LoadCT(unsigned long address);
  std::vector<DTYPE> ct_vars;

  void ClearMOList(void);
  std::list<MemOpp> moList;

  std::vector<DTYPE> memory;

  //keeping the stack as a seperate memory allows comparisons
  //of the main ram after execution between transforms of the program.
  //the exact machinations of the executing code on/to the stack are
  //irrelevant (provided the results are correct).

  DTYPE StackLoad(unsigned long address);
  void StackStore(unsigned long address, DTYPE value);
  std::vector<DTYPE> stack;


  /* registers */
  DTYPE regA;
  DTYPE regB;
  DTYPE regC;
  DTYPE regD;
  DTYPE regE;
  DTYPE regF; //?stack frame?
  DTYPE AT; //assembler temp
  DTYPE SP; //stack pointer
};

#endif //EXECSTATE_INCLUDED
