/*
  Copyright 2014
  John Culp, All Rights Reserved
*/

#include <assert.h>
#include <iostream>

#include "FullOppList.hh"

#include "Opps.hh"

using namespace std;

FullOppList::FullOppList(void)
{
  //cout << "FullOppList::FullOppList" << endl;
}

FullOppList::FullOppList(const Opp &rOpp)
{
  //cout << "FullOppList::FullOppList with opp" << endl;
  AddStatement(rOpp);
}

FullOppList::FullOppList(const FullOppList &rhs)
{
  //cout << "FullOppList::FullOppList copy constructor" << endl;
  *this = rhs;
}

FullOppList::~FullOppList(void)
{
  //cout << "FullOppList::~FullOppList" << endl;
  Clear();
}

const FullOppList& FullOppList::operator=(const FullOppList &rhs)
{
  //cout << "full opp list operator=" << endl;
  Clear();

  list<Opp*>::const_iterator cPlace = rhs.opps.begin();
  while(cPlace != rhs.opps.end() )
    {
      Opp* pOpp = *cPlace;
      assert(pOpp);
      AddStatement(*pOpp);
      cPlace++;

    };

  return *this;
}

void FullOppList::AddStatement(const Opp &rOpp)
{
  opps.push_back( rOpp.Clone() );
}

void FullOppList::Exec(ExecState &rState)
{
  list<Opp*>::const_iterator cPlace = opps.begin();
  while(cPlace != opps.end() )
    {
      Opp* pOpp = *cPlace;
      assert(pOpp);
      pOpp->Exec( rState );
      cPlace++;
    };
}

void FullOppList::Clear(void)
{
  while( !opps.empty() )
    {
      cout << "popping item" << endl;
      Opp* pOpp = opps.front();
      opps.pop_front();
      delete pOpp;
      pOpp = 0;
    };
}

void FullOppList::PrintTree(unsigned prevID, PT_State &rPTS)
{

  unsigned cID = prevID;
  bool ignore_edge = true;
  list<Opp*>::iterator cPlace = opps.begin();
  while(cPlace != opps.end() )
    {
      Opp* pOpp = *cPlace;
      assert(pOpp);
      pOpp->PrintTree(cID, rPTS);

      if(!ignore_edge)
	{
	  rPTS.rOut << "edge [color=red];\n";
	  rPTS.rOut << prevID << " -> " << cID << '\n';
	  rPTS.rOut << "edge [color=black];\n";
	}else
	ignore_edge = false;

      prevID = cID;
      cID = rPTS.GetNextID();

      cPlace++;
    };
}

void FullOppList::AssignUniqueIDs(ID_Gen &rGen)
{
  list<Opp*>::iterator cPlace = opps.begin();
  while(cPlace != opps.end() )
    {
      Opp* pOpp = *cPlace;
      assert(pOpp);

      pOpp->AssignUniqueIDs(rGen);

      cPlace++;
    };  
}

void FullOppList::PushVarTypes(VarStack &rDefs)
{
  list<Opp*>::iterator cPlace = opps.begin();
  while(cPlace != opps.end() )
    {
      Opp* pOpp = *cPlace;
      assert(pOpp);

      pOpp->PushVarTypes(rDefs);

      cPlace++;
    };
}

void FullOppList::AssignProperName(const std::string &rName)
{
  cout << "FullOppList::AssignProperName" << endl;
  if(opps.empty() || rName.empty() )
    {
      assert(false);
      return;
    };

  Opp *pOpp =  opps.front();
  if( CodeVarWrite *pCVW = dynamic_cast<CodeVarWrite*>(pOpp) )
    {
      pCVW->name = rName;
      pCVW->isGlobal = false;
    }else
    assert(false);
}

void FullOppList::DebugTestProperName(void)
{
  assert(!opps.empty());
  Opp *pOpp =  opps.front();
  CodeVarWrite *pCVW = dynamic_cast<CodeVarWrite*>(pOpp);
  assert(pCVW);

  assert( !pCVW->isGlobal);
}

