/*
  Copyright 2010, 2014
  John Culp, All Rights Reserved
*/

#include <assert.h>
#include <sstream>

#include "IE_Call.hh"

#include "Opps.hh"

using namespace std;

CallArg::CallArg(void) :
  FullOppList(), ctID(0)
{}

CallArg::CallArg(const IE &rIE) :
  FullOppList( CodeVarWrite("", rIE) ), ctID(0)
{}

CallArg::CallArg(const CallArg &rhs) :
  FullOppList(rhs), ctID(rhs.ctID)
{}

CallArg::~CallArg(void)
{}

const CallArg& CallArg::operator=(const CallArg &rhs)
{
  *static_cast<FullOppList*>(this) = static_cast<const FullOppList&>(rhs);
  ctID = rhs.ctID;

  return *this;
}


// **********************

ExprList::ExprList(void)
{
  //cout << "ExprList::ExprList" << endl;
}

ExprList::ExprList(const IE &rIE)
{
  //cout << "ExprList::ExprList with IE" << endl;
  AddIE(rIE);
}  

ExprList::ExprList(const ExprList &rhs)
{
  //cout << "ExprList::ExprList copyc 1: " << rhs.exprList.size() << endl;
  *this = rhs;
  //cout << "ExprList::ExprList copyc 2: " << rhs.exprList.size() << endl;
  //cout << "ExprList::ExprList copyc 3: " << exprList.size() << endl;
}
 
ExprList::~ExprList(void)
{
  Clear();
}

const ExprList& ExprList::operator=(const ExprList&rhs)
{
  Clear();
  
  list<CallArg>::const_iterator cPlace = rhs.exprList.begin();
  while(cPlace != rhs.exprList.end())
    {
      const CallArg &rCA = (*cPlace);
      exprList.push_back(rCA);
      cPlace++;
    };

  cout << "ExprList::ExprList operator = " << exprList.size() << endl;

  return *this;
}

void ExprList::AddIE(const IE &rIE)
{
  exprList.push_back( CallArg(rIE) );
  //cout << "ExprList::AddIE size: " << exprList.size() << endl;
}

void ExprList::Clear(void)
{
  exprList.clear();
  /*
  while( !exprList.empty() )
    {
      IE *pIE = exprList.front();
      exprList.pop_front();
      delete pIE;
      pIE = 0;
    }
  */
}

// ********************************

IE_Call::IE_Call(const std::string &rFnName) :
  IE(), ct_id(0), functionName(rFnName)
{}

IE_Call::IE_Call(const std::string &rFnName, ExprList &rEL) :
  IE(), ct_id(0), functionName(rFnName), exprList(rEL)
{}

IE_Call::IE_Call(const IE_Call &rhs) :
  IE(rhs), ct_id(rhs.ct_id), functionName(rhs.functionName), exprList(rhs.exprList)
{}

IE_Call::~IE_Call(void)
{}

IE* IE_Call::Clone(void) const
{
  return new IE_Call(*this);
}

void IE_Call::GetGName(std::string &rName)
{
  rName = "IE_Call";
}

DTYPE IE_Call::Exec(ExecState &rState)
{
  assert(false);
}

void IE_Call::PrintTree(unsigned myID, PT_State &rPTS)
{
  string type;
  GetGName(type);

  rPTS.rOut << myID << " [label=\"" << type << "\"];\n";

  //cout << "IE_Call::PrintTree size: " << exprList.exprList.size() << endl;

  list<CallArg>::iterator cPlace = exprList.exprList.begin();
  while( cPlace != exprList.exprList.end() )
    {
      CallArg &rCA = *cPlace;
      unsigned childID = rPTS.GetNextID();
      rCA.PrintTree(childID, rPTS);
      rPTS.rOut << childID << " -> " << myID << '\n';

      cPlace++;
    };
}

void IE_Call::AssignUniqueIDs(ID_Gen &rGen)
{
  ct_id = rGen.GetCTID();

  list<CallArg>::iterator cPlace = exprList.exprList.begin();
  while( cPlace != exprList.exprList.end() )
    {
      CallArg &rCA = *cPlace;
      rCA.ctID = rGen.GetCTID();
      rCA.AssignUniqueIDs(rGen);
      cPlace++;
    };
}

void IE_Call::PushVarTypes(VarStack &rDefs)
{
  cout << "IE_Call::PushVarTypes" << endl;
  rDefs.RegisterCT( ct_id );

  list<CallArg>::iterator cPlace = exprList.exprList.begin();
  while( cPlace != exprList.exprList.end() )
    {
      CallArg &rCA = *cPlace;

      ostringstream name;
      name << rDefs.fnName << "_ct_" << rCA.ctID;

      rCA.AssignProperName( name.str() );
      rDefs.RegisterCT( rCA.ctID );
      
      rCA.PushVarTypes(rDefs);
      cPlace++;
    };
}

P_TYPE IE_Call::GetType(void)
{
  return P_FLOAT;
}


