/*
  Copyright 2010, 2014
  John Culp, All Rights Reserved
*/

#include <assert.h>
#include <string.h>
#include <sstream>

#include "IExpr.hh"

using namespace std;

/*
ExecRes::ExecRes(void) :
  valid(false)
{}

ExecRes::ExecRes(DTYPE tvalue) :
  valid(true), value(tvalue)
{};

ExecRes::~ExecRes(void)
{}

void ExecRes::GetValue(void)
{
  if(!valid) assert(false);
  return value;
}
*/



// **************

// *****************

IE::IE(void)
{}

IE::IE(const IE &rhs) :
  loc(rhs.loc)
{}

IE::~IE(void)
{}

DTYPE IE::Exec(ExecState &rState)
{
  return 100005; //random number
}

void IE::PrintTree(unsigned myID, PT_State &rPTS)
{
  //cout << "unimplemented print tree" << endl;
  string type;
  GetGName(type);

  rPTS.rOut << myID << " [label=\"" << type << "\"];\n";
}

void IE::AssignUniqueIDs(ID_Gen &rGen)
{
  string type;
  GetGName(type);

  cout << "IE::AssignUniqueIDs default type: " << type << endl;
}

void IE::PushVarTypes(VarStack &)
{}

void IE::SetLoc(const TVCC::location &rloc)
{
  loc = rloc;
}

// **************
// **************

IE_Binary::IE_Binary(const IE &rA, const IE &rB) :
  IE(), pA( rA.Clone() ), pB( rB.Clone() )
{}

IE_Binary::IE_Binary(const IE_Binary &rhs) :
  IE(rhs)
{
  pA = rhs.pA->Clone();
  pB = rhs.pB->Clone();
}

IE_Binary::~IE_Binary(void)
{
  delete pA; pA = 0;
  delete pB; pB = 0;
}

void IE_Binary::PrintTree(unsigned myID, PT_State &rPTS)
{
  unsigned aID = rPTS.GetNextID();
  unsigned bID = rPTS.GetNextID();

  string type;
  GetGName(type);

  rPTS.rOut << myID << " [label=\"" << type << "\"];\n";

  rPTS.rOut << aID << " -> " << myID << '\n';
  rPTS.rOut << bID << " -> " << myID << '\n';

  pA->PrintTree(aID, rPTS);
  pB->PrintTree(bID, rPTS);
}

void IE_Binary::UpdateA(IE *pNewA)
{
  pA = pNewA;
}

void IE_Binary::UpdateB(IE *pNewB)
{
  pB = pNewB;
}

void IE_Binary::AssignUniqueIDs(ID_Gen &rGen)
{
  pA->AssignUniqueIDs(rGen);
  pB->AssignUniqueIDs(rGen);
}

void IE_Binary::PushVarTypes(VarStack &rDefs)
{
  pA->PushVarTypes(rDefs);
  pB->PushVarTypes(rDefs);
}

std::string IE_Binary::GetANM(void) const
{
  return string("error_default");
}

P_TYPE IE_Binary::GetType(void)
{
  P_TYPE aType = pA->GetType();
  P_TYPE bType = pB->GetType();

  if( aType == bType ) return aType;

  return P_ERROR;
}

// ***************

IE_Add::IE_Add(const IE &rA, const IE &rB) :
  IE_Binary(rA, rB)
{}

IE_Add::IE_Add(const IE_Add &rhs) :
  IE_Binary(rhs)
{}

IE_Add::~IE_Add(void)
{}

IE* IE_Add::Clone(void) const
{
  return new IE_Add(*this);
}

DTYPE IE_Add::Exec(ExecState &rState)
{
  assert(pA);
  assert(pB);

  return pA->Exec(rState) + pB->Exec(rState);
}

void IE_Add::GetGName(std::string &rName)
{
  rName = "ADD";
}

std::string IE_Add::GetANM(void) const
{
  return string("ADD");
}

// ***************

IE_Sub::IE_Sub(const IE &rA, const IE &rB) :
  IE_Binary(rA, rB)
{}

IE_Sub::IE_Sub(const IE_Sub &rhs) :
  IE_Binary(rhs)
{}

IE_Sub::~IE_Sub(void)
{}

IE* IE_Sub::Clone(void) const
{
  return new IE_Sub(*this);
}

DTYPE IE_Sub::Exec(ExecState &rState)
{
  assert(pA);
  assert(pB);

  return pA->Exec(rState) - pB->Exec(rState);
}

void IE_Sub::GetGName(std::string &rName)
{
  rName = "SUB";
}

std::string IE_Sub::GetANM(void) const
{
  return string("SUB");
}

// ***************

IE_Mul::IE_Mul(const IE &rA, const IE &rB) :
  IE_Binary(rA, rB)
{}

IE_Mul::IE_Mul(const IE_Mul &rhs) :
  IE_Binary(rhs)
{}

IE_Mul::~IE_Mul(void)
{}

IE* IE_Mul::Clone(void) const
{
  return new IE_Mul(*this);
}

DTYPE IE_Mul::Exec(ExecState &rState)
{
  assert(pA);
  assert(pB);

  return pA->Exec(rState) * pB->Exec(rState);
}

void IE_Mul::GetGName(std::string &rName)
{
  rName = "MUL";
}

std::string IE_Mul::GetANM(void) const
{
  return string("MUL");
}

// **************

IE_Div::IE_Div(const IE &rA, const IE &rB) :
  IE_Binary(rA, rB)
{}

IE_Div::IE_Div(const IE_Div &rhs) :
  IE_Binary(rhs)
{}

IE_Div::~IE_Div(void)
{}

IE* IE_Div::Clone(void) const
{
  return new IE_Div(*this);
}

DTYPE IE_Div::Exec(ExecState &rState)
{
  assert(pA);
  assert(pB);

  return pA->Exec(rState) / pB->Exec(rState);
}

std::string IE_Div::GetANM(void) const
{
  return string("DIV");
}

void IE_Div::GetGName(std::string &rName)
{
  rName = "DIV";
}

// *********************

IE_CMP::IE_CMP(const IE &rA, const IE &rB, CMP_TYPE tcType) :
  IE_Binary(rA, rB), cType(tcType)
{}

IE_CMP::IE_CMP(const IE_CMP &rhs) :
  IE_Binary(rhs), cType(rhs.cType)
{}

IE_CMP::~IE_CMP(void)
{}

IE* IE_CMP::Clone(void) const
{
  return new IE_CMP(*this);
}

DTYPE IE_CMP::Exec(ExecState &rState)
{}

void IE_CMP::GetGName(std::string &rName)
{
  ostringstream name;
  name << "CMP_" << cType;
}

P_TYPE IE_CMP::GetType(void)
{
  return P_SINT;
}

// *********************

IE_Logic::IE_Logic(const IE &rA, const IE &rB, LOGIC_OP tcType) :
  IE_Binary(rA, rB), cType(tcType), oppID(0)
{
  cout << "IE_Logic::IE_Logic" << endl;
}

IE_Logic::IE_Logic(const IE_Logic &rhs) :
  IE_Binary(rhs), cType(rhs.cType), oppID(rhs.oppID)
{
  cout << "IE_Logic::IE_Logic copy" << endl;
}

IE_Logic::~IE_Logic(void)
{}

IE* IE_Logic::Clone(void) const
{
  return new IE_Logic(*this);
}

DTYPE IE_Logic::Exec(ExecState &rState)
{}

void IE_Logic::GetGName(std::string &rName)
{
  ostringstream name;
  name << "LOGIC_OPP_" << cType;
}

P_TYPE IE_Logic::GetType(void)
{
  return P_SINT;
}

void IE_Logic::AssignUniqueIDs(ID_Gen &rGen)
{
  oppID = rGen.GetID();
  cout << "IE_Logic::AssignUniqueIDs assigned: " << oppID << endl;
  IE_Binary::AssignUniqueIDs(rGen);
  cout << "IE_Logic::AssignUniqueIDs assigned: " << oppID << endl;
}

// *********************

IE_Bin_Opp::IE_Bin_Opp(const IE &rA, const IE &rB, BIN_OP tcType) :
  IE_Binary(rA, rB), cType(tcType)
{}

IE_Bin_Opp::IE_Bin_Opp(const IE_Bin_Opp &rhs) :
  IE_Binary(rhs), cType(rhs.cType)
{}

IE_Bin_Opp::~IE_Bin_Opp(void)
{}

IE* IE_Bin_Opp::Clone(void) const
{
  return new IE_Bin_Opp(*this);
}

DTYPE IE_Bin_Opp::Exec(ExecState &rState)
{

}

void IE_Bin_Opp::GetGName(std::string &rName)
{
  ostringstream name;
  name << "BIN_OP_" << cType; 
}

std::string IE_Bin_Opp::GetANM(void) const
{
  switch(cType)
    {
    case BIN_AND:
      return string("AND");
    case BIN_OR:
      return string("OR");
    case BIN_XOR:
      return string("XOR");
    };
}

P_TYPE IE_Bin_Opp::GetType(void)
{
  return P_SINT;
}

// ******************************* end binary IExprs

IE_Unary::IE_Unary(const IE &rIn) :
  pSrc( rIn.Clone() )
{}

IE_Unary::IE_Unary(const IE_Unary &rhs) :
  IE(rhs), pSrc( rhs.pSrc->Clone() )
{}

IE_Unary::~IE_Unary(void)
{
  delete pSrc; pSrc = 0;
}

void IE_Unary::AssignUniqueIDs(ID_Gen &rGen)
{
  pSrc->AssignUniqueIDs(rGen);
}
  
void IE_Unary::PushVarTypes(VarStack &rDefs)
{
  pSrc->PushVarTypes(rDefs);
}

void IE_Unary::PrintTree(unsigned myID, PT_State &rPTS)
{
  unsigned srcID = rPTS.GetNextID();

  string type;
  GetGName(type);

  rPTS.rOut << myID << " [label=\"" << type << "\"];\n";
  rPTS.rOut << srcID << " -> " << myID << '\n';

  pSrc->PrintTree(srcID, rPTS);
}

void IE_Unary::UpdateSrc(IE *pNewSrc)
{
  pSrc = pNewSrc;
}

// ************

IE_Bin_Not::IE_Bin_Not(const IE &rIn) :
  IE_Unary(rIn)
{}

IE_Bin_Not::IE_Bin_Not(const IE_Bin_Not &rhs) :
  IE_Unary(rhs)
{}

IE_Bin_Not::~IE_Bin_Not(void)
{}

IE* IE_Bin_Not::Clone(void) const
{
  return new IE_Bin_Not(*this);
}

void IE_Bin_Not::GetGName(std::string &rName)
{
  rName = "NOT";
}

DTYPE IE_Bin_Not::Exec(ExecState &rState)
{}

P_TYPE IE_Bin_Not::GetType(void)
{
  return P_SINT;
}

// *******************************

IE_Cast::IE_Cast(P_TYPE t_dest_type, const IE &rIn) :
  IE_Unary(rIn), dest_type(t_dest_type)
{}

IE_Cast::IE_Cast(const IE_Cast &rhs) :
  IE_Unary(rhs), dest_type(rhs.dest_type)
{}

IE_Cast::~IE_Cast(void)
{}

IE* IE_Cast::Clone(void) const
{
  return new IE_Cast(*this);
}

void IE_Cast::GetGName(std::string &rName)
{
  ostringstream name;
  name << "Cast_to " << dest_type;

  rName = name.str();
}

DTYPE IE_Cast::Exec(ExecState &rState)
{}

P_TYPE IE_Cast::GetType(void)
{
  return dest_type;
}

// *******************************

IE_Shift::IE_Shift(const IE &rIn, unsigned tOffset, SHIFT_DIR tDir) :
  IE_Unary(rIn), offset(tOffset), dir(tDir)
{}

IE_Shift::IE_Shift(const IE_Shift &rhs) :
  IE_Unary(rhs), offset(rhs.offset), dir(rhs.dir)
{}

IE_Shift::~IE_Shift(void)
{}

IE* IE_Shift::Clone(void) const
{
  return new IE_Shift(*this);
}

void IE_Shift::GetGName(std::string &rName)
{
  rName = "shift_";
}

DTYPE IE_Shift::Exec(ExecState &rState)
{}

std::string IE_Shift::GetANM(void) const
{
  switch(dir)
    {
    case RIGHT:
      return string("RSR");
    case LEFT:
      return string("RSL");
    };
  return string("ERROR");
}

P_TYPE IE_Shift::GetType(void)
{
  return P_SINT;
}

// ******************************* end binary IExprs

IE_Imm::IE_Imm(DTYPE tValue, P_TYPE tType) :
  IE(), value(tValue), reg(REG_INVALID), myType(tType)
{}

IE_Imm::IE_Imm(float f_value) :
  IE(), value(0), reg(REG_INVALID), myType(P_FLOAT)
{
  memcpy(&value, &f_value, 4);
}

IE_Imm::IE_Imm(const IE_Imm &rhs) :
  IE(rhs), value(rhs.value), reg(rhs.reg), myType(rhs.myType)
{}

IE_Imm::~IE_Imm(void)
{}

IE* IE_Imm::Clone(void) const
{
  return new IE_Imm(*this);
}

DTYPE IE_Imm::Exec(ExecState &rState)
{
  return value;
}

void IE_Imm::GetGName(std::string &rName)
{
  ostringstream sstream;
  sstream << "IMM value: " << value;
  rName = sstream.str();
}

P_TYPE IE_Imm::GetType(void)
{
  return myType;
}

// ***************

// **************

IE_CodeVarRead::IE_CodeVarRead(const std::string &rName) :
  IE(), name(rName), isGlobal(false), myType(P_ERROR)
{
  cout << "IE_CodeVarRead::IE_CodeVarRead" << endl;
}

IE_CodeVarRead::IE_CodeVarRead(const IE_CodeVarRead &rhs) :
  IE(rhs), name(rhs.name), isGlobal(rhs.isGlobal), myType(rhs.myType)
{
  cout << "IE_CodeVarRead::IE_CodeVarRead copy" << endl;
}

IE_CodeVarRead::~IE_CodeVarRead(void)
{}

IE* IE_CodeVarRead::Clone(void) const
{
  return new IE_CodeVarRead(*this);
}

void IE_CodeVarRead::GetGName(std::string &rName)
{
  rName = "CodeVarRead";
}

DTYPE IE_CodeVarRead::Exec(ExecState &rState)
{

}

void IE_CodeVarRead::PushVarTypes(VarStack &rDefs)
{
  std::string context;
  myType = rDefs.GetVarType(name, context, isGlobal);
  if(!context.empty())
    name = context + "_" + name;
}

P_TYPE IE_CodeVarRead::GetType(void)
{
  return myType;
}

// **************

IE_CodeVarReadOffset::IE_CodeVarReadOffset(const std::string &rName, const IE &rOffset) :
  IE(), name(rName), isGlobal(false), pOffsetSrc( rOffset.Clone() ), myType(P_ERROR)
{}
  
IE_CodeVarReadOffset::IE_CodeVarReadOffset(const IE_CodeVarReadOffset &rhs) :
  IE(rhs), name(rhs.name), isGlobal(rhs.isGlobal), pOffsetSrc( rhs.pOffsetSrc->Clone() ), myType(rhs.myType)
{}

IE_CodeVarReadOffset::~IE_CodeVarReadOffset(void)
{
  delete pOffsetSrc;
  pOffsetSrc = 0;
}

IE* IE_CodeVarReadOffset::Clone(void) const
{
  return new IE_CodeVarReadOffset(*this);
}

void IE_CodeVarReadOffset::GetGName(std::string &rName)
{
  rName = "CodeVarReadOffset";
}

DTYPE IE_CodeVarReadOffset::Exec(ExecState &rState)
{}

void IE_CodeVarReadOffset::AssignUniqueIDs(ID_Gen &rGen)
{
  pOffsetSrc->AssignUniqueIDs(rGen);
}

void IE_CodeVarReadOffset::PushVarTypes(VarStack &rDefs)
{
  std::string context;
  myType = rDefs.GetVarType(name, context, isGlobal);
  if(!context.empty())
    name = context + "_" + name;

  pOffsetSrc->PushVarTypes(rDefs);
}

P_TYPE IE_CodeVarReadOffset::GetType(void)
{
  return myType;
}

// **************


IE_CodeVarAddress::IE_CodeVarAddress(const std::string &rName) :
  IE(), name(rName)
{}

IE_CodeVarAddress::IE_CodeVarAddress(const IE_CodeVarAddress &rhs) :
  IE(rhs), name(rhs.name)
{}

IE_CodeVarAddress::~IE_CodeVarAddress(void)
{}

IE* IE_CodeVarAddress::Clone(void) const
{
  return new IE_CodeVarAddress(*this);
}

void IE_CodeVarAddress::GetGName(std::string &rName)
{
  ostringstream ostr;
  ostr << "Code_Var_Address: " << name;
  rName = ostr.str();
}

P_TYPE IE_CodeVarAddress::GetType(void)
{
  return P_SINT;
}

// **************

IE_CompTemp::IE_CompTemp(IE *ptSrc) :
  IE(), id(0), pSrc(ptSrc), stackAddr(0)
{}

IE_CompTemp::IE_CompTemp(const IE_CompTemp &rhs) :
  IE(rhs), id(rhs.id), pSrc(rhs.pSrc), stackAddr(rhs.stackAddr)
{}

IE_CompTemp::~IE_CompTemp(void)
{
  delete pSrc;
  pSrc = 0;
}

IE* IE_CompTemp::Clone(void) const
{
  return new IE_CompTemp(*this);
}

void IE_CompTemp::GetGName(std::string &rName)
{
  rName = "CT";
}

DTYPE IE_CompTemp::Exec(ExecState &rState)
{
  return pSrc->Exec(rState);
}

void IE_CompTemp::PrintTree(unsigned myID, PT_State &rPTS)
{
  string type;
  GetGName(type);

  unsigned srcID =  rPTS.GetNextID();
  rPTS.rOut << srcID << " -> " << myID << '\n';
  rPTS.rOut << myID << " [label=\"" << type << "_" << id << "\"];\n";

  pSrc->PrintTree(srcID, rPTS);
}

void IE_CompTemp::AssignUniqueIDs(ID_Gen &rGen)
{
  id = rGen.GetCTID();
  pSrc->AssignUniqueIDs(rGen);
}

void IE_CompTemp::PushVarTypes(VarStack &rDefs)
{
  rDefs.RegisterCT(id);
  pSrc->PushVarTypes(rDefs);
}

P_TYPE IE_CompTemp::GetType(void)
{
  return pSrc->GetType();
}

// ****************

IE_CompTempLoad::IE_CompTempLoad(unsigned t_ct_id, P_TYPE tType, REG_N treg) :
  IE(), ct_id(t_ct_id), var_type(tType), reg(treg)
{}

IE_CompTempLoad::IE_CompTempLoad(const IE_CompTempLoad &rhs) :
  IE(rhs), ct_id(rhs.ct_id), var_type(rhs.var_type), reg(rhs.reg)
{}

IE_CompTempLoad::~IE_CompTempLoad(void)
{}

IE* IE_CompTempLoad::Clone(void) const
{
  return new IE_CompTempLoad(*this);
}

void IE_CompTempLoad::GetGName(std::string &rName)
{
  ostringstream out;
  out << "CTL_" << ct_id << "_" << reg;
  rName = out.str();
}

DTYPE IE_CompTempLoad::Exec(ExecState &rState)
{
  DTYPE value = rState.LoadCT(ct_id);

  return value;
}

P_TYPE IE_CompTempLoad::GetType(void)
{
  return var_type;
}

// **************

