/*
  Copyright 2010, 2014
  John Culp, All Rights Reserved
*/

#ifndef IEXPR_INCLUDED
#define IEXPR_INCLUDED

#include "ExecState.hh"
#include "PT_State.hh"
#include "ID_Gen.hh"
#include "TVCC_Defs.hh"
#include "VarStack.hh"
#include "location.hh"
#include "FullOppList.hh"

//all IExpr's return some value when executed.
//Every IE owns the pointers to the objects within them

/*
class ExecRes
{
public:
  ExecRes(void); //makes invalid
  ExecRes(DTYPE value); //makes valid
  virtual ~ExecRes(void);

  DTYPE GetValue(void); //asserts false if !valid

  bool valid;
  DTYPE value;
};
*/

class IE
{
public:
  IE(void);
  IE(const IE &rhs);
  virtual ~IE(void);

  virtual IE* Clone(void) const = 0;
  virtual void GetGName(std::string &rName) = 0; //get graph name

  //all opps return some value (even if they don't... ?bug in future?)
  //returns of void result in the return of value 0
  virtual DTYPE Exec(ExecState &rState);
  virtual void PrintTree(unsigned myID, PT_State &rPTS);
  virtual void AssignUniqueIDs(ID_Gen &rGen);
  virtual void PushVarTypes(VarStack &rDefs);
  virtual P_TYPE GetType(void) = 0; //returns the type of data returned in DTYPE
  virtual void SetLoc(const TVCC::location &rloc);

    TVCC::location loc;
private:
  const IE& operator=(const IE &rhs);
};

// ***************

// ************************* start binary expressions ************************

class IE_Binary : public IE //one of many instructions that take two arguments and return a value.
{
public:
  virtual ~IE_Binary(void);
protected:
  IE_Binary(const IE &rA, const IE &rB);
  IE_Binary(const IE_Binary &rhs);

public: //? or friend compiler?
  void UpdateA(IE *pNewA);
  void UpdateB(IE *pNewB);

  virtual void PrintTree(unsigned myID, PT_State &rPTS);
  virtual std::string GetANM(void) const; //assembler mnemonic
  virtual void AssignUniqueIDs(ID_Gen &rGen);
  virtual void PushVarTypes(VarStack &rDefs);
  virtual P_TYPE GetType(void);

  IE *pA, *pB;
  //ValueSrc aSrc, bSrc;
private:
  const IE_Binary& operator=(const IE_Binary &rhs);
};

class IE_Add : public IE_Binary
{
public:
  IE_Add(const IE &rA, const IE &rB);
  IE_Add(const IE_Add &rhs);
  virtual ~IE_Add(void);

  virtual IE* Clone(void) const;
  virtual DTYPE Exec(ExecState &rState);
  virtual void GetGName(std::string &rName);
  virtual std::string GetANM(void) const;

private:
  const IE_Add& operator=(const IE_Add &rhs);
};

class IE_Sub : public IE_Binary
{
public:
  IE_Sub(const IE &rA, const IE &rB);
  IE_Sub(const IE_Sub &rhs);
  virtual ~IE_Sub(void);

  virtual IE* Clone(void) const;
  virtual DTYPE Exec(ExecState &rState);
  virtual void GetGName(std::string &rName);
  virtual std::string GetANM(void) const;

private:
  const IE_Sub& operator=(const IE_Sub &rhs);
};

class IE_Mul : public IE_Binary
{
public:
  IE_Mul(const IE &rA, const IE &rB);
  IE_Mul(const IE_Mul &rhs);
  virtual ~IE_Mul(void);

  virtual IE* Clone(void) const;
  virtual DTYPE Exec(ExecState &rState);
  virtual void GetGName(std::string &rName);
  virtual std::string GetANM(void) const;

private:
  const IE_Mul& operator=(const IE_Mul &rhs);
};

class IE_Div : public IE_Binary
{
public:
  IE_Div(const IE &rA, const IE &rB);
  IE_Div(const IE_Div &rhs);
  virtual ~IE_Div(void);

  virtual IE* Clone(void) const;
  virtual DTYPE Exec(ExecState &rState);
  virtual void GetGName(std::string &rName);
  virtual std::string GetANM(void) const;

private:
  const IE_Div& operator=(const IE_Div &rhs);
};

class IE_CMP : public IE_Binary
{
public:
  enum CMP_TYPE{EQ, NEQ, GT, GTE, LT, LTE};

  IE_CMP(const IE &rA, const IE &rB, CMP_TYPE tcType);
  IE_CMP(const IE_CMP &rhs);
  virtual ~IE_CMP(void);


  virtual IE* Clone(void) const;
  virtual DTYPE Exec(ExecState &rState);
  virtual void GetGName(std::string &rName);
  //most of the comparisons return non-zero or zero, unlike
  //'c' where I think the results are guarenteed to be 0 or 1
  virtual P_TYPE GetType(void);

  CMP_TYPE cType;

private:
  const IE_CMP& operator=(const IE_CMP &rhs);
};

class IE_Logic : public IE_Binary
{
public:
  enum LOGIC_OP{LOGIC_AND, LOGIC_OR}; //ie && and ||

  IE_Logic(const IE &rA, const IE &rB, LOGIC_OP tcType);
  IE_Logic(const IE_Logic &rhs);
  virtual ~IE_Logic(void);

  virtual IE* Clone(void) const;
  virtual DTYPE Exec(ExecState &rState);
  virtual void GetGName(std::string &rName);
  //just going to return a sint.
  virtual P_TYPE GetType(void);
  virtual void AssignUniqueIDs(ID_Gen &rGen);

  unsigned oppID; //used for and
  LOGIC_OP cType;

private:
  const IE_Logic& operator=(const IE_Logic &rhs);
};

class IE_Bin_Opp : public IE_Binary
{
public:
  enum BIN_OP{BIN_AND, BIN_OR, BIN_XOR}; //ie & and | and ^

  IE_Bin_Opp(const IE &rA, const IE &rB, BIN_OP tcType);
  IE_Bin_Opp(const IE_Bin_Opp &rhs);
  virtual ~IE_Bin_Opp(void);


  virtual IE* Clone(void) const;
  virtual DTYPE Exec(ExecState &rState);
  virtual void GetGName(std::string &rName);
  virtual std::string GetANM(void) const;
  //just going to return a sint.
  virtual P_TYPE GetType(void);

  BIN_OP cType;

private:
  const IE_Bin_Opp& operator=(const IE_Bin_Opp &rhs);
};

class IE_Unary : public IE
{
protected:
  IE_Unary(const IE &rIn);
  IE_Unary(const IE_Unary &rhs);
  virtual ~IE_Unary(void);

public:
  virtual void AssignUniqueIDs(ID_Gen &rGen);
  virtual void PushVarTypes(VarStack &rDefs);
  virtual void PrintTree(unsigned myID, PT_State &rPTS);
  
  void UpdateSrc(IE *pNewSrc);
  IE *pSrc;
  
private:
  const IE_Unary& operator=(const IE_Unary &rhs);
};

class IE_Bin_Not : public IE_Unary
{
public:
  IE_Bin_Not(const IE &rIn);
  IE_Bin_Not(const IE_Bin_Not &rhs);
  virtual ~IE_Bin_Not(void);

  virtual IE* Clone(void) const;
  virtual void GetGName(std::string &rName);
  virtual DTYPE Exec(ExecState &rState);
  //just going to return a sint.
  virtual P_TYPE GetType(void);

private:
  const IE_Bin_Not& operator=(const IE_Bin_Not &rhs);
};

class IE_Cast : public IE_Unary
{
public:
  IE_Cast(P_TYPE t_dest_type, const IE &rIn);
  IE_Cast(const IE_Cast &rhs);
  virtual ~IE_Cast(void);

  virtual IE* Clone(void) const;
  virtual void GetGName(std::string &rName);
  virtual DTYPE Exec(ExecState &rState);
  virtual P_TYPE GetType(void);

  P_TYPE dest_type;

private:
  const IE_Cast& operator=(const IE_Cast &rhs);
};

class IE_Shift : public IE_Unary //a unary for us because shift amount is a constant
{
public:
  enum SHIFT_DIR{RIGHT, LEFT};

  IE_Shift(const IE &rIn, unsigned tOffset, SHIFT_DIR tDir);
  IE_Shift(const IE_Shift &rhs);
  virtual ~IE_Shift(void);

  virtual IE* Clone(void) const;
  virtual void GetGName(std::string &rName);
  virtual DTYPE Exec(ExecState &rState);
  virtual std::string GetANM(void) const;
  //just going to return a sint.
  //we are always going to do a logical shift however!!
  //I don't think I should care about what type of data I recieve as
  //long as it is integer
  virtual P_TYPE GetType(void);

  unsigned offset;
  SHIFT_DIR dir;

private:
  const IE_Bin_Not& operator=(const IE_Bin_Not &rhs);
};

//IE_Immediate
class IE_Imm : public IE
{
public:
  IE_Imm(DTYPE tValue, P_TYPE tType);
  IE_Imm(float value);
  IE_Imm(const IE_Imm &rhs);
  virtual ~IE_Imm(void);

  virtual IE* Clone(void) const;
  virtual void GetGName(std::string &rName);
  virtual DTYPE Exec(ExecState &rState);
  //type is whatever parser tells us
  virtual P_TYPE GetType(void);

  DTYPE value;
  REG_N reg;
  P_TYPE myType;
private:
  const IE_Imm& operator=(const IE_Imm &rhs);
};

class IE_CodeVarRead : public IE
{
public:
  IE_CodeVarRead(const std::string &rName);
  IE_CodeVarRead(const IE_CodeVarRead &rhs);
  virtual ~IE_CodeVarRead(void);

  virtual IE* Clone(void) const;
  virtual void GetGName(std::string &rName);

  virtual DTYPE Exec(ExecState &rState);
  virtual void PushVarTypes(VarStack &rDefs);
  //type is whatever parser defined that type as
  virtual P_TYPE GetType(void);

  std::string name; //name changed by PushVarTypes
  bool isGlobal; //established via PushVarTypes
  P_TYPE myType; //established via PushVarTypes
private:
  const IE_CodeVarRead& operator=(const IE_CodeVarRead &rhs);
};

class IE_CodeVarReadOffset : public IE
{
public:
  IE_CodeVarReadOffset(const std::string &rName, const IE &rOffset);
  IE_CodeVarReadOffset(const IE_CodeVarReadOffset &rhs);
  virtual ~IE_CodeVarReadOffset(void);

  virtual IE* Clone(void) const;
  virtual void GetGName(std::string &rName);

  virtual DTYPE Exec(ExecState &rState);
  virtual void AssignUniqueIDs(ID_Gen &rGen);
  virtual void PushVarTypes(VarStack &rDefs);
  //type is whatever the type is that we are reading an offset from?
  virtual P_TYPE GetType(void);

  std::string name;
  bool isGlobal;
  IE *pOffsetSrc;
  P_TYPE myType;

private:
  const IE_CodeVarReadOffset& operator=(const IE_CodeVarReadOffset &rhs);
};

class IE_CodeVarAddress : public IE
{
public:
  IE_CodeVarAddress(const std::string &rName);
  IE_CodeVarAddress(const IE_CodeVarAddress &rhs);
  virtual ~IE_CodeVarAddress(void);

  virtual IE* Clone(void) const;
  virtual void GetGName(std::string &rName);
  //just going to return a sint. it will be easier to perform pointer
  //arithmatic if I don't have to worry about sign conventions
  virtual P_TYPE GetType(void);

  std::string name;

private:
  const IE_CodeVarAddress& operator=(const IE_CodeVarAddress &rhs);
};

// an autogenerated variable that might be on the stack
// (or it may be optimized away)
// it is placed between opps; it is a place-holder for first few passes
// until register allocation?
class IE_CompTemp : public IE
{
public:
  IE_CompTemp(IE *ptSrc);
  IE_CompTemp(const IE_CompTemp &rhs);
  virtual ~IE_CompTemp(void);

  virtual IE* Clone(void) const;
  virtual void GetGName(std::string &rName);
  virtual DTYPE Exec(ExecState &rState);
  virtual void PrintTree(unsigned myID, PT_State &rPTS);
  virtual void AssignUniqueIDs(ID_Gen &rGen);
  virtual void PushVarTypes(VarStack &rDefs);
  //returns the type of the assigned expression
  virtual P_TYPE GetType(void);

  unsigned id;
  IE *pSrc; //the opp that generates the value that is on the stack
  unsigned long stackAddr; //not assigned initially

private:
  const IE_CompTemp& operator=(const IE_CompTemp &rhs);
};

class IE_CompTempLoad : public IE
{
public:
  IE_CompTempLoad(unsigned t_ct_id, P_TYPE tType, REG_N treg);
  IE_CompTempLoad(const IE_CompTempLoad &rhs);
  virtual ~IE_CompTempLoad(void);

  virtual IE* Clone(void) const;
  virtual void GetGName(std::string &rName);
  virtual DTYPE Exec(ExecState &rState);
  //returns the previously established type of the compiler temporary
  virtual P_TYPE GetType(void);

  unsigned ct_id;
  P_TYPE var_type;
  REG_N reg;

private:
  const IE_CompTempLoad& operator=(const IE_CompTempLoad &rhs);
};

#endif //IEXPR_INCLUDED
