/*
  Copyright 2014
  John Culp, All Rights Reserved
*/

#include <assert.h>
#include <fstream>

#include "Program.hh"

using namespace std;

MemBlock::MemBlock(void)
{}

MemBlock::MemBlock(int toffset, const ParamDeclList &rPDList) :
  offset(toffset), decList(rPDList)
{}

MemBlock::MemBlock(const MemBlock &rhs) :
  offset(rhs.offset), decList(rhs.decList)
{}

MemBlock::~MemBlock(void)
{}

const MemBlock & MemBlock::operator=(const MemBlock &rhs)
{
  offset = rhs.offset;
  decList = rhs.decList;
  return *this;
}

void MemBlock::EmitCodePrefix(std::ostream &rOut)
{
  rOut << ".segment data_1\n";
}

int MemBlock::Size(void)
{
  int size = 0;
  std::list<ParamDecl>::iterator cPlace = decList.paramList.begin();
  while( cPlace != decList.paramList.end() )
    {
      size += (*cPlace).size;
      cPlace++;
    };

  return size;
}

void MemBlock::EmitCode(std::ostream &rOut)
{
  
  //rOut << rFD.name << ":\n";
  int cAddress = offset;
  rOut << ".ORG " << cAddress << "\n";

  std::list<ParamDecl>::iterator cPlace = decList.paramList.begin();
  while( cPlace != decList.paramList.end() )
    {
      const ParamDecl &rPD = *cPlace;
      rOut << rPD.name << endl;
      cAddress += rPD.size;
      //if( (*cPlace).size != 1 )
      rOut << ".ORG " << cAddress << "\n";
	
      cPlace++;
    };

  rOut << "\n\n";
}

void MemBlock::EmitStackCode(std::ostream &rOut)
{
  int cAddress = offset;
  rOut << ".ORG " << cAddress << "\n";

  std::list<ParamDecl>::iterator cPlace = decList.paramList.begin();
  while( cPlace != decList.paramList.end() )
    {
      const ParamDecl &rPD = *cPlace;
      rOut << rPD.name << endl;
      if( (rPD.size != 1) || true ){
	cAddress -= rPD.size;
	rOut << ".ORG " << cAddress << "\n";
      }else
	cAddress -= rPD.size;

      cPlace++;
    };
}

// ******************

FunctionDef::FunctionDef(void) :
  FullOppList()
{}

FunctionDef::FunctionDef(const std::string &rTName, const ParamDeclList &rPDL, const FullOppList &rFOL) :
  FullOppList(rFOL), args(rPDL), name(rTName)
{}

FunctionDef::FunctionDef(const std::string &rTName, const FullOppList &rFOL) :
  FullOppList(rFOL), name(rTName)
{}

FunctionDef::FunctionDef(const FunctionDef &rhs)
{
  *this = rhs;
}

FunctionDef::~FunctionDef(void)
{}

const FunctionDef& FunctionDef::operator=(const FunctionDef &rhs)
{
  Clear();

  name = rhs.name;
  idGen = rhs.idGen;

  args = rhs.args;

  function_params = function_params;
  declared_vars = declared_vars;
  compiler_temps = compiler_temps;
  compiler_special_vars = compiler_special_vars;

  static_cast<FullOppList&>(*this) = static_cast<const FullOppList&>(rhs);

  return *this;
}

void FunctionDef::FillFunctionParams(void)
{
  list<ParamDecl>::iterator cPlace = args.paramList.begin();
  while( cPlace != args.paramList.end() )
    {
      ParamDecl &rPD = *cPlace;
      function_params.decList.paramList.push_back( ParamDecl(rPD.type, name + "_" + rPD.name) );
      cPlace++;
    };
}

void FunctionDef::FillLocalsFromParameters(VarStack &rVS)
{
  list<ParamDecl>::iterator cPlace = args.paramList.begin();
  while( cPlace != args.paramList.end() )
    {
      ParamDecl &rPD = *cPlace;
      rVS.AddVar( STV_Def(rPD.type, rPD.name) );
      cPlace++;
    };
}

// ******************

Program::Program(void)
{}

Program::~Program(void)
{};

void Program::Exec(ExecState &rEState)
{
  //search for a function with the correct entry name, then execute that.

  /*
  //for_each(pgm.opps.begin(), pgm.opps.end(), Exec(execState) );
  list<Opp*>::const_iterator cPlace = opps.begin();
  while(cPlace != opps.end() )
    {
      Opp* pOpp = *cPlace;
      assert(pOpp);
      pOpp->Exec( rEState );
      cPlace++;
    };
  */
}

void Program::PrintTree(const std::string &rName)
{
  cout << "Program::PrintTree 1" << endl;
  ofstream jjunk( rName.c_str() );
  if(!jjunk) return;

  cout << "Program::PrintTree 2" << endl;

  PT_State pts(jjunk);

  list<FunctionDef>::iterator cPlace = fdList.begin();
  while(cPlace != fdList.end())
    {
      cout << "in here" << endl;
      FunctionDef &rFD = *cPlace;

      pts.Start();

      unsigned prevID = pts.GetNextID(); //def is not rooted
      pts.rOut << prevID << '\n';
      pts.rOut << prevID << " [label=\"" << "START_" << "wrong_name" << "\"];\n";

      rFD.PrintTree(prevID, pts);

      pts.End();

      cPlace++;
    };

  jjunk.close();
}

void Program::FillGlobalsFromMBList(VarStack &rVS)
{
  unsigned ii = 0;
  std::list<MemBlock>::iterator cPlace = mbList.begin();
  while(cPlace != mbList.end())
    {
      MemBlock &rBlk = *cPlace;

      std::list<ParamDecl>::iterator dPlace = rBlk.decList.paramList.begin();
      while(dPlace != rBlk.decList.paramList.end() )
	{
	  ParamDecl &rPD = *dPlace;
	  rVS.AddVar( STV_Def(rPD.type, rPD.name) );
	  dPlace++;
	};


      ii++;
      cPlace++;
    };
}

void Program::PrintBlocks(void)
{
  unsigned ii = 0;
  std::list<MemBlock>::iterator cPlace = mbList.begin();
  while(cPlace != mbList.end())
    {
      MemBlock &rBlk = *cPlace;

      cout << "block number: " << ii << endl;
      
      std::list<ParamDecl>::iterator dPlace = rBlk.decList.paramList.begin();
      while(dPlace != rBlk.decList.paramList.end() )
	{
	  ParamDecl &rPD = *dPlace;
	  cout << ParamDecl::Translate(rPD.type) << " " << rPD.name << endl;
	  dPlace++;
	};


      cout << endl << endl;
      ii++;
      cPlace++;
    };
}
