/*
  Copyright 2014
  John Culp, All Rights Reserved
*/

#ifndef PROGRAM_INCLUDED
#define PROGRAM_INCLUDED

#include <list>
#include <string>

#include "Opps.hh"
#include "ExecState.hh"
#include "TVCC_Parser_Objs.hh"
#include "FullOppList.hh"
#include "ID_Gen.hh"
#include "VarStack.hh"

class MemBlock
{
public:
  MemBlock(void);
  MemBlock(int toffset, const ParamDeclList &rPDList);
  MemBlock(const MemBlock &rhs);
  ~MemBlock(void);

  const MemBlock & operator=(const MemBlock &rhs);

  int Size(void);

  //generates a definition block.
  void EmitCodePrefix(std::ostream &rOut);
  void EmitCode(std::ostream &rOut);
  void EmitStackCode(std::ostream &rOut);

  int offset;
  ParamDeclList decList;
};

class FunctionDef : public FullOppList
{
public:
  FunctionDef(void);
  //hack. we should steal list contents, not copy from object about to be destructed
  FunctionDef(const std::string &rTName, const ParamDeclList &rPDL, const FullOppList &rFOL);
  FunctionDef(const std::string &rTName, const FullOppList &rFOL);
  FunctionDef(const FunctionDef &rhs);
  ~FunctionDef(void);

  const FunctionDef& operator=(const FunctionDef &rhs);

  //fills the MemBlock function_params with args
  void FillFunctionParams(void);
  void FillLocalsFromParameters(VarStack &rVS);

  ID_Gen idGen; //every function has its own chunk of variables
  
  ParamDeclList args;

  std::string name;
  MemBlock function_params;
  MemBlock declared_vars;
  MemBlock compiler_temps;
  MemBlock compiler_special_vars;
};

class Program
{
public:
  Program(void);
  virtual ~Program(void);

  void Exec(ExecState &rEState);
  void PrintTree(const std::string &rName);
  void PrintTree(FunctionDef &rFD, PT_State &rPTS);
  void FillGlobalsFromMBList(VarStack &rVS);

  void PrintBlocks(void);

  std::list<MemBlock> mbList; //for globals
  std::list<FunctionDef> fdList;
};

#endif //PROGRAM_INCLUDED
