/*
  Copyright 2014
  John Culp, All Rights Reserved
*/

#include <assert.h>
#include <iostream>

#include "VarStack.hh"

using namespace std;

STV_Def::STV_Def(void) :
  var_type(P_ERROR)
{}

STV_Def::STV_Def(P_TYPE ttype, const std::string &rName) :
  var_type(ttype), name(rName)
{}

STV_Def::STV_Def(const STV_Def &rhs) :
  var_type(rhs.var_type), name(rhs.name)
{}

STV_Def::~STV_Def(void)
{}

const STV_Def & STV_Def::operator=(const STV_Def &rhs)
{
  var_type = rhs.var_type;
  name = rhs.name;

  return *this;
}

// ***********************

VarContext::VarContext(const std::string &rName) :
  name(rName)
{}

VarContext::VarContext(const VarContext &rhs) :
  varDefs(rhs.varDefs), name(rhs.name)
{}

VarContext::~VarContext(void)
{}

VarContext& VarContext::operator=(const VarContext &rhs)
{
  varDefs = rhs.varDefs;
  name = rhs.name;

  return *this;
}

P_TYPE VarContext::FindVar(const std::string &rName)
{
  std::list<STV_Def>::reverse_iterator cPlace = varDefs.rbegin();
  while( cPlace != varDefs.rend() )
    {
      if( rName == (*cPlace).name) return (*cPlace).var_type;

      cPlace++;
    };

  return P_ERROR;
}


// ***********************

VarStack::VarStack(const std::string &rFnName) :
  fnName(rFnName)
{}

VarStack::VarStack(const VarStack& rhs) :
  cStack(rhs.cStack), ct_list(rhs.ct_list), ct_special_list(rhs.ct_special_list)
{}

VarStack::~VarStack(void)
{}

bool VarStack::AddVar(const STV_Def &rVar)
{
  assert( !cStack.empty() );

  list<VarContext>::reverse_iterator cPlace = cStack.rbegin();
  while(cPlace != cStack.rend() )
    {
      P_TYPE ret = (*cPlace).FindVar(rVar.name);
      if( ret != P_ERROR )
	{
	  cerr << "variable name collision. name: " << rVar.name << endl;
	  return false;
	};

      cPlace++;
    }

  cStack.back().varDefs.push_back( rVar );
  //cout << "VarStack::AddVar" << endl;
  return true;
}

P_TYPE VarStack::GetVarType(const std::string &rName, std::string &rContext, bool &rIsGlobal)
{
  if( VarIsSpecialCT(rName) )
    {
      rContext = "";
      rIsGlobal = false; //by definition not.
      return P_SINT; //hack?
    };

  list<VarContext>::reverse_iterator cPlace = cStack.rbegin();
  while(cPlace != cStack.rend() )
    {
      VarContext &rVC = *cPlace;
      P_TYPE ret = rVC.FindVar(rName);
      if( ret != P_ERROR )
	{
	  if( rVC.name == "global" ) //really tacky
	    {
	      rContext = "";
	      rIsGlobal = true;
	    }else
	    {
	      rContext = fnName;
	      rIsGlobal = false;
	    };
	  return ret;
	};

      cPlace++;
    }

  rContext = "";
  rIsGlobal = true; //probably an error.
  return P_ERROR;
}

void VarStack::PushContext(const std::string &rName)
{
  cStack.push_back( VarContext(rName) );
}

void VarStack::PopContext(void)
{
  if( !cStack.empty() )
    cStack.pop_back();
}

void VarStack::Print(ostream &rOut)
{
  list<VarContext>::reverse_iterator cPlace = cStack.rbegin();
  while(cPlace != cStack.rend() )
    {
      VarContext &rVC = *cPlace;
      
      rOut << "stack level: " << rVC.name << endl;
      
      cPlace++;
    }
}


void VarStack::RegisterCT(unsigned ct_id)
{
  cout << "VarStack::RegisterCT num: " << ct_id << endl;
  ct_list.push_back(ct_id);
}
 
void VarStack::RegisterCT(const std::string &rCT)
{
  cout << "VarStack::RegisterCT string" << endl;
  ct_special_list.push_back(rCT);
}

bool VarStack::VarIsSpecialCT(const std::string &rName)
{
  list<string>::iterator cPlace = ct_special_list.begin();
  while(cPlace != ct_special_list.end())
    {
      if(*cPlace == rName) return true;
      cPlace++;
    };

  return false;
}
