/*
  Copyright 2014
  John Culp, All Rights Reserved
*/

#ifndef VARSTACK_INCLUDED
#define VARSTACK_INCLUDED

#include <string>
#include <list>

#include "TVCC_Defs.hh"

// stack variable definition
class STV_Def
{
public:
  STV_Def(void);
  STV_Def(P_TYPE ttype, const std::string &rName);
  STV_Def(const STV_Def &rhs);
  virtual ~STV_Def(void);

  const STV_Def & operator=(const STV_Def &rhs);

  P_TYPE var_type;
  std::string name;
};

class VarContext
{
public:
  VarContext(const std::string &rName);
  VarContext(const VarContext &rhs);
  ~VarContext(void);

  VarContext& operator=(const VarContext &rhs);

  //if returns type P_ERROR var by that name is not found.
  P_TYPE FindVar(const std::string &rName);

  std::list<STV_Def> varDefs;
  std::string name;
};

class VarStack
{
public:
  VarStack(const std::string &rFnName);
  VarStack(const VarStack& rhs);
  ~VarStack(void);

  //will warn about name collisions
  //refuses to add variables with type P_ERROR
  bool AddVar(const STV_Def &rVar);

  //returns P_ERROR if var name not found. prints a warning.
  //if the variable is in the global context or not defined
  //rContext is set to a zero length string.
  //rIsGlobal is set the the variable is a global.
  P_TYPE GetVarType(const std::string &rName, std::string &rContext, bool &rIsGlobal);

  void PushContext(const std::string &rName);
  void PopContext(void);

  void Print(std::ostream &rOut);

  void RegisterCT(unsigned ct_id);
  void RegisterCT(const std::string &rCT); //used by while/for/if for special names

  //hack. lists should be accessed by functions
  //bigger hack: no types associated with ct_specials?
  std::list<unsigned> ct_list;
  std::list<std::string> ct_special_list;
  bool VarIsSpecialCT(const std::string &rName);


  std::list<VarContext> cStack;
  std::string fnName; //used as the context for all variables declared in the function
private:
  const VarStack& operator=(const VarStack &rhs); //not implemented
};

#endif //VARSTACK_INCLUDED
