/*
  Copyright 2010, 2014
  John Culp, All Rights Reserved
*/

#include <assert.h>
#include <iostream>
#include <fstream>
#include <algorithm>

#include "TVCC_Lexer.hh"
#include "TVCC_Parser.hh"

#include "ExecState.hh"
#include "Program.hh"
#include "Compiler.hh"

using namespace std;

int DoIt(TVCC_Parser &rMyParser, Program **pPgm, ostream &rOut);

int main(int argc, char* argv[])
{
  /*
    where this mess started :)
    Program pgm;

    //*0 = 12*(3+4) - 5 -- store value 79 at address 0
    pgm.opps.push_back( new OppStore(0, OppSub( OppMul( OppImm(12), OppAdd( OppImm(3), OppImm(4)) ), OppImm(5) ) ) );
    pgm.opps.push_back( new OppStore(0, OppImm(5) ) );
  */

  if(argc != 3)
    {
      cerr << "improper usage.\n"
	   << "tvcc <infile> <outfile>" << endl;
      return -1;
    }

  //string inName(argv[1]);
  ifstream fin(argv[1]);
  if(!fin)
    {
      cerr << "unable to open input file: " << argv[1] << endl;
      return -1;
    };

  ofstream fout(argv[2]);
  if(!fout)
    {
      cerr << "unable to open output file: " << argv[2] << endl;
      return -2;
    };

  TVCC_Lexer myLexer;
  TVCC_Parser myparser(&myLexer);
  myLexer.switch_streams(&fin, &cout);

  Program *pPgm;
  bool error = false;

  int res = -5;
  try{
    res = DoIt(myparser, &pPgm, fout);
  }catch(const char *pErr)
    {
      cerr << "caught error: " << pErr << endl;
      error = true;
    };

  if(error && pPgm )
    pPgm->PrintTree("jjunk-err.graph");

  fin.close();

  fout.close();

  return res;
};

int DoIt(TVCC_Parser &rMyParser, Program **pPPgm, ostream &rOut)
{
  int res2 = rMyParser.parse();

  cout << "parse result: " << res2 << endl;
  if(res2 != 0 )
    {
      cerr << "parse failure" << endl;
      return -1;
    };

  Program *pPgm = rMyParser.GetProgram();
  if( !pPgm )
    {
      cerr << "no program!" << endl;
      return 5;
    };

  cout << "mem blocks list size: " << pPgm->mbList.size() << endl;
  pPgm->PrintBlocks();

  *pPPgm = pPgm;

  Compiler comp(*pPgm);
  
  comp.DoIt();

  // execState.Reset();
  // pgm.Exec(execState);
  // std::list<MemOpp> newMemOpps( execState.moList );
  // if( !CompareMOLists(initMemOpps, newMemOpps) ) return -2;
  // cout << "pass 1 ok" << endl;

  // execState.Reset();
  // pgm.Exec(execState);
  // std::list<MemOpp> newMemOpps2( execState.moList );
  // if( !CompareMOLists(initMemOpps, newMemOpps2) ) return -3;
  // cout << "pass 2 ok" << endl;

  std::list<MemBlock>::iterator ePlace= pPgm->mbList.begin();
  while(ePlace != pPgm->mbList.end() )
    {
      (*ePlace).EmitCode(rOut);
      ePlace++;
    };

  unsigned offsetStart = 100;

  std::list<FunctionDef>::iterator hPlace= pPgm->fdList.begin();
  while(hPlace != pPgm->fdList.end() )
    {
      FunctionDef &rFD = *hPlace;
      /*
      rFD.function_params.offset = offsetStart;
      rFD.declared_vars.offset = rFD.function_params.offset + rFD.function_params.Size();
      rFD.compiler_temps.offset = rFD.declared_vars.offset + rFD.declared_vars.Size();
      rFD.compiler_special_vars.offset = rFD.compiler_temps.offset + rFD.compiler_temps.Size();

      offsetStart = rFD.compiler_special_vars.offset + rFD.compiler_special_vars.Size();

      rFD.function_params.EmitCode(rOut);
      rFD.declared_vars.EmitCode(rOut);
      rFD.compiler_temps.EmitCode(rOut);
      rFD.compiler_special_vars.EmitCode(rOut);
      */
      hPlace++;
    };


  rOut << ".SEGMENT program\n";
  rOut << ".ORG 0\n";
  rOut << comp.OG() << "LI %H 1000\n"; /* init the stack pointer -> compiled code does use the stack! */
  rOut << comp.OG()  << "CALL main\n";
  rOut << comp.OG() << "HLT\n";

  std::list<FunctionDef>::iterator cPlace= pPgm->fdList.begin();
  while(cPlace != pPgm->fdList.end() )
    {
      comp.EmitCode(*cPlace, rOut);
      cPlace++;
    };

  cout << "success" << endl;
  return 0;
};
