
 /*
  Copyright 2014
  John Culp, All Rights Reserved
 */


%{
#include <stdlib.h>
void yyerror(char *);
#include "TVCC_Parser_Objs.hh"
#include "Opps.hh"
#include "Program.hh"
class TVCC_Lexer;
#include "tvcc.tab.hh"
#include "TVCC_Lexer.hh"
%}
%option noyywrap
blank [ \t]

%{
  // Code run each time a pattern is matched.
  # define YY_USER_ACTION  loc.columns (yyleng);
%}


%%

%{
  // Code run each time yylex is called.
  loc.step ();
%}

{blank}+   loc.step ();
[\n]+      loc.lines (yyleng); loc.step ();

[0-9]+      { return TVCC::parser::make_INTEGER(atoi(yytext), loc); }

[0-9]+.[0-9]+  { return TVCC::parser::make_FLOAT_IMM(strtod(yytext,0), loc); }

mem_block { return TVCC::parser::make_MEM_BLOCK(loc); }

void { return TVCC::parser::make_VOID(loc); }

uint { return TVCC::parser::make_UINT(P_UINT, loc); }

sint { return TVCC::parser::make_SINT(P_SINT, loc); }

float { return TVCC::parser::make_FLOAT(P_FLOAT, loc); }

return { return TVCC::parser::make_RETURN(loc); }

while { return TVCC::parser::make_WHILE(loc); }

for { return TVCC::parser::make_FOR(loc); }

if { return TVCC::parser::make_IF(loc); }

else { return TVCC::parser::make_ELSE(loc); }

">" {
 return TVCC::parser::make_GT_OP(loc);
}

"<" {
 return TVCC::parser::make_LT_OP(loc);
}

">=" {
 return TVCC::parser::make_GTE_OP(loc);
}

"<=" {
 return TVCC::parser::make_LTE_OP(loc);
}

"==" {
 return TVCC::parser::make_EQ_OP(loc);
}

"!=" {
 return TVCC::parser::make_NEQ_OP(loc);
}

"&&" {
 return TVCC::parser::make_LOGIC_AND_OP(loc);
}

"||" {
 return TVCC::parser::make_LOGIC_OR_OP(loc);
}

"&" {
 return TVCC::parser::make_BIN_AND_OP(loc);
}

"|" {
 return TVCC::parser::make_BIN_OR_OP(loc);
}

"~" {
 return TVCC::parser::make_BIN_NOT_OP(loc);
}

"^" {
 return TVCC::parser::make_BIN_XOR_OP(loc);
}

"<<" {
 return TVCC::parser::make_LSHIFT_OP(loc);
}

">>" { return TVCC::parser::make_RSHIFT_OP(loc); }

"-" { return TVCC::parser::make_MINUS(loc); }

"+" { return TVCC::parser::make_PLUS(loc); }

"*"  { return TVCC::parser::make_STAR(loc); }

"/"  { return TVCC::parser::make_SLASH(loc); }

";"  { return TVCC::parser::make_SEMIC(loc); }

","  { return TVCC::parser::make_COMMA(loc); }

"(" { return TVCC::parser::make_L_PAREN(loc); }

")" { return TVCC::parser::make_R_PAREN(loc); }

"[" { return TVCC::parser::make_L_BRACKET(loc); }

"]" { return TVCC::parser::make_R_BRACKET(loc); }

"{" { return TVCC::parser::make_L_CBRACKET(loc); }

"}" { return TVCC::parser::make_R_CBRACKET(loc); }

"=" { return TVCC::parser::make_EQUAL(loc); }

%{
/* I really wanted to do this, but could not figure out how */
/* [-+;()=}*{/]|"["|"]"      { return TVCC::parser::symbol_type(*yytext, loc); } */
%}


[a-zA-Z_]+[a-zA-Z0-9_]* {
std::string myID(yytext);
std::cerr << "*******************************   returning identifier: " << yytext << std::endl;
return TVCC::parser::make_IDENTIFIER(myID, loc);
}

"/*"    {
 int c;
while((c = yyinput()) != 0)
 {
     if(c == '\n')
     {
        loc.lines(1);
       /* loc.step(); */
     }
     else if(c == '*')
     {
          if((c = yyinput()) == '/')
             break;
          else
             unput(c);
     }
 }
}

<<EOF>>    { return TVCC::parser::make_END(loc); }
.           yyerror("invalid character");
%%
