
 /*
  Copyright 2014
  John Culp, All Rights Reserved
 */

%skeleton "lalr1.cc"
%language "c++"
%define api.namespace {TVCC}
%parse-param {TVCC_Lexer* myLexer}
%lex-param {TVCC_Lexer* myLexer}
%debug

%precedence LOGIC_OR_OP
%precedence LOGIC_AND_OP
%precedence BIN_OR_OP
%precedence BIN_XOR_OP
%precedence BIN_AND_OP
%left EQ_OP NEQ_OP
%left GT_OP LT_OP GTE_OP LTE_OP
%left LSHIFT_OP RSHIFT_OP
%left PLUS MINUS
%left STAR SLASH

%locations
%{
  #include <iostream>

  #include "TVCC_Parser_Objs.hh"
  #include "Opps.hh"
  #include "IExpr.hh"
  #include "Program.hh"
  class TVCC_Lexer;
  #include "tvcc.tab.hh"
  #include "TVCC_Lexer.hh"

namespace TVCC
   { 
      TVCC::parser::symbol_type yylex(TVCC_Lexer* myLexer);
   };

  Program *pMyProgram = 0;

/*
-union {
Opp *pOpp;
int imm_int;
Program *pProgram;
char id_name[256];
}
-type <pOpp> expr var_assign_stmt shift_expr unary_expr array_assign_stmt
-type <pProgram> program;
-type <imm_int> INTEGER;
-type <id_name> IDENTIFIER;
*/
%}

%token            END    0     "end of file"
%token GT_OP LT_OP GTE_OP LTE_OP EQ_OP NEQ_OP LOGIC_AND_OP LOGIC_OR_OP
%token BIN_AND_OP BIN_OR_OP BIN_XOR_OP BIN_NOT_OP
%token LSHIFT_OP RSHIFT_OP
%token MEM_BLOCK VOID WHILE FOR IF ELSE
%token RETURN

%token <std::string> IDENTIFIER;
%token <float> FLOAT_IMM;
%token <int> INTEGER;
%token MINUS PLUS EQUAL STAR SLASH
%token SEMIC COMMA L_PAREN R_PAREN L_BRACKET R_BRACKET L_CBRACKET R_CBRACKET
%token <P_TYPE> UINT SINT FLOAT

%define api.token.constructor
%define api.value.type variant
%define parse.assert

%type <Opp*> var_assign_stmt array_assign_stmt call_stmt declare_stmt while_stmt for_stmt if_stmt return_stmt
%type <Opp*> all_stmts
%type <Program*> program;
%type <ParamDecl> parameter_declaration
%type <ParamDeclList> parameter_declaration_list
%type <P_TYPE> type_specifier
%type <MemBlock> mem_block
%type <FunctionDef> function_def
%type <IE *> expr shift_expr unary_expr
%type <FullOppList> full_statement_list
%type <ParamDecl> def_arg
%type <ParamDeclList> def_arg_list
%type <ExprList> call_arg_list

%start program


%%
program:
       program mem_block { $$ = $1; ($$)->mbList.push_back($2); std::cout << "mem block to list" << std::endl; }
       | program function_def { $$ = $1; ($$)->fdList.push_back($2); std::cout << "function def to list" << std::endl; }
       | {pMyProgram = new Program();  $$ = pMyProgram; }
        ;

expr:
        INTEGER                                  { $$ = new IE_Imm($1, P_SINT); }
        | FLOAT_IMM                              { $$ = new IE_Imm($1); }
        | L_PAREN expr R_PAREN                   { $$ = $2->Clone(); }
        | shift_expr                             { $$ = $1->Clone(); }
        | unary_expr                             { $$ = $1->Clone(); }
        | IDENTIFIER L_BRACKET expr R_BRACKET    { $$ = new IE_CodeVarReadOffset($1, *$3); /* array read via pointer arithmatic */ } 
        | IDENTIFIER            { $$ = new IE_CodeVarRead($1); }
        | expr PLUS expr         { $$ = new IE_Add(*$1, *$3); $$->SetLoc(yyla.location); }
        | expr MINUS expr         { $$ = new IE_Sub(*$1, *$3); $$->SetLoc(yyla.location); }
        | expr STAR expr         { $$ = new IE_Mul(*$1, *$3); $$->SetLoc(yyla.location); }
        | expr SLASH expr         { $$ = new IE_Div(*$1, *$3); $$->SetLoc(yyla.location); }
        | expr GT_OP expr         { $$ = new IE_CMP(*$1, *$3, IE_CMP::GT); $$->SetLoc(yyla.location); }
        | expr LT_OP expr         { $$ = new IE_CMP(*$1, *$3, IE_CMP::LT);  $$->SetLoc(yyla.location); }
        | expr GTE_OP expr         { $$ = new IE_CMP(*$1, *$3, IE_CMP::GTE); $$->SetLoc(yyla.location); }
        | expr LTE_OP expr         { $$ = new IE_CMP(*$1, *$3, IE_CMP::LTE); $$->SetLoc(yyla.location); }
        | expr EQ_OP expr         { $$ = new IE_CMP(*$1, *$3, IE_CMP::EQ); $$->SetLoc(yyla.location);  }
        | expr NEQ_OP expr         { $$ = new IE_CMP(*$1, *$3, IE_CMP::NEQ); $$->SetLoc(yyla.location); }
        | expr LOGIC_AND_OP expr         { $$ = new IE_Logic(*$1, *$3, IE_Logic::LOGIC_AND); $$->SetLoc(yyla.location); }
        | expr LOGIC_OR_OP  expr        { $$ = new IE_Logic(*$1, *$3, IE_Logic::LOGIC_OR); $$->SetLoc(yyla.location);  }
        | expr BIN_AND_OP expr        { $$ = new IE_Bin_Opp(*$1, *$3, IE_Bin_Opp::BIN_AND); $$->SetLoc(yyla.location); }
        | expr BIN_OR_OP  expr       { $$ = new IE_Bin_Opp(*$1, *$3, IE_Bin_Opp::BIN_OR); $$->SetLoc(yyla.location); }
        | expr BIN_XOR_OP  expr       { $$ = new IE_Bin_Opp(*$1, *$3, IE_Bin_Opp::BIN_XOR); $$->SetLoc(yyla.location); }
        | IDENTIFIER L_PAREN call_arg_list R_PAREN { $$ = new IE_Call($1, $3);  $$->SetLoc(yyla.location); std::cout << "adding new function call IEXPR" << std::endl; };
        ;

shift_expr:
        expr LSHIFT_OP INTEGER         { $$ = new IE_Shift(*$1, $3, IE_Shift::LEFT ); $$->SetLoc(yyla.location); }
        | expr RSHIFT_OP INTEGER         { $$ = new IE_Shift(*$1, $3, IE_Shift::RIGHT ); $$->SetLoc(yyla.location); }
        ;

unary_expr :
        BIN_AND_OP IDENTIFIER         { $$ = new IE_CodeVarAddress($2); $$->SetLoc(yyla.location); } 
        | BIN_NOT_OP expr               { $$ = new IE_Bin_Not(*$2); $$->SetLoc(yyla.location); }
        | L_PAREN type_specifier R_PAREN expr { $$ = new IE_Cast($2, *$4); $$->SetLoc(yyla.location); }

declare_stmt:
        type_specifier IDENTIFIER SEMIC { $$ = new CodeVarDec($1, $2); std::cout << "found new declare stmt!!!" << std::endl; }
        | type_specifier IDENTIFIER EQUAL expr SEMIC { $$ = new CodeVarDec($1, $2, *($4) ); std::cout << "found new declare stmt!!!" << std::endl; }
        ;

var_assign_stmt:
        IDENTIFIER EQUAL expr SEMIC { $$ = new CodeVarWrite($1, *$3); std::cout << "found var_assign_stmt: " << $1 << std::endl; }
        ;

array_assign_stmt:
        IDENTIFIER L_BRACKET expr R_BRACKET EQUAL expr SEMIC { $$ = new CodeVarWriteOffset($1, *$6, *$3); std::cout << "found array_assign_stmt: " << $1 << std::endl; }
        ;

call_stmt:
        IDENTIFIER L_PAREN R_PAREN SEMIC { $$ = new OppCall($1); std::cout << "found function call 1" << std::endl; }
        | IDENTIFIER L_PAREN call_arg_list R_PAREN SEMIC { $$ = new OppCall($1, $3); std::cout << "found function call 2" << std::endl; };

while_stmt:
         WHILE L_PAREN expr R_PAREN L_CBRACKET full_statement_list R_CBRACKET SEMIC  { $$ = new OppWhile( *($3), $6); std::cout << "found while statement call" << std::endl; }
	 ;

for_stmt:
         FOR L_PAREN var_assign_stmt expr SEMIC var_assign_stmt R_PAREN L_CBRACKET full_statement_list R_CBRACKET SEMIC { $$ = new OppFor( *($3), *($4), *($6), $9); std::cout << "found for statement" << std::endl; }
	 ;

if_stmt:
         IF L_PAREN expr R_PAREN L_CBRACKET full_statement_list R_CBRACKET SEMIC { $$ = new OppIf( *($3), $6); }
         | IF L_PAREN expr R_PAREN L_CBRACKET full_statement_list R_CBRACKET ELSE  L_CBRACKET full_statement_list R_CBRACKET SEMIC { $$ = new OppIf( *($3), $6, $10); }
	 ;

return_stmt:
       RETURN SEMIC { $$ = new OppReturn(); }
       | RETURN expr SEMIC { $$ = new OppReturn( *($2) ); }

all_stmts :
       declare_stmt          { $$ = $1; }
       | var_assign_stmt     { $$ = $1; }
       | array_assign_stmt   { $$ = $1; }
       | call_stmt           { $$ = $1; }
       | while_stmt          { $$ = $1; }
       | for_stmt            { $$ = $1; }
       | if_stmt             { $$ = $1; }
       | return_stmt             { $$ = $1; }
       ;

full_statement_list:
        all_stmts { $$ = FullOppList(*$1); };
        | full_statement_list all_stmts { ($1).AddStatement(*$2); $$ = $1; };
	;

type_specifier:
        UINT { $$= $1;}
        | SINT { $$= $1;}
        | FLOAT { $$= $1;}
	;

mem_block :
       MEM_BLOCK INTEGER L_CBRACKET parameter_declaration_list R_CBRACKET SEMIC { $$ = MemBlock($2, $4); std::cout << "found mem block" << std::endl; };
       ;

function_def :
       VOID  IDENTIFIER L_PAREN VOID R_PAREN L_CBRACKET full_statement_list R_CBRACKET { $$ = FunctionDef($2, $7); };
       | VOID IDENTIFIER L_PAREN def_arg_list R_PAREN L_CBRACKET full_statement_list R_CBRACKET { $$ = FunctionDef($2, $4, $7); };
       | type_specifier IDENTIFIER L_PAREN VOID R_PAREN L_CBRACKET full_statement_list R_CBRACKET { $$ = FunctionDef($2, $7); };
       | type_specifier IDENTIFIER L_PAREN def_arg_list R_PAREN L_CBRACKET full_statement_list R_CBRACKET { $$ = FunctionDef($2, $4, $7); };
       ;

call_arg_list :
         call_arg_list COMMA expr { $$ = $1; $$.AddIE(*$3); std::cout << "appending to call arg list" << std::endl; };
        | expr { $$ = ExprList(*$1); std::cout << "creating call arg list" << std::endl; };

def_arg_list :
        def_arg_list COMMA def_arg { $$ = $1; $$.AddDecl($3); };
        | def_arg { $$ = ParamDeclList($1); };

def_arg :
        type_specifier IDENTIFIER { $$ = ParamDecl($1, $2); };

parameter_declaration_list:
        parameter_declaration  { $$ = ParamDeclList($1); };
        | parameter_declaration_list parameter_declaration { $$ = $1; $$.AddDecl($2); };
	;

parameter_declaration : 
       type_specifier IDENTIFIER SEMIC { $$ = ParamDecl($1, $2); };
       | type_specifier IDENTIFIER L_BRACKET INTEGER R_BRACKET SEMIC { $$ = ParamDecl($1, $2, $4); };
       ;

%%
void yyerror(char *s) {
  std::cerr << "wow got error: " << s << std::endl;
}
