/*
  Copyright 2010,2014
  John Culp, All Rights Reserved
*/

#ifndef SIMGUI_INCLUDED
#define SIMGUI_INCLUDED

#include <libGx/GxMainWin.hh>
#include <libGx/GxColumn.hh>
#include <libGx/GxButton.hh>
#include <libGx/GxRow.hh>
#include <libGx/GxLabel.hh>
#include <libGx/GxPulldownMenu.hh>
#include <libGx/GxEditWin.hh>
#include <libGx/GxTable.hh>
#include <libGx/GxPopupWin.hh>

#include <libImg/SimpleImg.hh>
#include <libImg/U32RGBAPixel.hh>
#include <libImg/XDrawingData.hh>

#include "Sim.hh"

enum DATA_FMT{DATA_HEX, DATA_UINT, DATA_SIGNED, DATA_FP};
//a bs utility function to keep implementation between WatchRow and RegRow identical
//should probably use inheritance instead.
void FormatValue(CPU::REG_TYPE value, DATA_FMT format, std::string &rStr);

class RegRow : public GxRow
{
public:
  RegRow(GxRealOwner *pOwner, const char *pName);
  virtual ~RegRow(void);

  void SetValue(unsigned long newValue);

  void SetFormat(DATA_FMT newFmt);

protected:
  void Fill(void);

  DATA_FMT fmt;
  unsigned long value;

  GxLabel regLabel;
  GxEditWin valueWin;
  GxPulldownMenu format;

  GxPulldownItem fmtHex;
  GxPulldownItem fmtUInt;
  GxPulldownItem fmtSigned;
  GxPulldownItem fmtFp;
};

class WatchRow : public GxRow
{
public:
  WatchRow(GxRealOwner *pOwner);
  virtual ~WatchRow(void);

  unsigned ParseAddr(void);
  void SetValue(unsigned tValue);

  void SetFormat(DATA_FMT newFmt);

  GxEditWin address;
  GxEditWin name;
  GxEditWin valueWin;
  GxPulldownMenu format;

protected:
  void Fill(void);

  DATA_FMT fmt;
  unsigned long value;

  GxPulldownItem fmtHex;
  GxPulldownItem fmtUInt;
  GxPulldownItem fmtSigned;
  GxPulldownItem fmtFp;
};

class ExecUntilDialog : public GxPopupWin
{
public:
  ExecUntilDialog(GxTopLevelWin *pTLWin);
  virtual ~ExecUntilDialog(void);

  GX_STATUS DoDialog(void);

  unsigned GetAddress(void);

protected:
  GX_STATUS eStat;
  void DoExit(GX_STATUS rStat);
	    
  GxColumn mainColumn;

  GxLabel addLabel;
  GxEditWin address;

  GxRow buttonRow;
  GxButton ok, cancel;
};

class DispWin : public GxCoreWin
{
public:
  DispWin(GxRealOwner *pOwner);
  virtual ~DispWin(void);

  SimpleImg<unsigned, U32RGBAPixel> img;

  void Draw(void);

  virtual void HandleEvent(const XEvent &rEvent);

protected:
  virtual void CreateXWindow(void);
  virtual void GetWindowData(XSetWindowAttributes &winAttributes, ULINT &valueMask);

  XDrawingData dwgData;
};

class SimGui : public GxMainWin
{
public:
  SimGui(GxDisplay *pOwner, Sim &rTSim);
  virtual ~SimGui(void);

protected:
  Sim &rSim;

  void DoLoad(void);
  void DoExec(void);
  void DoExecUntil(void);
  void DoStop(void);
  void DoStep(void);
  void DoSaveCfg(void);
  void DoLoadCfg(void);
  void DoBlankMem(void);
  void DoLoadMem(void);
  void DoBlankFB(void);
  void DoUpdateGui(void);
  void DoQuit(void);

  bool stopExec;
  void ExecCore(bool stop_at_addr, unsigned pc_address);
  void UpdateGui(void);

  GxRow mainRow;

  GxColumn bCol;

  GxButton loadFile;
  GxButton exec;
  GxButton execUntil;
  GxButton stop;
  GxButton singleStep;
  GxButton saveCfg;
  GxButton loadCfg;
  GxButton blankMem;
  GxButton loadMem;
  GxButton blankFB;
  GxButton updateGui;
  GxButton quit;

  GxColumn regCol;
  GxRow insRow;
  RegRow pcRow;
  GxEditWin disassembly;
  GxTable regTable; //two registers per row
  RegRow aRow;
  RegRow bRow;
  RegRow cRow;
  RegRow dRow;
  RegRow eRow;
  RegRow fRow;
  RegRow gRow;
  RegRow hRow;

  GxLabel wrLabel;
  WatchRow wr0;
  WatchRow wr1;
  WatchRow wr2;
  WatchRow wr3;
  WatchRow wr4;
  WatchRow wr5;
  WatchRow wr6;
  WatchRow wr7;
  WatchRow wr8;
  WatchRow wr9;
  
  WatchRow *wrs[10]; //a faster way of indexing

  DispWin dwin;
};

#endif //SIMGUI_INCLUDED
