/*
  Copyright 2009,2014
  John Culp, All Rights Reserved
*/

#ifndef CPU_INCLUDED
#define CPU_INCLUDED

#include <string>
#include <vector>

// ************* start hackish cut and paste from assembler. *********************
typedef unsigned long MN;
const MN MN_NOP = 0x00;
const MN MN_LUI = 0x01;
const MN MN_LLI = 0x02;
const MN MN_ADD = 0x03;
const MN MN_SUB = 0x04;
const MN MN_ADDL = 0x05;
const MN MN_AND = 0x06;
const MN MN_OR  = 0x07;
const MN MN_XOR = 0x08;
const MN MN_NOT = 0x09;
const MN MN_BSET = 0x0a;
const MN MN_BCLR = 0x0b;
const MN MN_RSL = 0x0c;
const MN MN_RSR = 0x0d;
const MN MN_MUL = 0x10;
const MN MN_CMP = 0x30;
const MN MN_SRI = 0x40;
const MN MN_SRR = 0x41;
const MN MN_LRI = 0x42;
const MN MN_LRR = 0x43;
const MN MN_SEQZ = 0x50;
const MN MN_SNEQZ = 0x51;
const MN MN_SBSET = 0x52;
const MN MN_SBCLR = 0x53;
const MN MN_JI = 0x60;
const MN MN_JR = 0x61;
const MN MN_HLT = 0x70;

const unsigned I_CMD_O = 25;
const unsigned I_CMD_MASK = 0xFE000000;
const unsigned I_SRCA_O = 22;
const unsigned I_SRCA_MASK = 0x01C00000;
const unsigned I_SRCB_O = 19;
const unsigned I_SRCB_MASK = 0x00380000;
const unsigned I_DEST_O = 16;
const unsigned I_DEST_MASK = 0x00070000;
// ************ end cut and paste *****************


template <class T>
class bc_vec : public std::vector<T>
{
public:
  virtual T& operator[](unsigned long id)
  {
    return std::vector<T>::at(id);
  }
};

class CPU
{
public:
  CPU(void);
  ~CPU(void);

  void LoadProg(const std::string &rFileName);
  void BlankData(void);
  void LoadData(const std::string &rFileName);

  void Reset(void); //sets PC= 0;
  bool SingleStep(void);
  void Execute(void); //resets execCount

  unsigned long ExecCount(void);
  void PrintState(void);

  //must be 4 bytes. checked via assert in construcor
  typedef unsigned REG_TYPE;
  typedef signed SIGN_REG_TYPE;

  //these were protected...
  REG_TYPE DataRead(REG_TYPE dwordAddr);
  void DataWrite(REG_TYPE dwordAddr, REG_TYPE value);
  
  //disassembles the instruction pointed to by the PC
  void Disassemble(std::string &rStr);
  void Dissassemble(REG_TYPE insn, unsigned &rCmd, unsigned &srcA, unsigned &srcB, unsigned &dest,
		    unsigned &imm) const;
protected:
  unsigned verbose; //0 off 1 every insn 2, every 2nd. etc.
  unsigned long execCount; //number of instructions executed

  const char* RegToStr(unsigned reg);

  REG_TYPE regA;
  REG_TYPE regB;
  REG_TYPE regC;
  REG_TYPE regD;
  REG_TYPE regE;
  REG_TYPE regF;
  REG_TYPE regG;
  REG_TYPE regH;

  unsigned PC;

  //these implement the dma control logic
  virtual REG_TYPE InvalidDataRead(REG_TYPE dwordAddr);
  virtual void InvalidDataWrite(REG_TYPE dwordAddr, REG_TYPE value);

  //virtual void DoDMA(REG_TYPE localAddr, REG_TYPE fbAddr, unsigned count, bool toFB);

  bc_vec<REG_TYPE> data_ram;
  bc_vec<REG_TYPE> program_rom;
  unsigned program_size; //high water mark of filled instrucions

  REG_TYPE GetRegValue(unsigned reg) const;
  void SetRegValue(unsigned reg, REG_TYPE value);
  long SignExtendImm(REG_TYPE imm) const;
  REG_TYPE GetImmBitMask(REG_TYPE imm) const; //converts 0-31 to a single-bit-set bitmask

  // ********* 
  void I_LUI(unsigned srcA, unsigned destReg, REG_TYPE data);
  void I_LLI(unsigned srcA, unsigned destReg, REG_TYPE data);
  void I_ADD(unsigned srcA, unsigned srcB, unsigned dest);
  void I_SUB(unsigned srcA, unsigned srcB, unsigned dest);
  void I_ADDL(unsigned srcA, unsigned dest, unsigned imm);
  void I_AND(unsigned srcA, unsigned srcB, unsigned dest);
  void I_OR(unsigned srcA, unsigned srcB, unsigned dest);
  void I_XOR(unsigned srcA, unsigned srcB, unsigned dest);
  void I_NOT(unsigned srcA, unsigned dest);
  void I_BSET(unsigned srcA, unsigned dest, unsigned imm);
  void I_BCLR(unsigned srcA, unsigned dest, unsigned imm);
  void I_RSL(unsigned src, unsigned dest, unsigned imm);
  void I_RSR(unsigned src, unsigned dest, unsigned imm);
  void I_MUL(unsigned srcA, unsigned srcB, unsigned dest);
  void I_CMP(unsigned srcA, unsigned srcB, unsigned dest);
  void I_SRI(unsigned srcA, REG_TYPE imm);
  void I_SRR(unsigned srcA, unsigned srcB, REG_TYPE imm);
  void I_LRI(unsigned dest, REG_TYPE imm);
  void I_LRR(unsigned srcA, unsigned dest, REG_TYPE imm);
};

#endif //CPU_INCLUDED
