/*
  file created 2014
  modified
  released to the Public Domain by the author
*/

#include <math.h>
#include <iostream>
#include <algorithm>
#include <ext/functional>
#include <assert.h>
#include <iomanip>
#include <arpa/inet.h>

#include <iostream>

#include "CB_EPP.hh"

using namespace std;
const unsigned NUM_REGISTERS = 7;
const unsigned STATUS_REGISTER = 6;
const unsigned COMMAND_BUS_ADDRESS_REGISTER = 4;
const unsigned COMMAND_BUS_CONTROL_REGISTER = 5;

CB_EPP::CB_EPP(EPPPort &rTPort) :
  CB_Base(),
  localDataRegisters(NUM_REGISTERS),
  remoteDataRegisters(NUM_REGISTERS),
  remoteDataValid(NUM_REGISTERS),
  rPort(rTPort)
{
  //generate(localDataRegisters.begin(), localDataRegisters.end(), identity<unsigned char>(0));
  //generate(remoteDataRegisters.begin(), remoteDataRegisters.end(), identity<unsigned char>(0) );
  //generate(remoteDataValid.begin(), remoteDataValid.end(), identity<bool>(false) );
  for(unsigned ii = 0; ii < NUM_REGISTERS; ii++)
    {
      localDataRegisters[ii] = 0;
      remoteDataRegisters[ii] = 0;
      remoteDataValid[ii] = false;
    };
}

CB_EPP::~CB_EPP(void)
{}

void CB_EPP::CB_Write(unsigned addr, unsigned value)
{
  Write32(value, 0);
  localDataRegisters[COMMAND_BUS_ADDRESS_REGISTER] = addr;
  FlushRegisters();

  rPort.WriteAddress(COMMAND_BUS_CONTROL_REGISTER);
  rPort.WriteData(0xF0);
}

unsigned CB_EPP::CB_Read(unsigned addr)
{
  localDataRegisters[COMMAND_BUS_ADDRESS_REGISTER] = addr;
  FlushRegisters();

  /* perform a command bus read */
  rPort.WriteAddress(COMMAND_BUS_CONTROL_REGISTER);
  rPort.WriteData(0x0F);

  /* read values from command bus registers*/

  rPort.WriteAddress(0);
  unsigned char byte3 = 0;
  rPort.ReadData(byte3);

  rPort.WriteAddress(1);
  unsigned char byte2 = 0;
  rPort.ReadData(byte2);

  rPort.WriteAddress(2);
  unsigned char byte1 = 0;
  rPort.ReadData(byte1);

  rPort.WriteAddress(3);
  unsigned char byte0 = 0;
  rPort.ReadData(byte0);

  DirtyRegisterCache(); //leave things sane for the next command

  unsigned value = ((unsigned long)byte3)<<24 | ((unsigned long)byte2)<<16 | ((unsigned long)byte1)<<8 | (unsigned long)byte0;
  return value;
}

void CB_EPP::DumpRegisters(void)
{
//   for(unsigned ii = 0; ii < 7; ii++)
//     {
//       rPort.WriteAddress(ii);

//       unsigned char value = 0;
//       rPort.ReadData(value);
//       cout << "reg ii: " << ii << " " << setw(2) << setfill('0') << hex << (int)value << dec << endl;
//     };
  unsigned char values[NUM_REGISTERS];
  for(unsigned ii = 0; ii < NUM_REGISTERS; ii++)
    {
      rPort.WriteAddress(ii);

      rPort.ReadData(values[ii]);
      cout << "reg ii: " << ii << " " << setw(2) << setfill('0') << hex << (int)values[ii] << dec << endl;
    };

  unsigned long address = ((unsigned long)values[0])<<16 | ((unsigned long)values[1])<<8 | ((unsigned long)values[2]);
  cout  << setw(8) << setfill('0') << hex << address << dec << " " << address << endl;
}

void CB_EPP::DumpLocalRegisterCache(void)
{
  for(unsigned ii = 0; ii < NUM_REGISTERS; ii++)
    {
     cout << "reg ii: " << ii << " " << setw(2) << setfill('0') << hex << (int)localDataRegisters[ii] << dec << endl;
    };
}

void CB_EPP::ClearRegisters(void)
{
  for(unsigned ii = 0; ii < NUM_REGISTERS-1; ii++) //minus one to not write to the status register
    {
      rPort.WriteAddress(ii);
      rPort.WriteData(0x00);
    };

  DirtyRegisterCache();
}

void CB_EPP::DirtyRegisterCache(void)
{
  for(unsigned ii = 0; ii < NUM_REGISTERS; ii++)
    remoteDataValid[ii] = false;
}

void CB_EPP::FlushRegisters(void)
{
  unsigned char tvcIntAddress = 0xFF; // an invalid value
  for(unsigned ii = 0; ii < NUM_REGISTERS; ii++)
    {
      if( !remoteDataValid[ii] || ( localDataRegisters[ii] != remoteDataRegisters[ii] ) )
	{
	  if( ii != COMMAND_BUS_CONTROL_REGISTER )
	    {
	      if( tvcIntAddress != ii )
		{
		  rPort.WriteAddress(ii);
		  tvcIntAddress = ii;
		};

	      rPort.WriteData( localDataRegisters[ii] );
	      remoteDataRegisters[ii] = localDataRegisters[ii];
	      remoteDataValid[ii] = true;
	      tvcIntAddress++;
	    };
	};
    };
}

void CB_EPP::Write32(unsigned long value, unsigned regOffset)
{
  localDataRegisters[regOffset + 0] = static_cast<unsigned char>( (value & 0xFF000000)>>24 );
  localDataRegisters[regOffset + 1] = static_cast<unsigned char>( (value & 0x00FF0000)>>16 );
  localDataRegisters[regOffset + 2] = static_cast<unsigned char>( (value & 0x0000FF00)>>8 );;
  localDataRegisters[regOffset + 3] = static_cast<unsigned char>( (value & 0x000000FF) );
}
