/*
  file created 2014
  modified
  released to the Public Domain by the author
*/

#include <iostream>

#include "CB_SimShim.hh"

using namespace std;

CB_SimShim::CB_SimShim(void) :
  CB_Base(), ShimXShaderCPU(),
  reg_01(0), reg_02(0),
  reg_11(0), reg_12(0),
  reg_21(0), reg_22(0),
  reg_e0(0), reg_e1(0)
{
  CreateRenderContext();
}

CB_SimShim::~CB_SimShim(void)
{}

void CB_SimShim::CB_Write(unsigned addr, unsigned value)
{
  //block set
  switch(addr)
    {
    case 0x01:
      reg_01 = value;
      return;
    case 0x02:
      reg_02 = value;
      return;
    case 0x03:
      DoBlockSet(value);
      return;
    default:
      break;
    };

  switch(addr)
    {
    case 0x11:
      reg_11 = value;
      return;
    case 0x12:
      reg_12 = value;
      return;
    case 0x13:
      DoBlockWrite();
      return;
    default:
      break;
    };

  switch(addr)
    {
    case 0x21:
      reg_21 = value;
      return;
    case 0x22:
      reg_22 = value;
      return;
    case 0x23:
      DoBlockRead();
      return;
    default:
      break;
    };

 switch(addr)
    {
    case 0xE0:
      reg_e0 = value;
      return;
    case 0xE1:
      reg_e1 = value;
      return;
    case 0xE2:
      program_rom[reg_e1] = reg_e0;
      return;
    case 0xE3:
      reg_e0 = program_rom[reg_e1];
      return;
    case 0xE4:
      data_ram[reg_e1] = reg_e0;
      return;
    case 0xE5:
      reg_e0 = data_ram[reg_e1];
      return;
    case 0xE6:
      //the unimplemented case actually does the right thing.
      cout << "forcing processor halt (unimplemented)" << endl;
      return;
    // case 0xE7:
    //   reg_e1 = value;
    //   return;
    case 0xE8:
      ResetAndExecute(); //doesn't actually return until done...
      return;
    default:
      break;
    };

 cout << "unknown register read addr: " << hex << addr << dec << endl;
}

unsigned CB_SimShim::CB_Read(unsigned addr)
{
  switch(addr)
    {
    case 0x22:
      return reg_22;
    case 0xE0:
      return reg_e0;
    case 0xE1:
      return reg_e1;
    case 0xE6:
      //single threaded program. if this can be called, we're halted;
      return 0x1;
    case 0xFF:
      return 0xF; //all FU's idle
    default:
      cout << "unknown register read addr: " << hex << addr << dec << endl;
      cout << "returning 0" << endl;
      return 0;
    }
}

void CB_SimShim::CB_Wait_Equ(unsigned addr, unsigned value)
{
  cout << "CB_SimShim::CB_Wait_Equ unimplemented?!" << endl;
}

void CB_SimShim::DoBlockRead(void)
{
  reg_22 = fb[ (reg_21>>2) ];
}

void CB_SimShim::DoBlockWrite(void)
{
  //cout << "CB_SimShim::DoBlockWrite reg_11: " << reg_11 << " reg_12: " << reg_12 << endl;

  fb[ (reg_11>>2) ] = reg_12;
}

void CB_SimShim::DoBlockSet(reg_t pix_value)
{
  unsigned addr_start = reg_01>>2;
  for(unsigned ii = 0; ii < (reg_02>>2); ii++)
    {
      fb[addr_start + ii] = reg_12;
    };
}
