/*
  file created 2014
  modified
  released to the Public Domain by the author
*/

#include <stdio.h>
#include <assert.h>
#include <string.h>

#include <iostream>
#include <iomanip>
#include <unistd.h>

#include "CB_USB.hh"

using namespace std;

CB_USB::CB_USB(void) :
  usb_handle(0)
{}

CB_USB::~CB_USB(void)
{}

void CB_USB::Init(void)
{
  int ret;

  assert( libusb_init(&lusb_context) == 0);
  libusb_set_debug(lusb_context, 20000);

  usb_handle = libusb_open_device_with_vid_pid(lusb_context,0x0547,0x2131);
  if(!usb_handle)
    {
      cout << "could not find device!!" << endl;
      assert(false);
    };

  int res = libusb_reset_device(usb_handle); 
  cout << "reset res: " << res << endl;

  assert( !libusb_set_configuration(usb_handle, 1) );
  assert( !libusb_claim_interface(usb_handle, 0) );
  assert( !libusb_set_interface_alt_setting(usb_handle, 0, 1) );

  cout << "usb purging read buffer" << endl;
  do{
    unsigned char in_buffer[1024];
    int xfer = 0;
    ret = libusb_bulk_transfer(usb_handle, 0x86, in_buffer, 1024, &xfer, 100);
    cout << "purge returned: " << ret << " xfer: " << xfer << endl;
  }while(ret >= 0);

  cout << "usb init success!" << endl;
}

void CB_USB::Test(void)
{
  unsigned value;

  value = CB_Read(0xFF);
  cout << "CB_USB::Test got value pre1: " << hex << value << dec << endl;

  value = CB_Read(0xFE);
  cout << "CB_USB::Test got value 1: " << hex << value << dec << endl;

  //value = 0x12345678;
  value = 0xDEADBEEF;
  CB_Write(0xFE, value);
  value = CB_Read(0xFE);

  cout << "CB_USB::Test got value 2: " << hex << value << dec << endl;
}


void CB_USB::CB_Write(unsigned addr, unsigned value)
{
  unsigned char out_buff[8];
  BuildWriteBuffer(CMD_WR, addr, value, out_buff);

  int xfer = 0;
  int ret = libusb_bulk_transfer(usb_handle, 0x02, out_buff, 8, &xfer, 100);

#if 0
  cout << "write got ret: " << ret << " xfer: " << xfer << endl;
#endif

  if( xfer != 8 || ret != 0)
    {
      cout << "write failure in CB_USB::CB_Write ret: " << ret << " xfer: " << xfer << endl;
      assert(false);
    };
}

unsigned CB_USB::CB_Read(unsigned addr)
{
  unsigned char out_buff[8];
  BuildWriteBuffer(CMD_RD, addr, 0, out_buff);
  
  int xfer = 0;
  int ret = libusb_bulk_transfer(usb_handle, 0x02, out_buff, 8, &xfer, 100);

#if 0
 cout << "cb read pre write got ret: " << ret << " xfer: " << xfer << endl;
#endif

  if( xfer != 8 || ret != 0)
    {
      cout << "pre value write failure in CB_USB::CB_Read ret: " << ret << endl;
      assert(false);
    };

  return Read_LL();
}

unsigned CB_USB::Read_LL(void)
{
  unsigned char in_buffer[512];
  for(unsigned ii = 0; ii < 6; ii++)
    in_buffer[ii] = 0;

  int xfer = 0;
  int ret = libusb_bulk_transfer(usb_handle, 0x86, in_buffer, 6, &xfer, 100);

#if 0
  cout << "CB_USB::Read_LL got ret: " << ret << " xfer: " << xfer << endl;
#endif

  if( ret != 0)
    {
      cout << "read error 1 in CB_USB::Read_LL ret: " << ret << endl;
      assert(false);
    };

  if( xfer != 6 )
    {
      cout << "read error 2 in CB_USB::Read_LL" << endl;
      cout << "expected read of 6 got: " << xfer << endl;
      assert(false);
    };

  char val_array[4];
  val_array[0] = in_buffer[4];
  val_array[1] = in_buffer[3];
  val_array[2] = in_buffer[2];
  val_array[3] = in_buffer[1];

  unsigned value;
  memcpy(&value, val_array, 4);

#if 0
  for(unsigned ii = 0; ii < 6; ii++)
    {
      cout << dec << ii << " " << hex
	   << setw(2) << setfill('0') << (0xFF&((unsigned)in_buffer[ii]))
	   << dec << endl;
    };
#endif

  return value;
}
