/*
  file created 2014
  based on files from 2006,2009
  modified 2014
  released to the Public Domain by the author
*/

#ifndef DRIVER_HL_INCLUDED
#define DRIVER_HL_INCLUDED

#include <vector>

#include "CB_Base.hh"
//#include "EngineBase.hh"
#include "EngineSWBase.hh"

class Driver_HL : public EngineSWBase
{
public:
  Driver_HL(CB_Base &rTCB);
  virtual ~Driver_HL(void);

  void PrintStatus(void);
  void TestRam(void);

  //fundamentally a hack. don't call. disables clearing
  //front and depth buffers, etc.
  void ConfigForRTLSim(bool tForSim = false);

  void LoadProgram(const std::string &rPrgm);
  void LoadDataRam(unsigned addr, unsigned value);
  void ClearHLTandReset(void);
  void SimWaitForHalt(void);
  bool Halted(bool printStatus = false);

  void Blank(unsigned char pixValue = 0);

  void DrawPixel(int xPos, int yPos, unsigned char pixValue);
  void DrawLine(int xStart, int yStart, int xEnd, int yEnd, unsigned char pixValue);
  void Read_FB(unsigned fbAddress, unsigned &rValue);
  void Write_FB(unsigned fbAddress, unsigned newValue, bool wait = true);
  void Block_Set(unsigned fbAddress, unsigned numBytes, unsigned newValue, bool wait = true);

  // **************** start EngineBase overloads *******************
  virtual void Setup(void);
  virtual void TriFace(TCoord rC1, TCoord rC2, TCoord rC3, TexC t1, TexC t2, TexC t3);
  virtual void ClearFB(void);
  virtual void ClearDB(void);
  virtual void UpLoadTexture(Texture &rTexture, unsigned textureID);
  virtual void SetCTexture(unsigned cTextureID);
  virtual void Flush(void);
//virtual void PushTForm(TForm addTForm);
//virtual void ClearTForm(void);
  // **************** end EngineBase overloads *********************

  virtual void PolyScanLine(unsigned sX, long startZ, unsigned eX, long endZ, unsigned yValue, TexC texStart, TexC texInc, EPIXEL pixValue);
  virtual void ClearFB(EPIXEL pixValue);
  virtual void ClearDB(unsigned depthVal);

protected:
  bool forSim;

  unsigned char EPixelToHWPixel(EPIXEL pixValue);
  unsigned HWPixelToBusWidth(unsigned char hwPixel);

  void WaitOnBW(void); //poll for completion of the block writer
  void WaitOnBR(void); //same for block reader
  void WaitOnBS(void); //same for block set
  void WaitOnPS(void); //same for poly scanline

  unsigned scanline_count;
  unsigned currentTextureID;

  CB_Base &rCB;
};

#endif //DRIVER_HL_INCLUDED
