/*
  file created 2006
  modified
  released to the Public Domain by the author
*/

#include <math.h>
#include <sys/io.h>
#include <unistd.h>
#include <assert.h>
#include <stdlib.h>

#include <iostream>
#include <iomanip>

#include "EPPPort.hh"

const int BASE_ADDR = 0x378; //lpt1
//const int BASE_ADDR = 0x278; //lpt2

const int DATAPORT = BASE_ADDR;
const int STATUSPORT = BASE_ADDR+1;
const int CONTROLPORT = BASE_ADDR +2;
const int ADDRESSPORT = BASE_ADDR + 3;
const int EPPDATAPORT = BASE_ADDR + 4;

using namespace std;

EPPPort::EPPPort(void) :
  scStat(true), abError(true)
{}

EPPPort::~EPPPort(void)
{}

void EPPPort::SetStatusCheck(bool tscStat)
{
  scStat = tscStat;
}

void EPPPort::SetAbortOnError(bool tabError)
{
  abError = tabError;
}

bool EPPPort::Initialize(void)
{
  int res = ioperm(DATAPORT, EPPDATAPORT - DATAPORT + 1, true);
  if( res == -1 )
    {
      cerr << " failed getting permission to bit-bang ports" << endl;
      return false;
    };

  outb(0x04, CONTROLPORT);

  return true;
}

bool EPPPort::CloseAll(void)
{
  int res = ioperm(DATAPORT, EPPDATAPORT - DATAPORT, false);
  if( res == -1 )
    {
      cerr << "error releasing ports" << endl;
      return false;
    };

  return true;
}

const unsigned NUM_DATA_REGISTERS = 12;

bool EPPPort::TestDataLinkQuality(void)
{
  cout << "starting EPPPort::TestDataLinkQuality" << endl;

  unsigned long bad_count = 0;
  for(unsigned jj = 0; jj < 65535; jj++)
    {
      //cout << jj << endl;
      unsigned char data_values[NUM_DATA_REGISTERS];
      for(unsigned ii = 0; ii < NUM_DATA_REGISTERS; ii++)
	{
	  data_values[ii] = (unsigned char)rint( (((double)random())/((double)RAND_MAX))* 255.00);
	};

      for(unsigned ii = 0; ii < NUM_DATA_REGISTERS; ii++)
	{
	  WriteAddress(ii);
	  WriteData(data_values[ii]);
	};
      
      unsigned char read_data[NUM_DATA_REGISTERS];
      for(unsigned ii = 0; ii < NUM_DATA_REGISTERS; ii++)
	{
	  WriteAddress(ii);
	  ReadData( read_data[ii] );
	};

      bool row_error = false;
      for(unsigned ii = 0; ii < NUM_DATA_REGISTERS; ii++)
	{
	  if( read_data[ii] != data_values[ii] )
	    {
	      bad_count++;
	      row_error = true;
	    };
	};

      if(row_error) //|| true)
	{
	  cout << "row error actual: ";
	  for(unsigned ii = 0; ii < NUM_DATA_REGISTERS; ii++)
	    cout << setfill('0') << setw(2) << hex << (int)data_values[ii] << "  ";
	  cout << endl;
	  cout << "row error read:   ";
	  for(unsigned ii = 0; ii < NUM_DATA_REGISTERS; ii++)
	    cout << setfill('0') << setw(2) << hex << (int)read_data[ii] << "  ";
	  cout << dec << endl;
	  cout << endl;
	};
    };

  cout << "\nEPPPort::TestDataLinkQuality done. Test stats: " << endl;
  cout << "bad count: " << bad_count << endl;
  cout << "percentage error: " << (double)bad_count/(double)(65535*NUM_DATA_REGISTERS) << '\n' << endl;

  if( bad_count )
    return false;
  else
    return true;
}

bool EPPPort::CheckAndClearTimeout(void)
{
  unsigned char status = inb(STATUSPORT);
    
  if( status & 0x01 )
    {
      status |= 0x01; //found a doc that says to clear error write a 1 to status bit ?!
      outb(status, STATUSPORT);
      return true;
    }else
      return false;
}

void EPPPort::WriteAddress(unsigned char address)
{
  outb(address, ADDRESSPORT);
  if(scStat)
    {
      bool timeout = CheckAndClearTimeout();
      if(abError && timeout)
	{
	  cout << "fatal timeout in EPPPort::WriteAddress" << endl;
	  assert(false);
	};
    };
}

void EPPPort::WriteData(unsigned char data)
{
  outb(data, EPPDATAPORT);
  if(scStat)
    {
      bool timeout = CheckAndClearTimeout();
      if(abError && timeout)
	{
	  cout << "fatal timeout in EPPPort::WriteData" << endl;
	  assert(false);
	};
    };
}

void EPPPort::ReadData(unsigned char &rData)
{
  rData = inb(EPPDATAPORT);
  if(scStat)
    {
      bool timeout = CheckAndClearTimeout();
      if(abError && timeout)
	{
	  cout << "fatal timeout in EPPPort::ReadData" << endl;
	  assert(false);
	};
    };
}

