/*
  file created 2014
  from files created & modified in 2006 and 2009
  modified 2014
  released to the Public Domain by the author
*/

#include <unistd.h>
#include <math.h>
#include <assert.h>
#include <iostream>

#include "EngineShaderDevel.hh"

using namespace std;

EngineShaderDevel::EngineShaderDevel(void)
{}

EngineShaderDevel::~EngineShaderDevel(void)
{}

void EngineShaderDevel::TriFace(TCoord rC1, TCoord rC2, TCoord rC3,
				TexC t1, TexC t2, TexC t3)
{
  rC1 = tForm*rC1;
  rC2 = tForm*rC2;
  rC3 = tForm*rC3;

  TriFace(rC1.x, rC1.y, rC1.z,
	  rC2.x, rC2.y, rC2.z,
	  rC3.x, rC3.y, rC3.z,
	  t1.x, t1.y, t2.x, t2.y, t3.x, t3.y);
}

void EngineShaderDevel::PolyScanLine(uint sX, sint startZ, uint eX, sint endZ, uint yValue,
				     float texStart_x, float textStart_y, float texInc_x, float texInc_y, uint pixValue)
{
  PolyScanLine(sX, startZ, eX,endZ, yValue, TexC(texStart_x, textStart_y), TexC(texInc_x, texInc_y), pixValue);
}

void EngineShaderDevel::PushTForm(TForm addTForm)
{
  tForm = addTForm*tForm;
}

void EngineShaderDevel::ClearTForm(void)
{
  tForm = CreateIdentity();
}

float j_sqrt(float in)
{
  float lb = 0;
  float hb = in;

  for(unsigned ii = 0; ii < 16; ii++)
    {
      float guess = 0.5*(lb+hb);
      float res = guess*guess;
      if( res > in){
	hb = guess;
      }else{
	lb = guess;
      };
    }

  return 0.5*(lb+hb);
}

void EngineShaderDevel::TriFace(float rC1_x, float rC1_y, float rC1_z,
				float rC2_x, float rC2_y, float rC2_z,
				float rC3_x, float rC3_y, float rC3_z,
				float t1_x, float t1_y, float t2_x, float t2_y, float t3_x, float t3_y)
{  
  float v1_x = rC2_x - rC1_x;
  float v1_y = rC2_y - rC1_y;
  float v1_z = rC2_z - rC1_z;

  float v2_x = rC3_x - rC1_x;
  float v2_y = rC3_y - rC1_y;
  float v2_z = rC3_z - rC1_z;

  float normal_x = v1_y*v2_z - v1_z*v2_y;
  float normal_y = v1_z*v2_x - v1_x*v2_z;
  float normal_z = v1_x*v2_y - v1_y*v2_x;

  float mag = j_sqrt(normal_x*normal_x + normal_y*normal_y + normal_z*normal_z);
  float invMag = 1.00/mag;

  normal_x = normal_x*invMag;
  normal_y = normal_y*invMag;
  normal_z = normal_z*invMag;

  rC1_z = rC1_z*40.00;
  rC2_z = rC2_z*40.00;
  rC3_z = rC3_z*40.00;
 
  float lightVec_x = 0.00;
  float lightVec_y = 0.00;
  float lightVec_z = 1.00;

  float intensity =  lightVec_x*normal_x + lightVec_y*normal_y + lightVec_z*normal_z;

  uint cVal = RI(intensity*255.00);
  if(cVal > 255)
    {
      cVal = 255;
    };

  /* EPIXEL color = cVal<<EPIXEL_RED_OFFSET | cVal<<EPIXEL_GREEN_OFFSET | cVal<<EPIXEL_BLUE_OFFSET; */
  //uint color = (cVal<<16) | (cVal<<0);
  EPIXEL color =  cVal<<EPIXEL_GREEN_OFFSET | cVal<<EPIXEL_BLUE_OFFSET;

  /* I'll just cheat and use the lightVec as a back face test. */
  if(intensity < 0.00 )
    {
      return;
    };

  sint x1 = RI(rC1_x);
  sint y1 = RI(rC1_y);
  sint z1 = RI(rC1_z);

  sint x2 = RI(rC2_x);
  sint y2 = RI(rC2_y);
  sint z2 = RI(rC2_z);

  sint x3 = RI(rC3_x);
  sint y3 = RI(rC3_y);
  sint z3 = RI(rC3_z);

  /*
  cout << x1 << " " << y1 << " " << z1 << endl;
  cout << x2 << " " << y2 << " " << z2 << endl;
  cout << x3 << " " << y3 << " " << z3 << endl;
  cout << endl;
  */

  /* cout << "before: " << y1 << " " << y2 << " " << y3 << endl; */

  sint temp;
  float tTmp_x;
  float tTmp_y;

  /* sort the values according to increasing y's */
  if( y1 >= y2 && y1 > y3) /* exchanging y1 and y3 */
    {
      temp = x3; x3 = x1; x1 = temp;
      temp = y3; y3 = y1; y1 = temp;
      temp = z3; z3 = z1; z1 = temp;
      tTmp_x = t3_x; t3_x = t1_x; t1_x = tTmp_x;
      tTmp_y = t3_y; t3_y = t1_y; t1_y = tTmp_y;

      if( y1 > y2) /* exchanging old y3 and y2 */
	{
	  temp = x2; x2 = x1; x1 = temp;
	  temp = y2; y2 = y1; y1 = temp;
	  temp = z2; z2 = z1; z1 = temp;
	  tTmp_x = t2_x; t2_x = t1_x; t1_x = tTmp_x;
	  tTmp_y = t2_y; t2_y = t1_y; t1_y = tTmp_y;
	};
    }else{
      if( y1 > y2)
	{
	  temp = x2; x2 = x1; x1 = temp;
	  temp = y2; y2 = y1; y1 = temp;
	  temp = z2; z2 = z1; z1 = temp;
	  tTmp_x = t2_x; t2_x = t1_x; t1_x = tTmp_x;
	  tTmp_y = t2_y; t2_y = t1_y; t1_y = tTmp_y;
	}else{
	  if( y2 > y3)
	    {
	      temp = x2; x2 = x3; x3 = temp;
	      temp = y2; y2 = y3; y3 = temp;
	      temp = z2; z2 = z3; z3 = temp;
	      tTmp_x = t2_x; t2_x = t3_x; t3_x = tTmp_x;
	      tTmp_y = t2_y; t2_y = t3_y; t3_y = tTmp_y;
	      if(y1 > y2)
		{
		  temp = x2; x2 = x1; x1 = temp;
		  temp = y2; y2 = y1; y1 = temp;
		  temp = z2; z2 = z1; z1 = temp;
		  tTmp_x = t2_x; t2_x = t1_x; t1_x = tTmp_x;
		  tTmp_y = t2_y; t2_y = t1_y; t1_y = tTmp_y;
		};
	    };
      };
  };

  /* cout << "after: " << y1 << " " << y2 << " " << y3 << endl; */

  if( (y1 == y2) && (y2 == y3) )
    {
      return;
    };

  /*
  if( y1 > y2 || y1 > y3) assert(false);
  if( y2 > y3) assert(false);
  */

  if( y1 == y2 )
    {
      if( x1 < x2 ){
	FlatBottomRast(x1,z1,t1_x,t1_y, x2,z2,t2_x,t2_y, y1, x3,y3,z3,t3_x,t3_y, color);
      }else{
	FlatBottomRast(x2,z2,t2_x,t2_y, x1,z1,t1_x,t1_y, y1, x3,y3,z3,t3_x,t3_y, color);
      };
      return;
    };

  if( y2 == y3 )
    {
      if( x2 < x3 ){
	FlatTopRast(x1,y1,z1,t1_x,t1_y, x2,z2,t2_x,t2_y, x3,z3,t3_x,t3_y, y2, color);
      }else{
	FlatTopRast(x1,y1,z1,t1_x,t1_y, x3,z3,t3_x,t3_y, x2,z2,t2_x,t2_y, y2, color);
      };
      return;
    };

  if( (y3-y1)*(x2-x1) - (y2 - y1)*(x3-x1) > 0)
    {
      /* cout << "right broken rast" << endl; */
      RightBrokenRast(x1,y1,z1,t1_x,t1_y, x2,y2,z2,t2_x,t2_y, x3,y3,z3,t3_x,t3_y, color);
      return;
    };

  /* cout << "left broken rast " << endl; */
  LeftBrokenRast(x1,y1,z1,t1_x,t1_y, x2,y2,z2,t2_x,t2_y, x3,y3,z3,t3_x,t3_y, color);
  return;
}

sint EngineShaderDevel::RI(float in)
{
  if(in < 0)
    return (sint)(in-0.5);
  else
    return (sint)(in+0.5);
}

void EngineShaderDevel::FlatBottomRast(sint blX, sint blZ, float blTC_x, float blTC_y,
				       sint brX, sint brZ, float brTC_x, float brTC_y,
				       sint bY, sint tX, sint tY, sint tZ, float tTC_x, float tTC_y, uint pixColor)
{
 
  float scale = 1.00/(float)(tY - bY);

  float leftTCInc_x = ( scale*(tTC_x - blTC_x) );
  float leftTCInc_y = ( scale*(tTC_y - blTC_y) );
  float rightTCInc_x = ( scale*(tTC_x - brTC_x) );
  float rightTCInc_y = ( scale*(tTC_y - brTC_y) );

  float leftTC_x = (blTC_x);
  float leftTC_y = (blTC_y);
  float rightTC_x = (brTC_x);
  float rightTC_y = (brTC_y);
  
  sint cY;
  sint startX;
  sint endX;
  sint startZ;
  sint endZ;

  for(cY = bY; cY < tY; cY = cY + 1)
    {
      startX = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tX - blX)) + blX );
      endX = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tX - brX)) + brX );

      startZ = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tZ - blZ)) + blZ );
      endZ = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tZ - brZ)) + brZ );

      scale = 1.00/(float)(endX - startX);
      float incTC_x = scale*(rightTC_x - leftTC_x);
      float incTC_y = scale*(rightTC_y - leftTC_y);

      PolyScanLine(startX, startZ, endX, endZ, cY, leftTC_x, leftTC_y, incTC_x, incTC_y, pixColor);

      leftTC_x = leftTC_x + leftTCInc_x;
      leftTC_y = leftTC_y + leftTCInc_y;
      rightTC_x = rightTC_x + rightTCInc_x;
      rightTC_y = rightTC_y + rightTCInc_y;
    };
}

void EngineShaderDevel::FlatTopRast(sint bX, sint bY, sint bZ, float bTC_x, float bTC_y,
				    sint tlX, sint tlZ, float tlTC_x, float tlTC_y,
				    sint trX, sint trZ, float trTC_x, float trTC_y, sint tY, uint pixColor)
{
  float scale = 1.00/(float)(tY - bY);
  float leftTCInc_x = ( scale*(tlTC_x - bTC_x) );
  float leftTCInc_y = ( scale*(tlTC_y - bTC_y) );
  float rightTCInc_x = ( scale*(trTC_x - bTC_x) );
  float rightTCInc_y = ( scale*(trTC_y - bTC_y) );

  float leftTC_x = bTC_x;
  float leftTC_y = bTC_y;
  float rightTC_x = bTC_x;
  float rightTC_y = bTC_y;

  sint cY;
  sint startX;
  sint endX;
  sint startZ;
  sint endZ;
  for(cY = bY; cY < tY; cY = cY+1)
    {
      startX = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tlX - bX)) + bX );
      endX = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(trX - bX)) + bX );

      startZ = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tlZ - bZ)) + bZ );
      endZ = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(trZ - bZ)) + bZ );

      scale = 1.00/(float)(endX - startX);
      float incTC_x = scale*(rightTC_x - leftTC_x);
      float incTC_y = scale*(rightTC_y - leftTC_y);

      PolyScanLine(startX, startZ, endX, endZ, cY, leftTC_x, leftTC_y, incTC_x, incTC_y, pixColor);

      leftTC_x = leftTC_x + leftTCInc_x;
      leftTC_y = leftTC_y + leftTCInc_y;
      rightTC_x = rightTC_x + rightTCInc_x;
      rightTC_y = rightTC_y + rightTCInc_y;
   };
}

void EngineShaderDevel::LeftBrokenRast(sint bX, sint bY, sint bZ, float bTC_x, float bTC_y,
				       sint lBrknX, sint lBrknY, sint lBrknZ, float lBrknTC_x, float lBrknTC_y,
				       sint tX, sint tY, sint tZ, float tTC_x, float tTC_y, uint pixColor)
{
  float scale = 1.00/(float)(lBrknY - bY);
  float leftTCInc_x = scale*(lBrknTC_x - bTC_x);
  float leftTCInc_y = scale*(lBrknTC_y - bTC_y);

  scale = 1.00/(float)(tY-bY);
  float rightTCInc_x = scale*(tTC_x - bTC_x);
  float rightTCInc_y = scale*(tTC_y - bTC_y);

  float leftTC_x = bTC_x;
  float leftTC_y = bTC_y;

  float rightTC_x = bTC_x;
  float rightTC_y = bTC_y;

  sint cY;
  sint startX;
  sint endX;
  sint startZ;
  sint endZ;
  float incTC_x;
  float incTC_y;

  for(cY = bY; cY < lBrknY; cY = cY + 1)
    {
      startX = RI( ((float)(cY - bY)/(float)(lBrknY - bY))*((float)(lBrknX - bX)) + bX );
      endX = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tX - bX)) + bX );

      startZ = RI( ((float)(cY - bY)/(float)(lBrknY - bY))*((float)(lBrknZ - bZ)) + bZ );
      endZ = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tZ - bZ)) + bZ );
 
      scale = 1.00/(float)(endX - startX);
      incTC_x = scale*(rightTC_x - leftTC_x);
      incTC_y = scale*(rightTC_y - leftTC_y);

      PolyScanLine(startX, startZ, endX, endZ, cY, leftTC_x, leftTC_y, incTC_x, incTC_y, pixColor);

      leftTC_x = leftTC_x + leftTCInc_x;
      leftTC_y = leftTC_y + leftTCInc_y;
      rightTC_x = rightTC_x + rightTCInc_x;
      rightTC_y = rightTC_y + rightTCInc_y;
    };

  leftTC_x = lBrknTC_x;
  leftTC_y = lBrknTC_y;
  scale = (1.00/(float)(tY - lBrknY));
  leftTCInc_x = scale*(tTC_x - lBrknTC_x);
  leftTCInc_y = scale*(tTC_y - lBrknTC_y);

  for(cY = lBrknY; cY < tY; cY = cY+1)
    {
      startX = RI( ((float)(cY - lBrknY)/(float)(tY - lBrknY))*((float)(tX - lBrknX)) + lBrknX );
      endX = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tX - bX)) + bX );

      startZ = RI( ((float)(cY - lBrknY)/(float)(tY - lBrknY))*((float)(tZ - lBrknZ)) + lBrknZ );
      endZ = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tZ - bZ)) + bZ );

      scale = 1.00/(float)(endX - startX);
      incTC_x = scale*(rightTC_x - leftTC_x);
      incTC_y = scale*(rightTC_y - leftTC_y);

      PolyScanLine(startX, startZ, endX, endZ, cY, leftTC_x, leftTC_y, incTC_x, incTC_y, pixColor);

      leftTC_x  = leftTC_x + leftTCInc_x;
      leftTC_y  = leftTC_y + leftTCInc_y;
      rightTC_x = rightTC_x + rightTCInc_x;
      rightTC_y = rightTC_y + rightTCInc_y;
    };
}

void EngineShaderDevel::RightBrokenRast(sint bX, sint bY, sint bZ, float bTC_x, float bTC_y,
					sint rBrknX, sint rBrknY, sint rBrknZ, float rBrknTC_x, float rBrknTC_y,
					sint tX, sint tY, sint tZ, float tTC_x, float tTC_y, uint pixColor)
{
  float leftTCInc_x  = (1.00/(float)(tY-bY) )*(tTC_x - bTC_x);
  float leftTCInc_y  = (1.00/(float)(tY-bY) )*(tTC_y - bTC_y);
  float rightTCInc_x = (1.00/(float)(rBrknY - bY) )*(rBrknTC_x - bTC_x);
  float rightTCInc_y = (1.00/(float)(rBrknY - bY) )*(rBrknTC_y - bTC_y);

  float leftTC_x  = bTC_x;
  float leftTC_y  = bTC_y;
  float rightTC_x = bTC_x;
  float rightTC_y = bTC_y;

  sint cY;
  sint startX;
  sint endX;
  sint startZ;
  sint endZ;
  float scale;
  float incTC_x;
  float incTC_y; 

  for(cY = bY; cY < rBrknY; cY = cY+1)
    {
      startX = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tX - bX)) + bX );
      endX = RI( ((float)(cY - bY)/(float)(rBrknY - bY))*((float)(rBrknX - bX)) + bX );

      startZ = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tZ - bZ)) + bZ );
      endZ = RI( ((float)(cY - bY)/(float)(rBrknY - bY))*((float)(rBrknZ - bZ)) + bZ );

      scale = 1.00/(float)(endX - startX);
      incTC_x = scale*(rightTC_x - leftTC_x);
      incTC_y = scale*(rightTC_y - leftTC_y);

      PolyScanLine(startX, startZ, endX, endZ, cY, leftTC_x, leftTC_y, incTC_x, incTC_y, pixColor);
      
      leftTC_x = leftTC_x + leftTCInc_x;
      leftTC_y = leftTC_y + leftTCInc_y;
      rightTC_x = rightTC_x + rightTCInc_x;
      rightTC_y = rightTC_y + rightTCInc_y;
    };

  rightTC_x = rBrknTC_x;
  rightTC_y = rBrknTC_y;
  scale = 1.00/(float)(tY - rBrknY);
  rightTCInc_x = scale*(tTC_x - rBrknTC_x);
  rightTCInc_y = scale*(tTC_y - rBrknTC_y);

  for(cY = rBrknY; cY < tY; cY=cY+1)
    {
      startX = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tX - bX)) + bX );
      endX = RI( ((float)(cY - rBrknY)/(float)(tY - rBrknY))*((float)(tX - rBrknX)) + rBrknX );

      startZ = RI( ((float)(cY - bY)/(float)(tY - bY))*((float)(tZ - bZ)) + bZ );
      endZ = RI( ((float)(cY - rBrknY)/(float)(tY - rBrknY))*((float)(tZ - rBrknZ)) + rBrknZ );

      scale = 1.00/(float)(endX - startX);
      incTC_x = scale*(rightTC_x - leftTC_x);
      incTC_y = scale*(rightTC_y - leftTC_y);

      PolyScanLine(startX, startZ, endX, endZ, cY, leftTC_x, leftTC_y, incTC_x, incTC_y,  pixColor);
       
      leftTC_x  = leftTC_x + leftTCInc_x;
      leftTC_y  = leftTC_y + leftTCInc_y;
      rightTC_x = rightTC_x + rightTCInc_x;
      rightTC_y = rightTC_y + rightTCInc_y;
   };
}
