/*
  file created 2009
  modified
  released to the Public Domain by the author
*/

#include <string.h>
#include <iostream>
#include <fstream>

using namespace std;

#include "ExtOffData.hh"

FullCoord::FullCoord(void)
{}

FullCoord::FullCoord(float tX, float tY, float tZ, float ttexX, float ttexY) :
  TCoord(tX, tY, tZ), texX(ttexX), texY(ttexY)
{}

FullCoord::FullCoord(const FullCoord &rhs) :
  TCoord(rhs), texX(rhs.texX), texY(rhs.texY)
{}

FullCoord::~FullCoord(void)
{}

const FullCoord& FullCoord::operator=(const FullCoord &rhs)
{
  static_cast<TCoord&>(*this) = rhs;
  texX = rhs.texX;
  texY = rhs.texY;

  return *this;
}


// **************************

ExtFaceData::ExtFaceData(void)
{}

ExtFaceData::ExtFaceData(unsigned tTexID, FullCoord tc1, FullCoord tc2, FullCoord tc3) :
  textureID(tTexID), c1(tc1), c2(tc2), c3(tc3)
{}

ExtFaceData::ExtFaceData(const ExtFaceData &rhs) :
  textureID(rhs.textureID), c1(rhs.c1), c2(rhs.c2), c3(rhs.c3)
{}

ExtFaceData::~ExtFaceData(void)
{}

const ExtFaceData & ExtFaceData::operator=(const ExtFaceData &rhs)
{
  textureID = rhs.textureID;
  c1 = rhs.c1;
  c2 = rhs.c2;
  c3 = rhs.c3;

  return *this;
}

// ********************

ExtOffData::ExtOffData(void)
{}

ExtOffData::~ExtOffData(void)
{
  while( !texList.empty() )
    {
      Texture *pTexture = texList.front();
      delete pTexture;
      pTexture = 0;
      texList.pop_front();
    };
}

void ExtOffData::ReadFullCoord(std::istream &rin, FullCoord &rCoord)
{
  float cX, cY, cZ, texX, texY;
  rin >> cX;
  rin >> cY;
  rin >> cZ;
  rin >> texX;
  rin >> texY;
  
  rCoord = FullCoord(cX, cY, cZ, texX, texY);
}

/* this implementation is absurdly fragile... */
bool ExtOffData::LoadFile(const std::string &rFileName)
{
  ifstream fin( rFileName.c_str() );
  if( !fin ) {
    cerr << "failed to open file: " << rFileName << endl;
    return false;
  };

  while(true)
    {
      char lBuffer[256];
      memset(lBuffer, 0, 256);
      fin.getline(lBuffer, 255);
      lBuffer[255] = '\0';

      if( !fin ) {
	cerr << "unexpected eof" << endl;
	return false;
      };

      //truncate buffer down to first whitespace
      string cmd(lBuffer);

      for(unsigned ii = 0; ii < cmd.size(); ii++)
	if( isspace( cmd[ii] ) ) {
	    cmd = cmd.substr(0, ii);
	    break;
	  };
      
      if(cmd == "texture") {
	unsigned textureID = 0;
	fin >> textureID;

	fin.getline(lBuffer, 255);
	lBuffer[255] = '\0';

	fin.getline(lBuffer, 255);
	lBuffer[255] = '\0';
	
	string fileName(lBuffer);
	Texture *pTexture = new Texture;
	if( !pTexture->FillTexture(fileName) )
	  {
	    cerr << "filling texture: " << fileName << "failed" << endl;
	    delete pTexture;
	    pTexture = 0;
	    return false;
	  }else
	  texList.push_back(pTexture);
      }else
	if( cmd == "face") {
	  unsigned textureID;
 	  fin >> textureID;

	  FullCoord c1, c2, c3;
	  ReadFullCoord(fin, c1);
	  ReadFullCoord(fin, c2);
	  ReadFullCoord(fin, c3);

	  fin.getline(lBuffer, 255);
	  lBuffer[255] = '\0';

	  faceList.push_back( ExtFaceData(textureID, c1, c2, c3) );
	}else
	  if( cmd == "eof" ) {
	    cout << "found sane eof" << endl;
	    break;
	  }else
	    {
	      cerr << "unknown command: " << cmd << endl;
	      return false;
	    };
    };

  return true;
}
