/*
  file created 2014
  modified
  released to the Public Domain by the author
*/

#ifndef SHIMXSHADERCPU_INCLUDED
#define SHIMXSHADERCPU_INCLUDED

#include <vector>
#include <string>
#include <fstream>

#include <libImg/SimpleImg.hh>
#include <libImg/U32RGBAPixel.hh>
#include <libImg/PXMUtil.hh>
#include <libImg/XImageWindow.hh>

#include "EngineSWBase.hh"
#include "CB_Base.hh"

#include "../cpu_sim/cpu.hh"

enum FB_DEPTH{FB_8BPP, FB_32BPP};

class ShimXShaderCPU : protected CPU
{
public:
  //the command bus object lets us create traces of command bus commands using
  //the CB_Logger for replaying in the HDL simulator
  ShimXShaderCPU(FB_DEPTH tFBDepth = FB_8BPP);
  virtual ~ShimXShaderCPU(void);

  virtual void CreateRenderContext(void);
  virtual void DeleteRenderContext(void);

  unsigned long long FullExecCount(void);
  void ResetFullExecCount(void);
protected:
  void Flush(void);
  
  virtual void ResetAndExecute(void);
  virtual void LocalWaitForHLT(void);

  FB_DEPTH fbDepth;
  std::string fname;

  REG_TYPE dma_cmd_0_reg;
  REG_TYPE dma_addr_0_reg;
  REG_TYPE dma_cmd_1_reg;
  REG_TYPE dma_addr_1_reg;
  REG_TYPE dma_cmd_2_reg;
  REG_TYPE dma_addr_2_reg;
  REG_TYPE dma_cmd_3_reg;
  REG_TYPE dma_addr_3_reg;

  unsigned long long lExecCount; //our running instruction count

  XImageWindow xImgWindow;

  unsigned currentTex; //zero is no texture is current
  Texture texArray[8];

  SimpleImg<unsigned, U32RGBAPixel> frameBuffer;
  void UpdateFB(void);
  bool visualFBChanged;

  bool refreshOnDMA; //only on dma writes of course
  unsigned refreshCount;

  std::ofstream fout;

  //cpu overloads. these implement the dma control logic
  virtual REG_TYPE InvalidDataRead(REG_TYPE dwordAddr);
  virtual void InvalidDataWrite(REG_TYPE dwordAddr, REG_TYPE value);
  virtual void DoDMA(REG_TYPE cmdReg, REG_TYPE fbAddr);

  std::vector<REG_TYPE> fb; //8MB buffer is size 2000000
};

#endif //SHIMXSHADERCPU_INCLUDED
