/*
  file created 2009
  modified
  released to the Public Domain by the author
*/

#include <math.h>

#include "TCoord.hh"

using namespace std;

TCoord::TCoord(void)
{}

TCoord::TCoord(float tX, float tY, float tZ) :
  x(tX), y(tY), z(tZ)
{}

TCoord::TCoord(const TCoord &rhs) :
  x(rhs.x), y(rhs.y), z(rhs.z)
{}

TCoord::~TCoord(void)
{}

const TCoord& TCoord::operator=(const TCoord &rhs)
{
  x = rhs.x;
  y = rhs.y;
  z = rhs.z;

  return *this;
}

void TCoord::Normalize(void)
{
  float mag = sqrt( pow(x, (float)2.0) + pow(y, (float)2.0) + pow(z, (float)2.0) );
  if( mag == 0.00 ) return; //hack. should be a small number

  float invMag = 1.00/mag;
  x *= invMag;
  y *= invMag;
  z *= invMag;
}

float TCoord::Mag(void) const
{
  return sqrt( pow(x, (float)2.00) + pow(y, (float)2.00) + pow(z, (float)2.00) );
}

const TCoord& operator*=(TCoord &lhs, const float scale)
{
  lhs.x *= scale;
  lhs.y *= scale;
  lhs.z *= scale;

  return lhs;
};

const TCoord& operator+=(TCoord &lhs, const TCoord &rhs)
{
  lhs.x += rhs.x;
  lhs.y += rhs.y;
  lhs.z += rhs.z;

  return lhs;
}

TCoord operator+(const TCoord &lhs, const TCoord &rhs)
{
  TCoord res;
  res.x = lhs.x + rhs.x;
  res.y = lhs.y + rhs.y;
  res.z = lhs.z + rhs.z;

  return res;
}

TCoord operator-(const TCoord &lhs, const TCoord &rhs)
{
  TCoord res;
  res.x = lhs.x - rhs.x;
  res.y = lhs.y - rhs.y;
  res.z = lhs.z - rhs.z;

  return res;
}

TCoord operator%(const TCoord &lhs, const TCoord &rhs)
{
  TCoord res;
  res.x = lhs.y*rhs.z - lhs.z*rhs.y;
  res.y = lhs.z*rhs.x - lhs.x*rhs.z;
  res.z = lhs.x*rhs.y - lhs.y*rhs.x;

  return res;
}

float operator*(const TCoord &lhs, const TCoord &rhs)
{
  return lhs.x*rhs.x + lhs.y*rhs.y + lhs.z*rhs.z;
}

ostream& operator<<(ostream &rout, TCoord &rCoord)
{
  rout << rCoord.x << " " << rCoord.y << " " << rCoord.z;

  return rout;
}

