/*
  Copyright 2014
  John Culp, All Rights Reserved
*/

#include <iostream>
#include <string.h>
#include <assert.h>
#include <math.h>

#include "cpu.hh"

using namespace std;

int AddTest(void);
int SubTest(void);
int MulTest(void);
int DivTest(void);
int CtofTest(void);
int CtoiTest(void);

int main(int argc, char *argv[])
{
  if(argc != 2)
    {
      cerr << "usage: fptest <add> <sub> <mul> <div> <ctoi> <ctof>" << endl;
      return -1;
    };

  string cmd(argv[1]);

  if(cmd == "add")
    {
      return AddTest();
    };

  if(cmd == "sub")
    {
      return SubTest();
    };

  if(cmd == "mul")
    {
      return MulTest();
    };

  if(cmd == "div")
    {
      return DivTest();
    };

  if(cmd == "ctoi")
    {
      return CtoiTest();
    };

  if(cmd == "ctof")
    {
      return CtofTest();
    };

  cerr << "unknown command: " << cmd << endl;
  return -1;
}

int AddTest(void)
{
  unsigned printInc = 0;

  signed argA = -1000000000;
  while(argA < 1000000000)
    {
      if( argA == 0 ) argA = 1; //prevent -inf
      signed ainc = (signed)4444.0*rint(log10(fabs(argA)));
      if( ainc < 1 ) ainc = 1;

      signed argB = -1000000000;
      while(argB < 1000000000)
	{
	  if( argB == 0 ) argB = 1; //prevent -inf
	  signed binc = (signed)4444.0*rint(log10(fabs(argB)));
	  if( binc < 1 ) binc = 1;

	  float fA = argA;
	  float fB = argB;
	  fA /= 1000.0;
	  fB /= 1000.0;
	  
	  printInc++;
	  if( !(printInc%10000) )
	    cout << "testing: " << fA << " " << fB << endl;
	  
	  CPU lcpu;
	  
	  lcpu.LoadProg("fptest-add.tbin");
	  lcpu.BlankData();
	  
	  CPU::REG_TYPE value;
	  
	  memcpy(&value, &fA, 4);
	  lcpu.DataWrite(0, value);
	  memcpy(&value, &fB, 4);
	  lcpu.DataWrite(1, value);
	  
	  lcpu.Execute();
	  
	  signed stack_start = 0xFF;
	  CPU::REG_TYPE ret_value = lcpu.DataRead(stack_start-7);
	  
	  float ret_float;
	  memcpy(&ret_float, &ret_value, 4);
	  
	  float des_value = fA+fB;
	  float error = ret_float - des_value;
	  
	  if( fabs(error) > fabs(0.01*des_value))
	    {
	      cout << "add failed result: " << ret_float << endl;
	      cout << "desired: " << des_value << endl;
	      return -3;
	    };
	  argB += binc;
	};
      argA += ainc;
    };

  cout << "add success!" << endl;
  return 0;
}

int SubTest(void)
{
  unsigned printInc = 0;

  signed argA = -1000000000;
  while(argA < 1000000000)
    {
      if( argA == 0 ) argA = 1; //prevent -inf
      signed ainc = (signed)4444.0*rint(log10(fabs(argA)));
      if( ainc < 1 ) ainc = 1;
      
      signed argB = -1000000000;
      while(argB < 1000000000)
	{
	  if( argB == 0 ) argB = 1; //prevent -inf
	  signed binc = (signed)4444.0*rint(log10(fabs(argB)));
	  if( binc < 1 ) binc = 1;

	  float fA = argA;
	  float fB = argB;
	  fA /= 1000.0;
	  fB /= 1000.0;
	  
	  printInc++;
	  if( !(printInc%10000) )
	    cout << "testing: " << fA << " " << fB << endl;
	  
	  CPU lcpu;
	  
	  lcpu.LoadProg("fptest-sub.tbin");
	  lcpu.BlankData();
	  
	  CPU::REG_TYPE value;

	  memcpy(&value, &fA, 4);
	  lcpu.DataWrite(0, value);
	  memcpy(&value, &fB, 4);
	  lcpu.DataWrite(1, value);
	  
	  lcpu.Execute();
	  
	  signed stack_start = 0xFF;
	  CPU::REG_TYPE ret_value = lcpu.DataRead(stack_start-7);
	  
	  float ret_float;
	  memcpy(&ret_float, &ret_value, 4);
	  
	  float des_value = fA-fB;
	  float error = ret_float - des_value;
	  
	  if( fabs(error) > fabs(0.01*des_value))
	    {
	      cout << "sub failed result: " << ret_float << endl;
	      cout << "desired: " << des_value << endl;
	      return -3;
	    };
	  argB += binc;
	};
      argA += ainc;
    };

  cout << "sub success!" << endl;
  return 0;
}

int MulTest(void)
{
  unsigned printInc = 0;

  signed argA = -1000000000;
  while(argA < 1000000000)
    {
      if( argA == 0 ) argA = 1; //prevent -inf
      signed ainc = (signed)4444.0*rint(log10(fabs(argA)));
      //cout << "ainc: " << ainc << endl;
      if( ainc < 1 ) ainc = 1;

      signed argB = -1000000000;
      while(argB < 1000000000)
	{
	  if( argB == 0 ) argB = 1; //prevent -inf
	  signed binc = (signed)4444.0*rint(log10(fabs(argB)));
	  if( binc < 1 ) binc = 1;

	  float fA = argA;
	  float fB = argB;
	  fA /= 1000.0;
	  fB /= 1000.0;

	  printInc++;
	  if( !(printInc%10000) )
	    cout << "testing: " << fA << " " << fB << endl;
	  
	  CPU lcpu;
	  
	  lcpu.LoadProg("fptest-mul.tbin");
	  lcpu.BlankData();
	  
	  CPU::REG_TYPE value;
	  
	  memcpy(&value, &fA, 4);
	  lcpu.DataWrite(0, value);
	  memcpy(&value, &fB, 4);
	  lcpu.DataWrite(1, value);
	  
	  lcpu.Execute();
	  
	  signed stack_start = 0xFF;
	  CPU::REG_TYPE ret_value = lcpu.DataRead(stack_start-7);
	  
	  float ret_float;
	  memcpy(&ret_float, &ret_value, 4);

	  float des_value = fA*fB;
	  float error = ret_float - des_value;
	  
	  if( fabs(error) > fabs(0.01*des_value) )
	    {
	      cout << "mul failed result: " << ret_float << endl;
	      cout << "desired: " << des_value << endl;
	      return -3;
	    };
	  argB += binc;
	};
      argA += ainc;
    };

  cout << "mul success!" << endl;
  return 0;
}

int DivTest(void)
{
  unsigned printInc = 0;

  signed argA = -1000000000;
  while(argA < 1000000000)
    {
      if( argA == 0 ) argA = 1; //prevent -inf
      signed ainc = (signed)4444.0*rint(log10(fabs(argA)));
      if( ainc < 1 ) ainc = 1;

      signed argB = -1000000000;
      while(argB < 1000000000)
	{
	  if( argB == 0 ) argB = 1; //prevent -inf
	  signed binc = (signed)4444.0*rint(log10(fabs(argB)));
	  if( binc < 1 ) binc = 1;

	  if(argB == 0)
	    {
	      argB += binc; //without this infinite loop
	      continue;
	    };

	  float fA = argA;
	  float fB = argB;
	  fA /= 1000.0;
	  fB /= 1000.0;

	  printInc++;
	  if( !(printInc%10000) )
	    cout << "testing: " << fA << " " << fB << endl;
	  
	  CPU lcpu;
	  
	  lcpu.LoadProg("fptest-div.tbin");
	  lcpu.BlankData();

	  CPU::REG_TYPE value1;
	  CPU::REG_TYPE value2;

	  memcpy(&value1, &fA, 4);
	  lcpu.DataWrite(0, value1);
	  memcpy(&value2, &fB, 4);
	  lcpu.DataWrite(1, value2);

	  //cout << "testing: " << hex << value1 << " " << hex << value2 << endl;

	  lcpu.Execute();

	  signed stack_start = 0xFF;
	  CPU::REG_TYPE ret_value = lcpu.DataRead(stack_start-7);

	  float ret_float;
	  memcpy(&ret_float, &ret_value, 4);
	  
	  float des_value = fA/fB;
	  float error = ret_float - des_value;
	  
	  if( fabs(error) > fabs(0.01*des_value) )
	    {
	      cout << "div failed result: " << ret_float << endl;
	      cout << "desired: " << des_value << endl;
	      return -3;
	    };
	  argB += binc;
	};
      argA += ainc;
    };

  cout << "div success!" << endl;
  return 0;
}

int CtoiTest(void)
{
  for(signed argA = -100000000; argA < 100000000; argA++)
    {
      float fA = argA;
      fA /= 1000.0;

      cout << "testing: " << fA << endl;
	
      CPU lcpu;
	
      lcpu.LoadProg("fptest-ctoi.tbin");
      lcpu.BlankData();

      CPU::REG_TYPE value;
      memcpy(&value, &fA, 4);
      lcpu.DataWrite(0, value);

      lcpu.Execute();

      signed stack_start = 0xFF;
      signed ret_value = lcpu.DataRead(stack_start-7);


      signed des_value = (signed)fA;

      if(des_value != ret_value)
	{
	  cout << "Ctoi failed result: " << ret_value << endl;
	  cout << "desired: " << des_value << endl;
	  return -3;
	};
      };

  cout << "ctoi success!" << endl;
  return 0;
}

int CtofTest(void)
{
  assert(sizeof(signed) == 4);

  for(signed argA = -10000000; argA < 10000000; argA++)
    {
      cout << "testing: " << argA << endl;
	
      CPU lcpu;
	
      lcpu.LoadProg("fptest-ctof.tbin");
      lcpu.BlankData();

      CPU::REG_TYPE value;
      memcpy(&value, &argA, 4);
      lcpu.DataWrite(0, value);

      lcpu.Execute();

      signed stack_start = 0xFF;
      CPU::REG_TYPE ret_value = lcpu.DataRead(stack_start-7);

      float ret_float;
      memcpy(&ret_float, &ret_value, 4);

      float error = ret_float - (float)argA;

      if( error > 0.25 )
	{
	  cout << "Ctof failed result: " << ret_float << endl;
	  cout << "desired: " << argA << endl;
	  return -3;
	};
      };

  cout << "ctof success!" << endl;
  return 0;
}
