-- Copyright 2014
-- John Culp, All Rights Reserved

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.shader_cpu_pack.all;

entity fetch is
  port(
    clock : in std_logic;
    reset : in std_logic;

    pc_to_fetch : out std_logic_vector(15 downto 0);
    ins_fetched : in  std_logic_vector(31 downto 0);

    new_addr : in std_logic_vector(15 downto 0);
    set_pc   : in std_logic;

    collision_in : in std_logic;

    hold_pc : in std_logic;

    instruction   : out std_logic_vector(31 downto 0);
    out_is_bubble : out std_logic;
    fetched_addr  : out std_logic_vector(15 downto 0)  --for debugging only
    );
end entity;

architecture behaviour of fetch is

  signal pc_reg  : unsigned(15 downto 0) := X"0000";
  signal prev_pc : unsigned(15 downto 0) := X"0000";
  signal prev_insn : std_logic_vector(31 downto 0) := (others => '0');
  
  signal reset_delay : std_logic;

  signal local_out_is_bubble : std_logic;
  signal local_hold_pc       : std_logic := '0';
begin

  
  
  --pc_to_fetch         <= std_logic_vector(prev_pc) when hold_pc = '1' else std_logic_vector(pc_reg);
  local_hold_pc <= '1' when (hold_pc = '1')  else '0'; --or (collision_in = '1')
  pc_to_fetch         <= std_logic_vector(prev_pc) when local_hold_pc = '1' else std_logic_vector(pc_reg);
  
  out_is_bubble       <= local_out_is_bubble;
  --everything takes a clock to update
  local_out_is_bubble <= '1' when (collision_in = '1') or (reset = '1') or (reset_delay = '1') or (set_pc = '1')  else '0'; --or (hold_pc = '1')
  --local_out_is_bubble <= '1' when (reset = '1') or (reset_delay = '1') or (set_pc = '1')  else '0';
  
  instruction <= ins_fetched;
  --instruction <= ins_fetched when local_out_is_bubble = '0' else prev_insn;

  fetched_addr <= std_logic_vector(prev_pc);
  --not right obviously
  --fetched_addr <= std_logic_vector(prev_pc) when local_out_is_bubble = '0' else (others=>'0');

  process(clock) is
  begin
    if rising_edge(clock) then
      if reset = '1' then
        pc_reg      <= X"0000";
        prev_pc     <= X"0000";
        reset_delay <= '1';
      else
        --process instruction fetch adjustments in the order of importance
        --the futher down the pipeline, the more important
        if set_pc = '1' then
          pc_reg      <= unsigned(new_addr);
          prev_pc     <= unsigned(new_addr);
          reset_delay <= '1';
        else
        
          if local_hold_pc /= '1' then
            reset_delay <= '0';

            if collision_in = '1' then
              pc_reg      <= prev_pc;
              reset_delay <= '1';
              --pc_reg <= pc_reg; --no change.
            else
              prev_pc <= pc_reg;
              pc_reg  <= pc_reg + 1;
              prev_insn <= ins_fetched;              
            end if;
          end if; --not hold
        end if;  --not jump
        
      end if;  --reset
    end if;  --clock
  end process;

end architecture;
