-- Copyright 2010,2014
-- John Culp, All Rights Reserved

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use std.textio.all;
use ieee.std_logic_textio.all;

use work.shader_cpu_pack.all;

entity load_store is
  port(
    clock : in std_logic;
    reset : in std_logic;

    skip_ins         : in std_logic;
    ins_opp          : in std_logic_vector(6 downto 0);
    ins_imm          : in std_logic_vector(15 downto 0);
    ins_imm_sign_ext : in std_logic_vector(31 downto 0);

    a_reg_value : in  std_logic_vector(31 downto 0);
    b_reg_value : in  std_logic_vector(31 downto 0);
    value_read  : out std_logic_vector(31 downto 0);
    busy        : out std_logic;

    d_ram_addr : out std_logic_vector(15 downto 0);
    d_ram_dout : out std_logic_vector(31 downto 0);
    d_ram_din  : in  std_logic_vector(31 downto 0);
    d_ram_we   : out std_logic;

    --at upper 8 bits of cpu address space
    ext_data_strobe  : out std_logic;   --perform a read or write
    ext_data_address : out std_logic_vector(7 downto 0);
    ext_data_out     : out std_logic_vector(31 downto 0);
    ext_data_in      : in  std_logic_vector(31 downto 0);
    ext_data_wr      : out std_logic;
    ext_data_busy    : in  std_logic
    );
end load_store;

architecture rtl of load_store is

  signal do_opp   : std_logic;
  signal do_write : std_logic;

  signal b_plus_imm : std_logic_vector(31 downto 0);

  signal ls_address_reg   : std_logic_vector(15 downto 0);
  signal ls_address       : std_logic_vector(15 downto 0);
  signal is_external_addr : std_logic;

  signal busy_in : std_logic;

  type LS_INT_STATE is (LS_IDLE, LS_PRE_STARTUP, LS_STARTUP, LS_EXT_WAIT, LS_DONE);
  signal ls_state : LS_INT_STATE;

begin

  do_opp <= '1' when skip_ins = '0' and ((ins_opp = INS_SRI) or (ins_opp = INS_LRI) or (ins_opp = INS_SRR) or (ins_opp = INS_LRR)) else '0';

  --second test makes sure we go busy immediatly on pre spinup of our state machine.
  --to stall processor waiting for external read.
  busy_in <= '1' when (do_opp = '1') or (ls_state /= LS_IDLE) else '0';
  busy    <= '0' when ls_state = LS_DONE                      else busy_in;

  do_write <= '1' when (ins_opp = INS_SRR) or (ins_opp = INS_SRI) else '0';

  b_plus_imm <= std_logic_vector(signed(b_reg_value) + signed(ins_imm_sign_ext));

  ls_address <=
    ins_imm                 when (ins_opp = INS_SRI) or (ins_opp = INS_LRI) else
    b_plus_imm(15 downto 0) when (ins_opp = INS_SRR) or (ins_opp = INS_LRR) else
    X"0000";  --hack?
  
  is_external_addr <= '1' when ls_address_reg (15 downto 8) = X"FF" else '0';

  value_read <= ext_data_in when is_external_addr = '1' and do_write = '0' and do_opp = '1' else
                d_ram_din when is_external_addr = '0' and do_write = '0' and do_opp = '1' else
                X"0000_0000";
  
  d_ram_addr <= ls_address_reg;
  --d_ram_we   <= '1' when (do_opp = '1') and (is_external_addr = '0') and (do_write = '1') and ls_state = LS_IDLE else '0';

  ext_data_address <= ls_address_reg(7 downto 0);
  --ext_data_wr      <= '1' when is_external_addr = '1' and do_write = '1' else '0';

  --unnecessary whens (keeps simulation cleaner)
  d_ram_dout   <= a_reg_value when do_write = '1' and is_external_addr = '0' else X"0000_0000";
  ext_data_out <= a_reg_value when do_write = '1' and is_external_addr = '1' else X"0000_0000";

  --yeah, this really needs a cleanup.
  process(clock) is
    variable next_state          : LS_INT_STATE;
    variable do_int_write_strobe : boolean;
    variable do_ext_write_strobe : boolean;
    variable do_ext_strobe       : boolean;
    variable out_line            : line;
  begin
    if rising_edge(clock) then
      next_state          := ls_state;
      do_ext_strobe       := false;
      do_int_write_strobe := false;
      do_ext_write_strobe := false;

      if reset = '1' then
        next_state := LS_IDLE;
      else                              --not reset
        case ls_state is
          when LS_IDLE =>
            ls_address_reg <= ls_address;
            if do_opp = '1' then
              next_state := LS_PRE_STARTUP;
            end if;
          when LS_PRE_STARTUP =>
            if is_external_addr = '1' then
              next_state    := LS_STARTUP;
              do_ext_strobe := true;
              if do_write = '1' then
                do_ext_write_strobe := true;
              end if;
            else  --  if is_external_addr = '0' then
              if do_write = '1' then
                do_int_write_strobe := true;
              end if;
              next_state := LS_DONE;
            end if;
          when LS_STARTUP =>
            if do_opp = '1' and do_write = '1' then
            --write(out_line, string'("ERS "));
            --hwrite(out_line, ls_address_reg);
            --write(out_line, string'(" "));
            --hwrite(out_line, a_reg_value);  --kinda
            --writeline(output, out_line);
            end if;
            if do_opp = '1' and do_write = '0' then
            --write(out_line, string'("ERL "));
            --hwrite(out_line, ls_address_reg);
            --writeline(output, out_line);
            end if;
            next_state := LS_EXT_WAIT;
          when LS_EXT_WAIT =>
            if ext_data_busy /= '1' then
              next_state := LS_DONE;
            end if;
          when LS_DONE =>
            if do_opp = '1' and do_write = '1' then
            --write(out_line, string'("RS "));
            --hwrite(out_line, ls_address_reg);
            --write(out_line, string'(" "));
            --hwrite(out_line, a_reg_value);  --kinda
            --writeline(output, out_line);
            end if;
            if do_opp = '1' and do_write = '0' then
            --write(out_line, string'("RL "));
            --hwrite(out_line, ls_address_reg);
            --writeline(output, out_line);
            end if;
            next_state := LS_IDLE;
          when others=>
            null;
        end case;
      end if;  --reset

      if do_int_write_strobe = true then
        d_ram_we <= '1';
      else
        d_ram_we <= '0';
      end if;

      if do_ext_write_strobe = true then
        ext_data_wr <= '1';
      else
        ext_data_wr <= '0';
      end if;

      if do_ext_strobe = true then
        ext_data_strobe <= '1';
      else
        ext_data_strobe <= '0';
      end if;

      ls_state <= next_state;
    end if;  --clock
  end process;

end architecture;

