-- Copyright 2014
-- John Culp, All Rights Reserved

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.shader_cpu_pack.all;

entity reg_bank is
  port(
    clock : in std_logic;
    --reset : in std_logic; --hack. no reset?

    port_a_addr : in std_logic_vector(2 downto 0);
    port_b_addr : in std_logic_vector(2 downto 0);

    --these ports are async read via the above addresses
    port_a : out std_logic_vector(31 downto 0);
    port_b : out std_logic_vector(31 downto 0);

    --port c is written on rising edge of clock if WE is high
    port_c_addr : in std_logic_vector(2 downto 0);
    port_c      : in std_logic_vector(31 downto 0);
    WE          : in std_logic
    );
end reg_bank;

architecture rtl of reg_bank is
  
  signal reg_000_a_out_port : std_logic_vector(31 downto 0);
  signal reg_000_b_out_port : std_logic_vector(31 downto 0);

  signal reg_001_a_out_port : std_logic_vector(31 downto 0);
  signal reg_001_b_out_port : std_logic_vector(31 downto 0);

  signal reg_010_a_out_port : std_logic_vector(31 downto 0);
  signal reg_010_b_out_port : std_logic_vector(31 downto 0);

  signal reg_011_a_out_port : std_logic_vector(31 downto 0);
  signal reg_011_b_out_port : std_logic_vector(31 downto 0);

  signal reg_100_a_out_port : std_logic_vector(31 downto 0);
  signal reg_100_b_out_port : std_logic_vector(31 downto 0);

  signal reg_101_a_out_port : std_logic_vector(31 downto 0);
  signal reg_101_b_out_port : std_logic_vector(31 downto 0);

  signal reg_110_a_out_port : std_logic_vector(31 downto 0);
  signal reg_110_b_out_port : std_logic_vector(31 downto 0);

  signal reg_111_a_out_port : std_logic_vector(31 downto 0);
  signal reg_111_b_out_port : std_logic_vector(31 downto 0);
  
begin

  reg_000 : std_reg
    generic map (REG_ADDR => "000")
    port map(clock    => clock, a_out_port => reg_000_a_out_port, b_out_port => reg_000_b_out_port,
             do_write => WE, reg_write_addr => port_c_addr, c_in_port => port_c);

  reg_001 : std_reg
    generic map (REG_ADDR => "001")
    port map(clock    => clock, a_out_port => reg_001_a_out_port, b_out_port => reg_001_b_out_port,
             do_write => WE, reg_write_addr => port_c_addr, c_in_port => port_c);

  reg_010 : std_reg
    generic map (REG_ADDR => "010")
    port map(clock    => clock, a_out_port => reg_010_a_out_port, b_out_port => reg_010_b_out_port,
             do_write => WE, reg_write_addr => port_c_addr, c_in_port => port_c);

  reg_011 : std_reg
    generic map (REG_ADDR => "011")
    port map(clock    => clock, a_out_port => reg_011_a_out_port, b_out_port => reg_011_b_out_port,
             do_write => WE, reg_write_addr => port_c_addr, c_in_port => port_c);

  reg_100 : std_reg
    generic map (REG_ADDR => "100")
    port map(clock    => clock, a_out_port => reg_100_a_out_port, b_out_port => reg_100_b_out_port,
             do_write => WE, reg_write_addr => port_c_addr, c_in_port => port_c);

  reg_101 : std_reg
    generic map (REG_ADDR => "101")
    port map(clock    => clock, a_out_port => reg_101_a_out_port, b_out_port => reg_101_b_out_port,
             do_write => WE, reg_write_addr => port_c_addr, c_in_port => port_c);

  reg_110 : std_reg
    generic map (REG_ADDR => "110")
    port map(clock    => clock, a_out_port => reg_110_a_out_port, b_out_port => reg_110_b_out_port,
             do_write => WE, reg_write_addr => port_c_addr, c_in_port => port_c);

  reg_111 : std_reg
    generic map (REG_ADDR => "111")
    port map(clock    => clock, a_out_port => reg_111_a_out_port, b_out_port => reg_111_b_out_port,
             do_write => WE, reg_write_addr => port_c_addr, c_in_port => port_c);


  port_a <=
    reg_000_a_out_port when port_a_addr = "000" else
    reg_001_a_out_port when port_a_addr = "001" else
    reg_010_a_out_port when port_a_addr = "010" else
    reg_011_a_out_port when port_a_addr = "011" else
    reg_100_a_out_port when port_a_addr = "100" else
    reg_101_a_out_port when port_a_addr = "101" else
    reg_110_a_out_port when port_a_addr = "110" else
    reg_111_a_out_port when port_a_addr = "111" else
    (others => '0');

  port_b <=
    reg_000_b_out_port when port_b_addr = "000" else
    reg_001_b_out_port when port_b_addr = "001" else
    reg_010_b_out_port when port_b_addr = "010" else
    reg_011_b_out_port when port_b_addr = "011" else
    reg_100_b_out_port when port_b_addr = "100" else
    reg_101_b_out_port when port_b_addr = "101" else
    reg_110_b_out_port when port_b_addr = "110" else
    reg_111_b_out_port when port_b_addr = "111" else
    (others => '0');

end rtl;

