-- Copyright 2014
-- John Culp, All Rights Reserved

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.shader_pack.all;
use std.textio.all;
use ieee.std_logic_textio.all;

use work.shader_pack.all;
use work.tb_pack.all;
use work.dma_sim_pack.all;

--ok we are going to have to:
--open a command data file whichi will:
--write to the shader's program ram to fill it
--(maybe re-read to test and validate that functionality)
--write to some data ram memory addresses. (fill in arguments)
--probably read the above too.
--write to a control register to release the cpu
--poll control register for execution to complete by HLT
--write new values into ram memory address for next polygon
--reset processor and wait for halt.
--the command data file will be single through
--will probably be created by capturing output of engine with a
--new c++ test generation backend. I expect the file to be big

entity simple_tb_2 is end simple_tb_2;

architecture behaviour of simple_tb_2 is
  signal clk     : std_logic := '0';
  signal sigterm : std_logic := '0';

  signal reset : std_logic := '0';

  --in and out names are wrt to the spi_slave.
  signal cmd_bus_data_in  : std_logic_vector(31 downto 0);
  signal cmd_bus_data_out : std_logic_vector(31 downto 0);
  signal cmd_bus_addr     : std_logic_vector(7 downto 0);
  signal cmd_bus_wr       : std_logic;
  signal cmd_bus_strobe   : std_logic;

  signal cb_addr_to_write : std_logic_vector(7 downto 0);
  signal cb_is_write      : std_logic;
  signal cb_to_write      : std_logic_vector(31 downto 0) := (others => '0');
  signal cb_was_read      : std_logic_vector(31 downto 0) := (others => '0');
  signal cb_cmd_strobe    : std_logic;
  signal cb_busy          : std_logic                     := '0';

  signal mu_bus_address : std_logic_vector(1 downto 0);
  signal mu_bus_data    : std_logic_vector(31 downto 0);
  signal mu_bus_write   : std_logic;
  signal mu_bus_strobe  : std_logic;
  signal mu_cpu_dma_irq : std_logic;
  signal mu_cpu_dma_sel : std_logic;

  type ADAPT_INT_STATE is (ADAPT_IDLE, ADAPT_WORKING, ADAPT_WAITING);
  signal adapt_state : ADAPT_INT_STATE;
  
  file stimulus : text open read_mode is "sim_commands.txt";
  --file stimulus : text open read_mode is "debug-blah.txt";

begin

  smb : shader_meta_block
    port map(clock           => clk, reset => reset,
             cmd_bus_data_in => cmd_bus_data_out, cmd_bus_data_out => cmd_bus_data_in, cmd_bus_addr => cmd_bus_addr,
             cmd_bus_wr      => cmd_bus_wr, cmd_bus_strobe => cmd_bus_strobe,
             mu_bus_address  => mu_bus_address, mu_bus_data => mu_bus_data, mu_bus_write => mu_bus_write, mu_bus_strobe => mu_bus_strobe,
             mu_cpu_dma_irq  => mu_cpu_dma_irq, mu_cpu_dma_sel => mu_cpu_dma_sel);

 mc : tb_mc
  port map(clk=>clk, reset=>reset,
           mu_bus_address=>mu_bus_address, mu_bus_data=>mu_bus_data, mu_bus_write=>mu_bus_write,
           mu_bus_strobe=>mu_bus_strobe, mu_cpu_dma_irq=>mu_cpu_dma_irq, mu_cpu_dma_sel=>mu_cpu_dma_sel);
  
  system_clock : process                --system clock
  begin
    wait for 50 ns;
    clkloop : loop
      wait for 5 ns;                    --100MHz (whups) clock (1/2 period) 
      clk <= not clk;
      if sigterm = '1' then
        exit;
      end if;
    end loop clkloop;
    wait for 1 us;
    wait;
  end process;

  system_reset : process
  begin
    wait for 20 ns;
    reset <= '1';
    wait for 99 ns;
    reset <= '0';
    wait;
  end process;

  cb_busy <= '1' when adapt_state /= ADAPT_IDLE else '0';
  
  --adapts the async test_process to the sync shader
  cb_process : process(clk)
    variable next_state : ADAPT_INT_STATE;
    variable do_cb_strobe : boolean;
  begin
    if rising_edge(clk) then
      next_state := adapt_state;
      do_cb_strobe := false;
      
      case adapt_state is
        when ADAPT_IDLE =>
          if cb_cmd_strobe = '1' then
            cmd_bus_data_out <= cb_to_write;
            cmd_bus_addr <= cb_addr_to_write;
            cmd_bus_wr <= cb_is_write;
            next_state := ADAPT_WORKING;
            do_cb_strobe := true;
          end if;
        when ADAPT_WORKING =>
          cb_was_read <= cmd_bus_data_in;
          next_state := ADAPT_WAITING;
        when ADAPT_WAITING =>
          if cb_cmd_strobe = '0' then
            next_state := ADAPT_IDLE;
          end if;
        when others=>
          null;
      end case;

      if do_cb_strobe = true then
        cmd_bus_strobe <= '1';
      else
        cmd_bus_strobe <= '0';
      end if;
      
      adapt_state <= next_state;
    end if;
  end process;


  test_process : process
    variable cmd_line           : line;
    variable out_line           : line;
    variable cmd                : std_logic_vector(31 downto 0);
    variable cmd_addr           : std_logic_vector(31 downto 0);
    variable cmd_data           : std_logic_vector(31 downto 0);
    variable local_cb_to_write : std_logic_vector(47 downto 0);
    variable local_cmd_is_write : std_logic;
    variable local_cmd_addr : std_logic_vector(7 downto 0);
    variable local_cmd_data : std_logic_vector(31 downto 0);
  begin
    cb_cmd_strobe <= '0';
    DMA_Init;

    wait for 5 us;
    loop
      readline(stimulus, cmd_line);
      --write(output, "reading cmd");
      hread(cmd_line, cmd);
      --write(output, "reading addr");
      hread(cmd_line, cmd_addr);
      --write(output, "reading data");
      hread(cmd_line, cmd_data);
      --write(out_line, string'("read_command: "));
      hwrite(out_line, cmd);
      write(out_line, string'(" "));
      hwrite(out_line, cmd_addr);
      write(out_line, string'(" "));
      hwrite(out_line, cmd_data);
      --writeline(output, out_line);
      --write(output, string'("before wait"));
      wait until cb_busy = '0' for 20 ns;
      --write(output, string'("after wait"));
      case cmd(3 downto 0) is
        when X"1" =>
          local_cmd_is_write := '1';
          local_cmd_addr := cmd_addr(7 downto 0);
          local_cmd_data := cmd_data;
        when X"2" =>
          local_cmd_is_write := '0';
          local_cmd_addr := cmd_addr(7 downto 0);
          local_cmd_data := cmd_data;
        when X"3" =>
          local_cmd_is_write := '0';
          local_cmd_addr := cmd_addr(7 downto 0);
          local_cmd_data := cmd_data;
        when others =>
          assert false report "invalid command from file" severity failure;
      end case;
      loop
        cb_addr_to_write <= local_cmd_addr;
        cb_is_write  <= local_cmd_is_write;
        cb_to_write  <= local_cmd_data;
        --write(output, string'("setting cmd strobe"));
        cb_cmd_strobe <= '1';
        wait for 20 ns;
        wait until cb_busy = '1' for 20 ns;
        cb_cmd_strobe <= '0';
        wait until cb_busy = '0';
        --write(output, string'("oh hey, cb_busy must be zero?"));
        assert cb_busy = '0' report "cb_busy not zero" severity failure;
        exit when cmd /= X"00000003" or ((cmd = X"00000003") and (cb_was_read(15 downto 0) = cmd_data(15 downto 0)));
        deallocate(out_line);
        write(out_line, string'("cmd3 result: "));
        hwrite(out_line, cb_was_read(31 downto 0));
        writeline(output, out_line);
        wait for 10 us;
      end loop;
      case cmd(3 downto 0) is
        when X"2" =>
          deallocate(out_line);
          write(out_line, string'("read_value: "));
          hwrite(out_line, cb_was_read(31 downto 0));
          writeline(output, out_line);
        when X"3" =>
          write(output, string'("cmd 3 completed!"));
        when others =>
          null;
      end case;
    end loop;

    DMA_Close;
  end process;

end behaviour;
