/*
  Copyright 2014
  John Culp, All Rights Reserved
*/

#include <iostream>
#include <unistd.h>
#include <vector>
#include <fstream>

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <libCb/CbCallback.hh>
#include <libImg/SimpleImg.hh>
#include <libImg/U32RGBAPixel.hh>
#include <libImg/PXMUtil.hh>
#include <libImg/XImageWindow.hh>

#include "../hdl-shader/dma_io_shared.h"

using namespace std;

const unsigned disp_w = 640;
const unsigned disp_h = 480;
int in_fd;
int out_fd;

void UpdateGui(XImageWindow &rX, std::vector<unsigned> &rFB)
{
  SimpleImg<unsigned, U32RGBAPixel> localImg(disp_w, disp_h);
  localImg.Initialize();

  for(unsigned cy = 0; cy < disp_h; cy++)
    for(unsigned cx = 0; cx < disp_w; cx++)
      {
#if 0
	//32bpp framebuffer
	localImg.Pixel(cx, cy) = rFB[cx + cy*disp_w];
#else
	//8bpp framebuffer
	unsigned pixel_byte = cx + cy*disp_w;
	unsigned quad_pix =  rFB[ (pixel_byte>>2) ];
	
	unsigned pix_byte_value;
	switch( pixel_byte&0x3 )
	  {
	  case 3:
	    pix_byte_value = (quad_pix&0x000000FF);
	    break;
	  case 2:
	    pix_byte_value = (quad_pix&0x0000FF00)>>8;
	    break;
	  case 1:
	    pix_byte_value = (quad_pix&0x00FF0000)>>16;
	    break;
	  default:
	    pix_byte_value = (quad_pix&0xFF000000)>>24;
	    break;
	  };

	unsigned red   = (pix_byte_value&0xE0)>>5;
	unsigned green = (pix_byte_value&0x1c)>>2;
	unsigned blue  = (pix_byte_value&0x3);
	
	red *= 255; red /= 7; //mul before divide
	green *= 255; green /= 7; //mul before divide
	blue *= 255; blue /= 3; //mul before divide
	
	U32RGBAPixel pixValue = (red<<24) | (green<<16) | (blue<<8);

	localImg.Pixel(cx, cy) = pixValue;
#endif
      };

  rX.UpdateWindow(localImg);
}

class MyImgWin : public XImageWindow
{
public:
  MyImgWin(unsigned tWidth, unsigned tHeight) :
    XImageWindow(tWidth, tHeight)
  {}

  virtual void TimerCB(void)
  {
    myCB();
  };

  CbVoidFO myCB;
};

std::vector<unsigned> fb;
MyImgWin *pXWin = 0; //so hacky.

void TimerCB(void);

int main(void)
{
  fb.resize(8000000); //hmmm. should be 2000000
  for(unsigned ii = 0; ii < fb.size(); ii++)
    fb[ii] = 0;

  MyImgWin xImgWindow(disp_w, disp_h);
  if( !xImgWindow.CreateAll() )
    {
      cerr << "failed to init x" << endl;
      return -1;
    };

  pXWin = &xImgWindow;

  xImgWindow.myCB.Assign( CbVoidPlain(&TimerCB) );
  UpdateGui(xImgWindow, fb);

  in_fd = out_fd = -1;

  cout << "starting server_in raw read open" << endl;
  in_fd = open("/home/john/tvc-all/hdl-shader/dma_server_in", O_RDONLY);
  cout << "raw read open returned" << endl;

  cout << "starting server out open" << endl;
  out_fd = open("/home/john/tvc-all/hdl-shader/dma_server_out", O_WRONLY);  
  cout << "raw write open returned" << endl;
 
  if(in_fd < 0)
    {
      cerr << "opening input in_fd failed" << endl;
      return -1;
    };

  if(out_fd < 0)
    {
      cout << "opening output out_fd failed" << endl;
      return -1;
    };

  cout << "in_fd: " << in_fd << " out_fd: " << out_fd << endl;

  xImgWindow.EventLoop();

  return 0;
};

void TimerCB(void)
{
  //cout << "TimerCB" << endl;

  unsigned writeCmdCount = 0;
  while(true)
    {
      fd_set read_set;
      FD_ZERO(&read_set);
      FD_SET(in_fd, &read_set);
      struct timeval timeout;
      timeout.tv_sec = 0;
      timeout.tv_usec = 0;

      int selres = select(in_fd+1, &read_set, 0, 0, &timeout);
      if(selres <= 0) return; //no read activity.

      unsigned command = Read_Int(in_fd);
      if(command == 0xF00FF00F)
	{
	  unsigned addr = Read_Int(in_fd);
	  unsigned count = Read_Int(in_fd);
	  cout << "doing a dma read from fb addr:" << addr << " count: " << count << endl;

	  for(unsigned ii = 0; ii < count; ii++)
	    {
	      unsigned value = fb[addr+ii];
	      cout << "write int: " << hex << value << endl;
	      Write_Int(out_fd, value);
	    }
	}else{
	if(command == 0xF00F0FF0)
	  {
	    writeCmdCount++;
	    unsigned addr = Read_Int(in_fd);
	    unsigned count = Read_Int(in_fd);
	    cout << "doing a dma write to fb addr: " << addr << " count: " << count << endl;

	    for(unsigned ii = 0; ii < count; ii++)
	      {
		unsigned value = Read_Int(in_fd);
		fb[addr+ii] = value;
		cout << "value: " << value << endl;
	      };
	  }else
	  {
	    cerr << "unknown command: " << hex << command << endl;
	    assert(false);
	  };
      };

      // if( !(writeCmdCount%10) ) //possibly skip
      // 	{
	  writeCmdCount = 0;
	  UpdateGui(*pXWin, fb);
	  //};
    };
}
