/*
  Copyright 2014
  John Culp, All Rights Reserved
*/

#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <assert.h>
#include <list>

using namespace std;

class MemValue
{
public:
  MemValue(void){};
  MemValue(unsigned a, unsigned v) :
    addr(a), value(v){};

  unsigned addr;
  unsigned value;
};

bool operator==(const MemValue &lhs, const MemValue &rhs)
{
  return (lhs.addr == rhs.addr) && (lhs.value == rhs.value);
} 


class ProcState
{
public:
  ProcState(void) :
    pc(0), a(0), b(0), c(0), d(0), e(0), f(0), g(0), h(0)
  {};

  friend ostream& operator<<(ProcState &rps, ostream &rout);
  friend bool operator!=(const ProcState &lhs, const ProcState &rhs);

  void Store(MemValue rNew)
  {
    list<MemValue>::iterator cPlace = ramList.begin();
    while(cPlace != ramList.end())
      {
	MemValue &rMV = *cPlace;
	if(rNew.addr == rMV.addr)
	  {
	    rMV.value = rNew.value;
	    cout << "updating address: " << rMV.addr << " with value: " << rMV.value << endl;
	    return;
	  }
	cPlace++;
      };

    cout << "writing address: " << rNew.addr << " with value: " << rNew.value << endl;
    ramList.push_back(rNew);
  };

  unsigned Load(unsigned addr)
  {
    list<MemValue>::iterator cPlace = ramList.begin();
    while(cPlace != ramList.end())
      {
	MemValue &rMV = *cPlace;
	if(addr == rMV.addr)
	  {
	    cout << "reading address: " << rMV.addr << " with value: " << rMV.value << endl;
	    return rMV.value;
	  }
	cPlace++;
      };

    cout << "ERROR UKNOWN ADDRESS READ: " << addr << " returning 0" << endl;
    return 0;
  };

  MemValue lastReadValue, lastWriteValue;

  unsigned pc, a, b, c, d, e, f, g, h;

  list<MemValue> ramList; //not sorted. yeah.
};

ostream& operator<<(ostream &rout, ProcState &rps)
{
  rout << hex << rps.pc << " " << rps.a << " " << rps.b << " " << rps.c << " " << rps.d << " " << rps.e << " " << rps.f << " " << rps.g << " " << rps.h << endl;
  return rout;
}

bool operator!=(const ProcState &lhs, const ProcState &rhs)
{
  if(lhs.pc != rhs.pc) return true;
  if(lhs.a != rhs.a)   return true;
  if(lhs.b != rhs.b)   return true;
  if(lhs.c != rhs.c)   return true;
  if(lhs.d != rhs.d)   return true;
  if(lhs.e != rhs.e)   return true;
  if(lhs.f != rhs.f)   return true;
  if(lhs.g != rhs.g)   return true;
  if(lhs.h != rhs.h)   return true;

  if( lhs.ramList != rhs.ramList )
    {
      cout << "ram lists fail" << endl;
      return true;
    }else
    cout << "memlists ok!" << endl;

  return false;
}

void AdvanceSim(ProcState &rSimState, ifstream &r_fin_sim, unsigned &line)
{

  while(true)
    {
      char linebuffer[256];
      r_fin_sim.getline(linebuffer, 255);
      linebuffer[255] = '\0';
      if(!r_fin_sim) assert(false);
      line++;

      if( string(linebuffer, 3) == "PC:")
	{
	  rSimState.pc = strtoul(linebuffer+4, 0, 16);
	  rSimState.a =  strtoul(linebuffer+15, 0, 16);
	  rSimState.b =  strtoul(linebuffer+26, 0, 16);
	  rSimState.c =  strtoul(linebuffer+37, 0, 16);
	  rSimState.d =  strtoul(linebuffer+48, 0, 16);
	  rSimState.e =  strtoul(linebuffer+59, 0, 16);
	  rSimState.f =  strtoul(linebuffer+70, 0, 16);
	  rSimState.g =  strtoul(linebuffer+81, 0, 16);
	  rSimState.h =  strtoul(linebuffer+92, 0, 16);
	  return; //we supposedly advanced pc.
	};

      if( string(linebuffer, 2) == "RS" )
	{
	  cout << "ram store" << endl;

	  char *pTail;
	  unsigned addr = strtoul(linebuffer+3, &pTail, 16);
	  char *pStart = pTail;
	  unsigned value = strtoul(pStart, &pTail, 16);

	  cout << "sim store: " << addr << " " << value << endl;
	  rSimState.Store( MemValue(addr, value) );

	  continue; //we have not yet advanced PC
	};

      if( string(linebuffer, 2) == "RL" )
	{
	  char *pTail;
	  unsigned addr = strtoul(linebuffer+3, &pTail, 16);
	  char *pStart = pTail;
	  unsigned value = strtoul(pStart, &pTail, 16);

	  cout << "sim load: "  << addr << " " << value << endl;
	  unsigned storedVal = rSimState.Load(addr);
	  if( storedVal != value )
	    {
	      cout << "sim load values don't match!!!!!" << endl;
	    };
	  continue; //we have not yet advanced PC
	};
    };
}

void AdvanceRtl(ProcState &rSimState, ifstream &r_fin_rtl, unsigned &line)
{
  while(true)
    {
      char linebuffer[256];
      r_fin_rtl.getline(linebuffer, 255);
      linebuffer[255] = '\0';
      if(!r_fin_rtl) assert(false);
      line++;

      if( string(linebuffer).substr(0,3) == "../" )
	{
	  cout << "ignoring line" << endl;
	  continue;
	};

      if( string(linebuffer).substr(0,3) == "SET" )
	{
	  char reg = linebuffer[4];
	  
	  unsigned value = strtoul(linebuffer+6, 0, 16);
	  cout << "set: " << reg << " " << value << endl;

	  switch(reg)
	    {
	    case 'A':
	      rSimState.a = value;
	      break;
	    case 'B':
	      rSimState.b = value;
	      break;
	    case 'C':
	      rSimState.c = value;
	      break;
	    case 'D':
	      rSimState.d = value;
	      break;
	    case 'E':
	      rSimState.e = value;
	      break;
	    case 'F':
	      rSimState.f = value;
	      break;
	    case 'G':
	      rSimState.g = value;
	      break;
	    case 'H':
	      rSimState.h = value;
	      break;
	    default:
	      cerr << "unknown reg" << endl;
	      assert(false);
	    };
	  continue;
	};

      if( string(linebuffer, 2) == "RS" )
	{
	  //cout << "ram store" << endl;
	  unsigned addr =  strtoul(linebuffer+3, 0, 16);
	  unsigned value = strtoul(linebuffer+8, 0, 16);
	  rSimState.Store( MemValue(addr, value) );
	  continue;
	};

      if( string(linebuffer, 2) == "RL" )
	{
	  //cout << "ram load" << endl;
	  unsigned addr =  strtoul(linebuffer+3, 0, 16);
	  rSimState.Load(addr);
	  continue;
	};

      if( string(linebuffer).substr(0,3) == "RET" ) //retired
	{
	  unsigned addr = strtoul(linebuffer+4, 0, 16);
	  cout << "ret: " << addr << endl;
	  rSimState.pc = addr;
	  return;
	};
    };

}

int main(void)
{
  cout << "hello world" << endl;

  ifstream fin_rtl("rtl-log.txt");
  ifstream fin_sim("sim-log.txt");

  if(!fin_rtl)
    {
      cerr << "failed to load rtl log" << endl;
      return -1;
    };

  if(!fin_sim)
    {
      cerr << "failed to load sim log" << endl;
      return -1;
    }

  ProcState sim, rtl;

  unsigned sim_line, rtl_line;
  sim_line = rtl_line = 0;

  while(true)
    {
      AdvanceSim(sim, fin_sim, sim_line);
      AdvanceRtl(rtl, fin_rtl, rtl_line);

	  cout << "sim: " << endl;
	  cout << sim;
	  cout << "rtl: " << endl;
	  cout << rtl;

      if(sim != rtl)
	{
	  cout << "mismatch!" << endl;
	  cout << "sim_line: " << dec << sim_line << " rtl_line: " << rtl_line << endl;
	  cout << "sim: " << endl;
	  cout << sim;
	  cout << "rtl: " << endl;
	  cout << rtl;

	  return 1;
	};
    };

  cout << "all good" << endl;
  return 0;
};
