--file created 2006
--modified 2009, 2014
--released to the Public Domain by the author

library IEEE;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity block_reader is
  port(
    --memory user access bus 
    mu_bus_address : in    std_logic_vector(1 downto 0);
    mu_bus_data    : inout std_logic_vector(31 downto 0);
    mu_bus_write   : in    std_logic;  --if high the MCU is going to write to a mu
    --if high the MCU expects the selected MU to present its data on the mu_bus_data or to
    --copy data from the mu_bus_data into its internal cache/registers.
    mu_bus_strobe  : in    std_logic;

    block_reader_irq    : out std_logic;
    block_reader_select : in  std_logic;

    --blah               
    tvc_clock : in std_logic;
    tvc_reset : in std_logic;

    --interface to the eppcu
    cmd_bus_data     : in  std_logic_vector(31 downto 0);
    cmd_bus_addr     : in  std_logic_vector(7 downto 0);
    cmd_bus_wr       : in  std_logic;
    cmd_bus_strobe   : in  std_logic;
    cmd_bus_data_out : out std_logic_vector(31 downto 0);

    idle : out std_logic
    );
end block_reader;

architecture Behavioral of block_reader is
  signal fb_address_reg : std_logic_vector(31 downto 0) := (others => '0');

  type PR_INT_STATE is (PR_IDLE, PR_WAITING, PR_MCU_DRIVEN);
  signal pr_state : PR_INT_STATE := PR_IDLE;

  signal pix_data_reg : std_logic_vector(31 downto 0);

begin
  idle             <= '1' when pr_state = PR_IDLE    else '0';
  block_reader_irq <= '1' when pr_state = PR_WAITING else '0';

  process(block_reader_select, mu_bus_write, mu_bus_address, fb_address_reg) is
    variable case_input : std_logic_vector(3 downto 0);
  begin
    case_input := block_reader_select & mu_bus_write & mu_bus_address;
    case case_input is
      when "1001" =>
        mu_bus_data <= X"00" & X"000001";
      when "1010" =>
        mu_bus_data <= fb_address_reg;
      when others =>
        mu_bus_data <= (others => 'Z');
    end case;
  end process;

  cmd_bus_data_out <= pix_data_reg when cmd_bus_addr = X"22" and cmd_bus_strobe = '1' else (others => 'Z');

  process (tvc_clock) is
    variable next_state : PR_INT_STATE;
  begin
    if tvc_clock = '1' and tvc_clock'event then
      next_state := pr_state;

      if tvc_reset = '1' then
        next_state := PR_IDLE;
      else                              --not reset
        case pr_state is
          when PR_IDLE =>
            if cmd_bus_strobe = '1' and cmd_bus_wr = '1' then
              case cmd_bus_addr is
                when X"21" =>
                  fb_address_reg <= cmd_bus_data(31 downto 0);
                when X"23" =>
                  next_state := PR_WAITING;
                when others =>
                  null;
              end case;
            end if;
          when PR_WAITING =>
            if block_reader_select = '1' then
              next_state := PR_MCU_DRIVEN;
            end if;
          when PR_MCU_DRIVEN =>
            if block_reader_select = '0' then
              next_state := PR_IDLE;
            end if;
            if mu_bus_strobe = '1' then
              pix_data_reg <= mu_bus_data;
            end if;
        end case;
      end if;  --reset

      pr_state <= next_state;
    end if;  --clock
  end process;
end Behavioral;
