-- Create Date:    03:03:41 09/16/2006
-- modified 2009, 2014
-- released to the Public Domain by the author

library IEEE;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

--HACK!!! this must break the transfer up into smaller chunks to prevent pix_pusher data starvation.
--right now this is handled in the driver, but this is quite tacky.

entity block_set is
  port(
    --memory user access bus
    mu_bus_address : in    std_logic_vector(1 downto 0);
    mu_bus_data    : inout std_logic_vector(31 downto 0);
    mu_bus_write   : in    std_logic;
    mu_bus_strobe  : in    std_logic;

    block_set_irq    : out std_logic;
    block_set_select : in  std_logic;

    tvc_clock : in std_logic;
    tvc_reset : in std_logic;

    --interface to the eppcu
    cmd_bus_data   : in std_logic_vector(31 downto 0);
    cmd_bus_addr   : in std_logic_vector(7 downto 0);
    cmd_bus_wr     : in std_logic;
    cmd_bus_strobe : in std_logic;

    idle : out std_logic
    );
end block_set;

architecture Behavioral of block_set is

  signal fb_address_reg : std_logic_vector(31 downto 2) := (others => '0');
  signal fb_num_reg     : std_logic_vector(23 downto 2) := (others => '0');
  signal fb_value_reg   : std_logic_vector(31 downto 0) := (others => '0');

  type BS_INT_STATE is (BS_IDLE, BS_WAITING, BS_MCU_DRIVEN);
  signal bs_state : BS_INT_STATE := BS_IDLE;

begin

  process(block_set_select, mu_bus_write, mu_bus_address, fb_address_reg, fb_num_reg, fb_value_reg) is
    variable case_input : std_logic_vector(3 downto 0);
  begin
    case_input := block_set_select & mu_bus_write & mu_bus_address;
    case case_input is
      when "1001" =>
        mu_bus_data <= X"80" & fb_num_reg & "00";
      when "1010" =>
        mu_bus_data <= fb_address_reg & "00";
      when "1000" =>
        mu_bus_data <= fb_value_reg;
      when others =>
        mu_bus_data <= (others => 'Z');
    end case;
  end process;

  idle          <= '1' when bs_state = BS_IDLE    else '0';
  block_set_irq <= '1' when bs_state = BS_WAITING else '0';

  process (tvc_clock) is
    variable next_state : BS_INT_STATE;
  begin
    if tvc_clock = '1' and tvc_clock'event then
      next_state := bs_state;

      if tvc_reset = '1' then
        next_state := BS_IDLE;
      else                              --not reset
        case bs_state is
          when BS_IDLE =>
            if cmd_bus_strobe = '1' and cmd_bus_wr = '1' then
              case cmd_bus_addr is
                when X"01" =>
                  fb_address_reg <= cmd_bus_data(31 downto 2);
                when X"02" =>
                  fb_num_reg <= cmd_bus_data(23 downto 2);
                when X"03" =>
                  fb_value_reg <= cmd_bus_data(31 downto 0);
                when X"04" =>
                  next_state := BS_WAITING;
                when others =>
                  null;
              end case;
            end if;
          when BS_WAITING =>
            if block_set_select = '1' then
              next_state := BS_MCU_DRIVEN;
            end if;
          when BS_MCU_DRIVEN =>
            if block_set_select = '0' then
              next_state := BS_IDLE;
            end if;
        end case;
      end if;  --reset

      bs_state <= next_state;
    end if;  --clock
  end process;

end Behavioral;

