--file created 2006
--modified 2009, 2014
--released to the Public Domain by the author

library IEEE;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity block_writer is
  port(
    --memory user access bus 
    mu_bus_address : in    std_logic_vector(1 downto 0);
    mu_bus_data    : inout std_logic_vector(31 downto 0);
    mu_bus_write   : in    std_logic;
    --not actually used here. we have only a single data value to present
    mu_bus_strobe  : in    std_logic;

    block_writer_irq    : out std_logic;
    block_writer_select : in  std_logic;

    tvc_clock : in std_logic;
    tvc_reset : in std_logic;

    --interface to the eppcu
    cmd_bus_data   : in std_logic_vector(31 downto 0);
    cmd_bus_addr   : in std_logic_vector(7 downto 0);
    cmd_bus_wr     : in  std_logic;
    cmd_bus_strobe : in std_logic;

    idle : out std_logic
    );
end block_writer;

architecture Behavioral of block_writer is
  signal fb_address_reg : std_logic_vector(31 downto 0) := (others => '0');
  signal fb_data_reg    : std_logic_vector(31 downto 0) := (others => '0');

  type BW_INT_STATE is (BW_IDLE, BW_MCU_WAITING, BW_MCU_DRIVEN, BW_COMPLETING);
  signal bw_state : BW_INT_STATE := BW_IDLE;

begin
  idle <= '1' when bw_state = BW_IDLE else '0';
  block_writer_irq <= '1' when bw_state = BW_MCU_WAITING else '0';
  
  process(block_writer_select, mu_bus_write, mu_bus_address, fb_address_reg, fb_data_reg) is
    variable case_input : std_logic_vector(3 downto 0);
  begin
    case_input := block_writer_select & mu_bus_write & mu_bus_address;
    case case_input is
      when "1001" =>
        mu_bus_data <= X"80" & X"000001";
      when "1010" =>
        mu_bus_data <= fb_address_reg;
      when "1000" =>
        mu_bus_data <= fb_data_reg;
      when others =>
        mu_bus_data <= (others => 'Z');
    end case;
  end process;

  process (tvc_clock)
    variable next_state : BW_INT_STATE;
  begin
    if tvc_clock = '1' and tvc_clock'event then
      next_state := bw_state;
      
      if tvc_reset = '1' then
        next_state := BW_IDLE;
      else                              --not reset
        case BW_state is
          when BW_IDLE =>
            if cmd_bus_strobe = '1' and cmd_bus_wr = '1' then
              case cmd_bus_addr is
                when X"11" =>
                  fb_address_reg <= cmd_bus_data(31 downto 0);
                when X"12" =>
                  fb_data_reg <= cmd_bus_data(31 downto 0);
                when X"13" =>
                  --fb_data_reg <= X"1234_5678";
                  next_state         := BW_MCU_WAITING;
                when others =>
                  null;
              end case;
            end if;
          when BW_MCU_WAITING =>
            if block_writer_select = '1' then
              next_state := BW_MCU_DRIVEN;
            end if;
          when BW_MCU_DRIVEN =>
            if block_writer_select = '0' then
              next_state := BW_COMPLETING;
            end if;
          when BW_COMPLETING =>
            next_state := BW_IDLE;
        end case;
      end if;  --reset

      bw_state <= NEXT_STATE;
    end if;  --clock
  end process;
end Behavioral;
