-- Copyright 2014
-- John Culp, All Rights Reserved

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.shader_cpu_pack.all;

entity compute is
  port(
    clock : in std_logic;
    reset : in std_logic;

    opcode       : in std_logic_vector(6 downto 0);
    a_reg        : in std_logic_vector(2 downto 0);
    b_reg        : in std_logic_vector(2 downto 0);
    imm          : in std_logic_vector(15 downto 0);
    imm_sign_ext : in std_logic_vector(31 downto 0);

    hold : in std_logic;

    a_reg_value : in std_logic_vector(31 downto 0);
    b_reg_value : in std_logic_vector(31 downto 0);

    a_reg_used     : out std_logic;
    b_reg_used     : out std_logic;
    alu_result     : out std_logic_vector(31 downto 0);
    skip_next_insn : out std_logic
    );
end entity;

architecture behaviour of compute is

  --if we assume the imm is a bit selection, this is the set bit
  --computed by the ALU for its usage, it outputs the value so we
  -- can also use it here.
  signal imm_bit_sel : std_logic_vector(31 downto 0);
begin

  my_alu : alu
    port map(opcode        => opcode, bus_a => a_reg_value, bus_b => b_reg_value, imm => imm, imm_sign_ext => imm_sign_ext,
             bit_shift_out => imm_bit_sel, result => alu_result);

  skip_next_insn <= '1' when opcode = INS_SEQZ and a_reg_value = X"0000_0000" else
                    '1' when opcode = INS_SNEQZ and a_reg_value /= X"0000_0000" else
                    '1' when opcode = INS_SBSET and (a_reg_value and imm_bit_sel) = imm_bit_sel else
                    '1' when opcode = INS_SBCLR and ((not a_reg_value) and imm_bit_sel) = imm_bit_sel else
                    '0';

  a_reg_used <=
    --INS_NOP
    '1' when opcode = INS_LUI else
    '1' when opcode = INS_LLI else
    '1' when opcode = INS_ADD else
    '1' when opcode = INS_SUB else
    '1' when opcode = INS_ADDL else
    '1' when opcode = INS_AND else
    '1' when opcode = INS_OR else
    '1' when opcode = INS_XOR else
    '1' when opcode = INS_NOT else
    '1' when opcode = INS_BSET else
    '1' when opcode = INS_BCLR else
    '1' when opcode = INS_RSL else
    '1' when opcode = INS_RSR else
    '1' when opcode = INS_MUL else
    '1' when opcode = INS_CMP else
    '1' when opcode = INS_SRI else
    '1' when opcode = INS_SRR else
    --INS_LRI
    --INS_LRR
    '1' when opcode = INS_SEQZ else
    '1' when opcode = INS_SNEQZ else
    '1' when opcode = INS_SBSET else
    '1' when opcode = INS_SBCLR else
    --INS_JI
    '1' when opcode = INS_JR else
    --INS_HLT
    '0';

  b_reg_used <=
    --INS_NOP
    --INS_LUI
    --INS_LLI
    '1' when opcode = INS_ADD else
    '1' when opcode = INS_SUB else
    --INS_ADDL
    '1' when opcode = INS_AND else
    '1' when opcode = INS_OR else
    '1' when opcode = INS_XOR else
    --INS_NOT
    --INS_BSET
    --INS_BCLR
    --INS_RSL
    --INS_RSR
    '1' when opcode = INS_MUL else
    '1' when opcode = INS_CMP else
    --INS_SRI
    '1' when opcode = INS_SRR else
    --INS_LRI
    '1' when opcode = INS_LRR else
    --INS_SEQZ
    --INS_SNEQZ
    --INS_SBSET
    --INS_SBCLR
    --JI
    --JR
    --HLT
    '0';
  
end architecture;
