-- Create Date:    03:49:58 08/11/2006
-- modified 2009, 2014
-- released to the Public Domain by the author

library IEEE;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

--this implements a 32 bit wide 16 location memory location for caching both depth and front buffer values from the framebuffer
--the 16 location depth is  selected based on the capabilities of the spartan II/III distributed ram

entity poly_cache is
  generic(WIDTH : positive);
  port(
    addr         : in  std_logic_vector(3 downto 0);
    data_out     : out std_logic_vector(WIDTH-1 downto 0);
    data_in      : in  std_logic_vector(WIDTH-1 downto 0);
    write_enable : in  std_logic;
    clk          : in  std_logic
    );
end poly_cache;

architecture Behavioral of poly_cache is

  --the xilinx distributed sram primitive
  component RAM16X1S
    port (
      D    : in  std_logic;
      WE   : in  std_logic;
      WCLK : in  std_logic;
      A0   : in  std_logic;
      A1   : in  std_logic;
      A2   : in  std_logic;
      A3   : in  std_logic;
      O    : out std_logic
      );
  end component;

  --signal int_output_data : std_logic_vector(31 downto 0);
  --signal output_data_reg : std_logic_vector(31 downto 0);
  
begin

  --data_out <= output_data_reg;
  
  g1 : for b in 0 to WIDTH-1 generate
    U1 : RAM16X1S
      port map(D => data_in(b), WE => write_enable, WCLK => clk, A0 => addr(0), A1 => addr(1), A2 => addr(2), A3 => addr(3), O => data_out(b));
  end generate;

  --make the reads syncronous
--      process(clk)
--      begin
--      if clk = '1' and clk'event then
--              output_data_reg <= int_output_data;
--      end if;
--      end process;

end Behavioral;

