-- Copyright 2014
-- John Culp, All Rights Reserved

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library UNISIM;
use UNISIM.VComponents.all;

use work.shader_cpu_pack.all;

entity ram_inst is
  port(
    clock : in std_logic;

    cpu_pc   : in  std_logic_vector(15 downto 0);
    cpu_inst : out std_logic_vector(31 downto 0);

    cb_int_strobe : in  std_logic;
    cb_int_we     : in  std_logic;
    cb_int_addr   : in  std_logic_vector(15 downto 0);
    cb_int_dout   : out std_logic_vector(31 downto 0);
    cb_int_din    : in  std_logic_vector(31 downto 0)
    );
end ram_inst;

architecture rtl of ram_inst is

  signal local_cb_we : std_logic;
  
  signal we_0, we_1, we_2, we_3 : std_logic;
  signal we_4, we_5, we_6, we_7 : std_logic;
  signal we_8, we_9, we_a, we_b : std_logic;
  signal we_c, we_d, we_e, we_f : std_logic;

  signal do_a_0, do_a_1, do_a_2, do_a_3 : std_logic_vector(31 downto 0);
  signal do_a_4, do_a_5, do_a_6, do_a_7 : std_logic_vector(31 downto 0);
  signal do_a_8, do_a_9, do_a_a, do_a_b : std_logic_vector(31 downto 0);
  signal do_a_c, do_a_d, do_a_e, do_a_f : std_logic_vector(31 downto 0);

  signal do_b_0, do_b_1, do_b_2, do_b_3 : std_logic_vector(31 downto 0);
  signal do_b_4, do_b_5, do_b_6, do_b_7 : std_logic_vector(31 downto 0);
  signal do_b_8, do_b_9, do_b_a, do_b_b : std_logic_vector(31 downto 0);
  signal do_b_c, do_b_d, do_b_e, do_b_f : std_logic_vector(31 downto 0);

  --preserve the register read at the rising edge to mux out the proper value
  --in next clock period when another address might be being presented
  signal addr_read_a_reg : std_logic_vector(15 downto 0);
  signal addr_read_b_reg : std_logic_vector(15 downto 0);

begin

  process(clock) is
  begin
    if rising_edge(clock) then
      addr_read_a_reg <= cpu_pc;
      addr_read_b_reg <= cb_int_addr;
    end if;
  end process;

  local_cb_we <= '1' when cb_int_we = '1' and cb_int_strobe = '1' else '0';

  we_0 <= local_cb_we when M_0( cb_int_addr ) else '0';
  we_1 <= local_cb_we when M_1( cb_int_addr ) else '0';
  we_2 <= local_cb_we when M_2( cb_int_addr ) else '0';
  we_3 <= local_cb_we when M_3( cb_int_addr) else '0';
  we_4 <= local_cb_we when M_4( cb_int_addr ) else '0';
  we_5 <= local_cb_we when M_5( cb_int_addr ) else '0';
  we_6 <= local_cb_we when M_6( cb_int_addr ) else '0';
  we_7 <= local_cb_we when M_7( cb_int_addr ) else '0';
  we_8 <= local_cb_we when M_8( cb_int_addr ) else '0';
  we_9 <= local_cb_we when M_9( cb_int_addr ) else '0';
  we_a <= local_cb_we when M_A( cb_int_addr ) else '0';
  we_b <= local_cb_we when M_B( cb_int_addr ) else '0';
  we_c <= local_cb_we when M_C( cb_int_addr ) else '0';
  we_d <= local_cb_we when M_D( cb_int_addr ) else '0';
  we_e <= local_cb_we when M_E( cb_int_addr ) else '0';
  we_f <= local_cb_we when M_F( cb_int_addr ) else '0';

  cpu_inst <=
    do_a_0 when M_0(addr_read_a_reg ) else
    do_a_1 when M_1(addr_read_a_reg ) else
    do_a_2 when M_2(addr_read_a_reg ) else
    do_a_3 when M_3(addr_read_a_reg ) else
    do_a_4 when M_4(addr_read_a_reg ) else
    do_a_5 when M_5(addr_read_a_reg ) else
    do_a_6 when M_6(addr_read_a_reg ) else
    do_a_7 when M_7(addr_read_a_reg ) else
    do_a_8 when M_8(addr_read_a_reg ) else
    do_a_9 when M_9(addr_read_a_reg ) else
    do_a_a when M_A(addr_read_a_reg ) else
    do_a_b when M_B(addr_read_a_reg ) else
    do_a_c when M_C(addr_read_a_reg ) else
    do_a_d when M_D(addr_read_a_reg ) else
    do_a_e when M_E(addr_read_a_reg ) else
    do_a_f when M_F(addr_read_a_reg ) else
    X"0000_0000";

  cb_int_dout <=
    do_b_0 when M_0(addr_read_b_reg ) else
    do_b_1 when M_1(addr_read_b_reg ) else
    do_b_2 when M_2(addr_read_b_reg ) else
    do_b_3 when M_3(addr_read_b_reg ) else
    do_b_4 when M_4(addr_read_b_reg ) else
    do_b_5 when M_5(addr_read_b_reg ) else
    do_b_6 when M_6(addr_read_b_reg ) else
    do_b_7 when M_7(addr_read_b_reg ) else
    do_b_8 when M_8(addr_read_b_reg ) else
    do_b_9 when M_9(addr_read_b_reg ) else
    do_b_a when M_A(addr_read_b_reg ) else
    do_b_b when M_B(addr_read_b_reg ) else
    do_b_c when M_C(addr_read_b_reg ) else
    do_b_d when M_D(addr_read_b_reg ) else
    do_b_e when M_E(addr_read_b_reg ) else
    do_b_f when M_F(addr_read_b_reg ) else
    X"0000_0000";

  in_ram_0 : RAMB16_S36_S36
    --generic map(INIT_00 => X"123456789abcdef0000000000000000000000000000000000123456789abcdef" )
    --generic map(INIT_00 =>   X"0000000000000efae0000000060a000002010000040101c8020000000400007b" )
    generic map(INIT_00 =>     X"02000000_04000000_c000000a_060a0000_02412000_04410456_02008000_04000123",
                INIT_01 =>     X"00000000_00000000_00000000_00000000_02412000_e0000000_02410000_04410000" )
    port map(ENA   => '1', ENB => '1', SSRA => '0', SSRB => '0', CLKA => clock, CLKB => clock,
             DIPA  => (others => '0'), DIPB => (others => '0'), DIA => (others => '0'), WEA => '0', DOA => do_a_0,
             ADDRA => cpu_pc(8 downto 0), DIB => cb_int_din, ADDRB => cb_int_addr(8 downto 0), WEB => we_0, DOB => do_b_0);

  in_ram_1 : RAMB16_S36_S36
    port map(ENA   => '1', ENB => '1', SSRA => '0', SSRB => '0', CLKA => clock, CLKB => clock,
             DIPA  => (others => '0'), DIPB => (others => '0'), DIA => (others => '0'), WEA => '0', DOA => do_a_1,
             ADDRA => cpu_pc(8 downto 0), DIB => cb_int_din, ADDRB => cb_int_addr(8 downto 0), WEB => we_1, DOB => do_b_1);

  in_ram_2 : RAMB16_S36_S36
    port map(ENA   => '1', ENB => '1', SSRA => '0', SSRB => '0', CLKA => clock, CLKB => clock,
             DIPA  => (others => '0'), DIPB => (others => '0'), DIA => (others => '0'), WEA => '0', DOA => do_a_2,
             ADDRA => cpu_pc(8 downto 0), DIB => cb_int_din, ADDRB => cb_int_addr(8 downto 0), WEB => we_2, DOB => do_b_2);

  in_ram_3 : RAMB16_S36_S36
    port map(ENA   => '1', ENB => '1', SSRA => '0', SSRB => '0', CLKA => clock, CLKB => clock,
             DIPA  => (others => '0'), DIPB => (others => '0'), DIA => (others => '0'), WEA => '0', DOA => do_a_3,
             ADDRA => cpu_pc(8 downto 0), DIB => cb_int_din, ADDRB => cb_int_addr(8 downto 0), WEB => we_3, DOB => do_b_3);

  in_ram_4 : RAMB16_S36_S36
    port map(ENA   => '1', ENB => '1', SSRA => '0', SSRB => '0', CLKA => clock, CLKB => clock,
             DIPA  => (others => '0'), DIPB => (others => '0'), DIA => (others => '0'), WEA => '0', DOA => do_a_4,
             ADDRA => cpu_pc(8 downto 0), DIB => cb_int_din, ADDRB => cb_int_addr(8 downto 0), WEB => we_4, DOB => do_b_4);

  in_ram_5 : RAMB16_S36_S36
    port map(ENA   => '1', ENB => '1', SSRA => '0', SSRB => '0', CLKA => clock, CLKB => clock,
             DIPA  => (others => '0'), DIPB => (others => '0'), DIA => (others => '0'), WEA => '0', DOA => do_a_5,
             ADDRA => cpu_pc(8 downto 0), DIB => cb_int_din, ADDRB => cb_int_addr(8 downto 0), WEB => we_5, DOB => do_b_5);

  in_ram_6 : RAMB16_S36_S36
    port map(ENA   => '1', ENB => '1', SSRA => '0', SSRB => '0', CLKA => clock, CLKB => clock,
             DIPA  => (others => '0'), DIPB => (others => '0'), DIA => (others => '0'), WEA => '0', DOA => do_a_6,
             ADDRA => cpu_pc(8 downto 0), DIB => cb_int_din, ADDRB => cb_int_addr(8 downto 0), WEB => we_6, DOB => do_b_6);

  in_ram_7 : RAMB16_S36_S36
    port map(ENA   => '1', ENB => '1', SSRA => '0', SSRB => '0', CLKA => clock, CLKB => clock,
             DIPA  => (others => '0'), DIPB => (others => '0'), DIA => (others => '0'), WEA => '0', DOA => do_a_7,
             ADDRA => cpu_pc(8 downto 0), DIB => cb_int_din, ADDRB => cb_int_addr(8 downto 0), WEB => we_7, DOB => do_b_7);

  in_ram_8 : RAMB16_S36_S36
    port map(ENA   => '1', ENB => '1', SSRA => '0', SSRB => '0', CLKA => clock, CLKB => clock,
             DIPA  => (others => '0'), DIPB => (others => '0'), DIA => (others => '0'), WEA => '0', DOA => do_a_8,
             ADDRA => cpu_pc(8 downto 0), DIB => cb_int_din, ADDRB => cb_int_addr(8 downto 0), WEB => we_8, DOB => do_b_8);

  in_ram_9 : RAMB16_S36_S36
    port map(ENA   => '1', ENB => '1', SSRA => '0', SSRB => '0', CLKA => clock, CLKB => clock,
             DIPA  => (others => '0'), DIPB => (others => '0'), DIA => (others => '0'), WEA => '0', DOA => do_a_9,
             ADDRA => cpu_pc(8 downto 0), DIB => cb_int_din, ADDRB => cb_int_addr(8 downto 0), WEB => we_9, DOB => do_b_9);

  in_ram_a : RAMB16_S36_S36
    port map(ENA   => '1', ENB => '1', SSRA => '0', SSRB => '0', CLKA => clock, CLKB => clock,
             DIPA  => (others => '0'), DIPB => (others => '0'), DIA => (others => '0'), WEA => '0', DOA => do_a_a,
             ADDRA => cpu_pc(8 downto 0), DIB => cb_int_din, ADDRB => cb_int_addr(8 downto 0), WEB => we_a, DOB => do_b_a);

  in_ram_b : RAMB16_S36_S36
    port map(ENA   => '1', ENB => '1', SSRA => '0', SSRB => '0', CLKA => clock, CLKB => clock,
             DIPA  => (others => '0'), DIPB => (others => '0'), DIA => (others => '0'), WEA => '0', DOA => do_a_b,
             ADDRA => cpu_pc(8 downto 0), DIB => cb_int_din, ADDRB => cb_int_addr(8 downto 0), WEB => we_b, DOB => do_b_b);

  in_ram_c : RAMB16_S36_S36
    port map(ENA   => '1', ENB => '1', SSRA => '0', SSRB => '0', CLKA => clock, CLKB => clock,
             DIPA  => (others => '0'), DIPB => (others => '0'), DIA => (others => '0'), WEA => '0', DOA => do_a_c,
             ADDRA => cpu_pc(8 downto 0), DIB => cb_int_din, ADDRB => cb_int_addr(8 downto 0), WEB => we_c, DOB => do_b_c);

  in_ram_d : RAMB16_S36_S36
    port map(ENA   => '1', ENB => '1', SSRA => '0', SSRB => '0', CLKA => clock, CLKB => clock,
             DIPA  => (others => '0'), DIPB => (others => '0'), DIA => (others => '0'), WEA => '0', DOA => do_a_d,
             ADDRA => cpu_pc(8 downto 0), DIB => cb_int_din, ADDRB => cb_int_addr(8 downto 0), WEB => we_d, DOB => do_b_d);

  in_ram_e : RAMB16_S36_S36
    port map(ENA   => '1', ENB => '1', SSRA => '0', SSRB => '0', CLKA => clock, CLKB => clock,
             DIPA  => (others => '0'), DIPB => (others => '0'), DIA => (others => '0'), WEA => '0', DOA => do_a_e,
             ADDRA => cpu_pc(8 downto 0), DIB => cb_int_din, ADDRB => cb_int_addr(8 downto 0), WEB => we_e, DOB => do_b_e);

  in_ram_f : RAMB16_S36_S36
    port map(ENA   => '1', ENB => '1', SSRA => '0', SSRB => '0', CLKA => clock, CLKB => clock,
             DIPA  => (others => '0'), DIPB => (others => '0'), DIA => (others => '0'), WEA => '0', DOA => do_a_f,
             ADDRA => cpu_pc(8 downto 0), DIB => cb_int_din, ADDRB => cb_int_addr(8 downto 0), WEB => we_f, DOB => do_b_f);

end architecture;
