-- Copyright 2014
-- John Culp, All Rights Reserved

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.shader_cpu_pack.all;

entity retire is
  port(
    clock : in std_logic;
    reset : in std_logic;

    insn      : in std_logic_vector(6 downto 0);
    insn_imm  : in std_logic_vector(15 downto 0);
    is_bubble : in std_logic;

    read_value_in : in std_logic_vector(31 downto 0);
    c_value_in    : in std_logic_vector(31 downto 0);

    --value to write may come from compute or load store. 
    --note: we may not write any value at all to a register
    reg_write_value : out std_logic_vector(31 downto 0);
    do_reg_write    : out std_logic;

    do_jump   : out std_logic;
    jump_addr : out std_logic_vector;

    do_halt : out std_logic
    );
end retire;

architecture behavioural of retire is
  signal maybe_reg_write : std_logic;
  signal skip_eval_in : std_logic := '1';  --hack. missing definition.
  signal local_insn   : std_logic_vector(6 downto 0);
begin

  local_insn <= (others => '0') when (is_bubble = '1') or (reset = '1') else insn;

  do_jump     <= '1' when (local_insn = INS_JI) or (local_insn = INS_JR) else '0';
  
  jump_addr <= insn_imm                when local_insn = INS_JI else
               c_value_in(15 downto 0) when local_insn = INS_JR else
               X"0000";  --hack. unnecessary but less distracting in sim.

  reg_write_value <= read_value_in when (local_insn = INS_LRI or local_insn = INS_LRR) else c_value_in;

  do_reg_write <= '0' when is_bubble = '1' else maybe_reg_write;
  
  maybe_reg_write <=
    '1' when local_insn = INS_LUI else
    '1' when local_insn = INS_LLI else
    '1' when local_insn = INS_ADD else
    '1' when local_insn = INS_SUB else
    '1' when local_insn = INS_ADDL else
    '1' when local_insn = INS_AND else
    '1' when local_insn = INS_OR else
    '1' when local_insn = INS_XOR else
    '1' when local_insn = INS_NOT else
    '1' when local_insn = INS_BSET else
    '1' when local_insn = INS_BCLR else
    '1' when local_insn = INS_RSL else
    '1' when local_insn = INS_RSR else
    '1' when local_insn = INS_MUL else
    '1' when local_insn = INS_CMP else
    '1' when local_insn = INS_LRI else
    '1' when local_insn = INS_LRR else
    '0';

  do_halt <= '1' when (local_insn = INS_HLT) else '0';

end behavioural;

