--file created 2014
--modified
--released to the Public Domain by the author

library IEEE;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

--  Uncomment the following lines to use the declarations that are
--  provided for instantiating Xilinx primitive components.
library UNISIM;
use UNISIM.VComponents.all;

use work.tvc_pack.all;

entity tvc_imp_usb is
  port (
    tvc_clock     : in  std_logic;
    tvc_reset_in  : in  std_logic;
    tvc_reset_out : out std_logic;      --for simulation
    led_out_0     : out std_logic;

    disp1_vgaR  : out std_logic_vector(2 downto 0);
    disp1_vgaG  : out std_logic_vector(2 downto 0);
    disp1_vgaB  : out std_logic_vector(2 downto 1);
    disp1_hSync : out std_logic;
    disp1_vSync : out std_logic;

    usb_fdata  : inout std_logic_vector(7 downto 0);
    usb_faddr  : out   std_logic_vector(1 downto 0);
    usb_slrd   : out   std_logic;
    usb_slwr   : out   std_logic;
    usb_sloe   : out   std_logic;
    usb_slcs   : out   std_logic;
    usb_int0   : out   std_logic;
    usb_pktend : out   std_logic;
    usb_flaga  : in    std_logic;
    usb_flagb  : in    std_logic;
    usb_flagc  : in    std_logic;
    usb_ifclk  : in    std_logic;

    mem_addr     : out   std_logic_vector(23 downto 1);
    nx_ram_clk   : out   std_logic;
    nx_ram_adv_n : out   std_logic;
    nx_ram_cre   : out   std_logic;
    nx_ram_ce_n  : out   std_logic;
    nx_ram_oe_n  : out   std_logic;
    nx_ram_we_n  : out   std_logic;
    nx_ram_lb_n  : out   std_logic;
    nx_ram_ub_n  : out   std_logic;
    nx_ram_wait  : in    std_logic;
    mem_dq       : inout std_logic_vector(15 downto 0)
    );
end tvc_imp_usb;

architecture Behavioral of tvc_imp_usb is

  --usb fifo signals
  signal fifo2_inbusy   : std_logic;
  signal fifo2_pktend   : std_logic;
  signal fifo2_inbyte   : std_logic_vector(7 downto 0);
  signal fifo2_inflag   : std_logic;
  signal fifo0_outbyte  : std_logic_vector(7 downto 0);
  signal fifo0_outflag  : std_logic;
  signal fifo_clk       : std_logic;
  signal fifo2_incount  : std_logic_vector(31 downto 0);  --not used
  signal fifo0_outcount : std_logic_vector(31 downto 0);  --not used
  signal fifo_debugvec  : std_logic_vector(7 downto 0);   --not used

  signal usb_data_in    : std_logic_vector(7 downto 0);
  signal usb_strobe_in  : std_logic;
  signal usb_in_busy    : std_logic;
  signal usb_data_out   : std_logic_vector(8 downto 0);
  signal usb_strobe_out : std_logic;
  signal usb_out_busy   : std_logic;

  signal reset          : std_logic;    --main logic synced reset
  signal clk            : std_logic;    --main logic clock
  signal clk_ram        : std_logic;    --same as above
  signal clk_ram_adv    : std_logic;
  signal clk_ram_adv_dcm : std_logic;
  signal clk_feed       : std_logic;
  signal clk_feed_logic : std_logic;

  signal dcm_0_locked   : std_logic;
  signal dcm_1_locked   : std_logic;
  signal dcm_2_locked   : std_logic;
  signal reset_pre_sync : std_logic;
  signal disp_raw_reset : std_logic;

  signal dcm_1_reset_in : std_logic;
  signal dcm_2_reset_in : std_logic;
  signal dcm_2_reset_in_real : std_logic;
  

  --directions are in the direction of the FUs
  signal cmd_bus_data_in  : std_logic_vector(31 downto 0);
  signal cmd_bus_addr     : std_logic_vector(7 downto 0);
  signal cmd_bus_strobe   : std_logic;
  signal cmd_bus_wr       : std_logic;
  signal cmd_bus_data_out : std_logic_vector(31 downto 0);
  signal busy_bits        : std_logic_vector(31 downto 0);

  signal dcm2_reset_sr : std_logic_vector(15 downto 0) := (others=>'1');
  
begin
  clk_ram_adv <= clk_ram_adv_dcm; --for implementation
  --clk_ram_adv <= clk_ram; --for simulation
  
  clk       <= clk_ram;
  --led_out_0 <= tvc_reset_in;
  led_out_0 <= reset;

  reset_pre_sync <= '1' when dcm_2_locked /= '1' or tvc_reset_in = '1' else '0';
  disp_raw_reset <= '1' when dcm_0_locked /= '1' or tvc_reset_in = '1' else '0';

  rsync : reset_sync
    generic map(init_state => '1')
    port map(clock         => clk, di => reset_pre_sync, do => reset);


  --usb_fifos_imp : usb_fifos
  --  port map(usb_fdata      => usb_fdata, usb_faddr => usb_faddr, usb_slrd => usb_slrd, usb_slwr => usb_slwr, usb_sloe => usb_sloe,
  --           usb_slcs       => usb_slcs, usb_int0 => usb_int0, usb_pktend => usb_pktend, usb_flaga => usb_flaga, usb_flagb => usb_flagb,
  --           usb_flagc      => usb_flagc, usb_ifclk => usb_ifclk,
  --           fifo2_inbusy=>fifo2_inbusy, fifo2_pktend=>fifo2_pktend, fifo2_inbyte   => fifo2_inbyte, fifo2_inflag => fifo2_inflag, fifo0_outbyte => fifo0_outbyte,
  --           fifo0_outflag  => fifo0_outflag, fifo_clk => fifo_clk, fifo2_incount => fifo2_incount,
  --           fifo0_outcount => fifo0_outcount, fifo_debugvec => fifo_debugvec);

  --usb_ca : usb_clock_adapter
  --  port map(tvc_clock      => clk, tvc_reset => reset,
  --           fifo2_inbyte   => fifo2_inbyte, fifo2_inbusy=>fifo2_inbusy, fifo2_pktend=>fifo2_pktend, fifo2_inflag => fifo2_inflag,
  --           fifo0_outbyte => fifo0_outbyte, fifo0_outflag => fifo0_outflag, fifo_clk => fifo_clk,
  --           usb_data_in    => usb_data_in, usb_strobe_in => usb_strobe_in, usb_data_out => usb_data_out,
  --           usb_strobe_out => usb_strobe_out, usb_out_busy => usb_out_busy);

  usb_fifos_imp : usb_io_async
    port map(clock         => clk, reset => reset,
             usb_ifclk     => fifo_clk, usb_fd => usb_fdata, usb_fifo_ad => usb_faddr, usb_flag_a => usb_flaga, usb_flag_b => usb_flagb, usb_flag_c => usb_flagc,
             usb_int0_n    => usb_int0, usb_slcs_n => usb_slcs, usb_slrd => usb_slrd, usb_slwr => usb_slwr, usb_sloe => usb_sloe, usb_pktend => usb_pktend,
             fifo2_inbusy  => usb_out_busy, fifo2_pktend => usb_data_out(8), fifo2_inbyte => usb_data_out(7 downto 0), fifo2_inflag => usb_strobe_out,
             fifo0_outbyte => usb_data_in, fifo0_outflag => usb_strobe_in, fifo0_busy => usb_in_busy);

  usb_cbm : usb_cb_master
    port map(clock           => clk, reset => reset, busy_bits => busy_bits,
             cmd_bus_data_in => cmd_bus_data_out, cmd_bus_data_out => cmd_bus_data_in, cmd_bus_addr => cmd_bus_addr, cmd_bus_wr => cmd_bus_wr, cmd_bus_strobe => cmd_bus_strobe,
             usb_data_in     => usb_data_in, usb_strobe_in => usb_strobe_in, usb_in_busy=>usb_in_busy,
             usb_data_out => usb_data_out, usb_strobe_out => usb_strobe_out, usb_out_busy => usb_out_busy);

  tvc_tl : tvc
    port map (tvc_clock       => clk, ram_clock => clk_ram, ram_clock_adv => clk_ram_adv,
              tvc_reset_in    => reset, disp_raw_clk => clk_feed, disp_raw_reset => disp_raw_reset, tvc_reset_out => tvc_reset_out,
              disp1_vgaR      => disp1_vgaR, disp1_vgaG => disp1_vgaG, disp1_vgaB => disp1_vgaB,
              disp1_hSync     => disp1_hSync, disp1_vSync => disp1_vSync,
              cmd_bus_data_in => cmd_bus_data_in, cmd_bus_addr => cmd_bus_addr, cmd_bus_strobe => cmd_bus_strobe,
              cmd_bus_wr      => cmd_bus_wr, cmd_bus_data_out => cmd_bus_data_out, busy_bits => busy_bits,
              mem_addr        => mem_addr, nx_ram_clk => nx_ram_clk, nx_ram_adv_n => nx_ram_adv_n, nx_ram_cre => nx_ram_cre,
              nx_ram_ce_n     => nx_ram_ce_n, nx_ram_oe_n => nx_ram_oe_n, nx_ram_we_n => nx_ram_we_n, nx_ram_lb_n => nx_ram_lb_n,
              nx_ram_ub_n     => nx_ram_ub_n, nx_ram_wait => nx_ram_wait, mem_dq => mem_dq);

  --this gets the input clock to a dedicated channel so that
  --it can be used across the entire chip.
  DCM_clk_feed : DCM_SP
    generic map (
      CLKDV_DIVIDE          => 2.0,  --  Divide by: 1.5,2.0,2.5,3.0,3.5,4.0,4.5,5.0,5.5,6.0,6.5
      --     7.0,7.5,8.0,9.0,10.0,11.0,12.0,13.0,14.0,15.0 or 16.0
      CLKFX_DIVIDE          => 2,       --  Can be any interger from 1 to 32
      CLKFX_MULTIPLY        => 2,       --  Can be any integer from 1 to 32
      CLKIN_DIVIDE_BY_2     => false,  --  TRUE/FALSE to enable CLKIN divide by two feature
      CLKIN_PERIOD          => 20.0,    --  Specify period of input clock
      CLKOUT_PHASE_SHIFT    => "NONE",  --  Specify phase shift of "NONE", "FIXED" or "VARIABLE" 
      CLK_FEEDBACK          => "1X",  --  Specify clock feedback of "NONE", "1X" or "2X" 
      DESKEW_ADJUST         => "SYSTEM_SYNCHRONOUS",  -- "SOURCE_SYNCHRONOUS", "SYSTEM_SYNCHRONOUS" or
                                        --     an integer from 0 to 15
      DLL_FREQUENCY_MODE    => "LOW",  -- "HIGH" or "LOW" frequency mode for DLL
      DUTY_CYCLE_CORRECTION => true,  --  Duty cycle correction, TRUE or FALSE
      PHASE_SHIFT           => 0,  --  Amount of fixed phase shift from -255 to 255
      STARTUP_WAIT          => true)  --  Delay configuration DONE until DCM_SP LOCK, TRUE/FALSE
    port map (
      CLK0   => clk_feed,               -- 0 degree DCM CLK ouptput
      --CLK180 => CLK180, -- 180 degree DCM CLK output
      --CLK270 => CLK270, -- 270 degree DCM CLK output
      --CLK2X => CLK2X,   -- 2X DCM CLK output
      --CLK2X180 => CLK2X180, -- 2X, 180 degree DCM CLK out
      --CLK90 => CLK90,   -- 90 degree DCM CLK output
      --CLKDV => CLKD,V   -- Divided DCM CLK out (CLKDV_DIVIDE)
      --CLKFX => ,          -- DCM CLK synthesis out (M/D)
      --CLKFX180 => clk_ram_adv, -- 180 degree CLK synthesis out
      LOCKED => dcm_0_locked,           -- DCM LOCK status output
      --PSDONE => PSDONE, -- Dynamic phase adjust done output
      --STATUS => STATUS, -- 8-bit DCM status bits output
      CLKFB  => clk_feed,               -- DCM clock feedback
      CLKIN  => tvc_clock,         -- Clock input (from IBUFG, BUFG or DCM)
      --PSCLK => PSCLK,   -- Dynamic phase adjust clock input
      --PSEN => PSEN,     -- Dynamic phase adjust enable input
      --PSINCDEC => PSINCDEC, -- Dynamic phase adjust increment/decrement
      RST    => tvc_reset_in            -- DCM asynchronous reset input
      );

  dcm_1_reset_in <= '1' when dcm_0_locked /= '1' or tvc_reset_in = '1' else '0';

  --generates 40MHz as input to next clock stage.
  DCM_clk_feed_logic : DCM_SP
    generic map (
      CLKDV_DIVIDE          => 2.0,  --  Divide by: 1.5,2.0,2.5,3.0,3.5,4.0,4.5,5.0,5.5,6.0,6.5
      --     7.0,7.5,8.0,9.0,10.0,11.0,12.0,13.0,14.0,15.0 or 16.0
      CLKFX_DIVIDE          => 5,       --  Can be any interger from 1 to 32
      CLKFX_MULTIPLY        => 8,       --  Can be any integer from 1 to 32
      CLKIN_DIVIDE_BY_2     => false,  --  TRUE/FALSE to enable CLKIN divide by two feature
      CLKIN_PERIOD          => 20.0,    --  Specify period of input clock
      CLKOUT_PHASE_SHIFT    => "NONE",  --  Specify phase shift of "NONE", "FIXED" or "VARIABLE" 
      CLK_FEEDBACK          => "NONE",  --  Specify clock feedback of "NONE", "1X" or "2X" 
      DESKEW_ADJUST         => "SYSTEM_SYNCHRONOUS",  -- "SOURCE_SYNCHRONOUS", "SYSTEM_SYNCHRONOUS" or
                                        --     an integer from 0 to 15
      DLL_FREQUENCY_MODE    => "LOW",  -- "HIGH" or "LOW" frequency mode for DLL
      DUTY_CYCLE_CORRECTION => true,  --  Duty cycle correction, TRUE or FALSE
      PHASE_SHIFT           => 0,  --  Amount of fixed phase shift from -255 to 255
      STARTUP_WAIT          => true)  --  Delay configuration DONE until DCM_SP LOCK, TRUE/FALSE
    port map (
      --CLK0 => CLK0,     -- 0 degree DCM CLK ouptput
      --CLK180 => CLK180, -- 180 degree DCM CLK output
      --CLK270 => CLK270, -- 270 degree DCM CLK output
      --CLK2X => CLK2X,   -- 2X DCM CLK output
      --CLK2X180 => CLK2X180, -- 2X, 180 degree DCM CLK out
      --CLK90 => CLK90,   -- 90 degree DCM CLK output
      --CLKDV => CLKD,V   -- Divided DCM CLK out (CLKDV_DIVIDE)
      CLKFX  => clk_feed_logic,         -- DCM CLK synthesis out (M/D)
      --CLKFX180 => clk_ram_adv, -- 180 degree CLK synthesis out
      LOCKED => dcm_1_locked,           -- DCM LOCK status output
      --PSDONE => PSDONE, -- Dynamic phase adjust done output
      --STATUS => STATUS, -- 8-bit DCM status bits output
      --CLKFB => CLKFB,   -- DCM clock feedback
      CLKIN  => clk_feed,          -- Clock input (from IBUFG, BUFG or DCM)
      --PSCLK => PSCLK,   -- Dynamic phase adjust clock input
      --PSEN => PSEN,     -- Dynamic phase adjust enable input
      --PSINCDEC => PSINCDEC, -- Dynamic phase adjust increment/decrement
      RST    => dcm_1_reset_in          -- DCM asynchronous reset input
      );

  dcm_2_reset_in <= '1' when dcm_1_locked /= '1' or tvc_reset_in = '1' else '0';

  --buffers the main clock; generates clk and clk_ram.
  --phasing was necessary for 80MHz sync ram.
  DCM_main : DCM_SP
    generic map (
      CLKDV_DIVIDE          => 2.0,  --  Divide by: 1.5,2.0,2.5,3.0,3.5,4.0,4.5,5.0,5.5,6.0,6.5
      --     7.0,7.5,8.0,9.0,10.0,11.0,12.0,13.0,14.0,15.0 or 16.0
      CLKFX_DIVIDE          => 2,       --  Can be any interger from 1 to 32
      CLKFX_MULTIPLY        => 2,       --  Can be any integer from 1 to 32
      CLKIN_DIVIDE_BY_2     => false,  --  TRUE/FALSE to enable CLKIN divide by two feature
      CLKIN_PERIOD          => 12.5,  --  Specify period of input clock -> 75Mhz
      CLKOUT_PHASE_SHIFT    => "NONE",  --  Specify phase shift of "NONE", "FIXED" or "VARIABLE" 
      CLK_FEEDBACK          => "1X",  --  Specify clock feedback of "NONE", "1X" or "2X" 
      DESKEW_ADJUST         => "SYSTEM_SYNCHRONOUS",  -- "SOURCE_SYNCHRONOUS", "SYSTEM_SYNCHRONOUS" or
                                        --     an integer from 0 to 15
      DLL_FREQUENCY_MODE    => "LOW",  -- "HIGH" or "LOW" frequency mode for DLL
      DUTY_CYCLE_CORRECTION => true,  --  Duty cycle correction, TRUE or FALSE
      PHASE_SHIFT           => 0,  --  Amount of fixed phase shift from -255 to 255
      STARTUP_WAIT          => true)  --  Delay configuration DONE until DCM_SP LOCK, TRUE/FALSE
    port map (
      CLK0   => clk_ram,                -- 0 degree DCM CLK ouptput
      --CLK180 => CLK180, -- 180 degree DCM CLK output
      CLK270 => clk_ram_adv_dcm,            -- 270 degree DCM CLK output
      --CLK2X => ,   -- 2X DCM CLK output
      --CLK2X180 => clk_ram_adv, -- 2X, 180 degree DCM CLK out
      --CLK90 => CLK90,   -- 90 degree DCM CLK output
      --CLKDV  => clk,                    -- Divided DCM CLK out (CLKDV_DIVIDE)
      --CLKFX => ,   -- DCM CLK synthesis out (M/D)
      --CLKFX180 => , -- 180 degree CLK synthesis out
      LOCKED => dcm_2_locked,           -- DCM LOCK status output
      --PSDONE => PSDONE, -- Dynamic phase adjust done output
      --STATUS => STATUS, -- 8-bit DCM status bits output
      CLKFB  => clk_ram,                -- DCM clock feedback
      CLKIN  => clk_feed_logic,    -- Clock input (from IBUFG, BUFG or DCM)
      --PSCLK => PSCLK,   -- Dynamic phase adjust clock input
      --PSEN => PSEN,     -- Dynamic phase adjust enable input
      --PSINCDEC => PSINCDEC, -- Dynamic phase adjust increment/decrement
      RST    =>  dcm_2_reset_in_real          -- DCM asynchronous reset input
      );
  --End of DCM_SP_inst instantiation

  dcm_2_reset_in_real <= dcm2_reset_sr(15);
  
  process(clk_feed_logic) is
  begin
    if rising_edge(clk_feed_logic) then
      if dcm_2_reset_in = '0' then
        dcm2_reset_sr <= dcm2_reset_sr(14 downto 0) & '0';
      end if;
    end if;
  end process;
  
end architecture;
