--file created 2014
--modified
--released to the Public Domain by the author

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.tvc_tb_pack.all;
use work.tvc_pack.all;

use std.textio.all;
use ieee.std_logic_textio.all;

--ok we are going to have to:
--open a command data file which will:
--write to the shader's program ram to fill it
--(maybe re-read to test and validate that functionality)
--write to some data ram memory addresses. (fill in arguments)
--probably read the above too.
--write to a control register to release the cpu
--poll control register for execution to complete by HLT
--write new values into ram memory address for next polygon
--reset processor and wait for halt.
--the command data file will be single through
--will probably be created by capturing output of engine with a
--new c++ test generation backend. I expect the file to be _big_

--format of file?
--cmd (1 hex digit) 

entity tvc_imp_usb_tb is end tvc_imp_usb_tb;

architecture behaviour of tvc_imp_usb_tb is
  signal clk           : std_logic := '0';
  signal sigterm       : std_logic := '0';
  signal reset         : std_logic := '0';
  signal tvc_reset_out : std_logic;

  signal cmd_wr_strobe    : std_logic;
  signal cmd_rd_strobe    : std_logic;
  signal cb_address       : std_logic_vector(7 downto 0);
  signal cb_data_to_write : std_logic_vector(31 downto 0);
  signal cb_read_data     : std_logic_vector(31 downto 0);
  signal usb_busy         : std_logic;

  signal usb_fd     : std_logic_vector(7 downto 0) := (others => '0');
  signal usb_faddr  : std_logic_vector(1 downto 0);
  signal usb_slrd   : std_logic;
  signal usb_slwr   : std_logic;
  signal usb_sloe   : std_logic;
  signal usb_slcs   : std_logic;
  signal usb_int0   : std_logic;
  signal usb_pktend : std_logic;
  signal usb_flaga  : std_logic                    := '0';
  signal usb_flagb  : std_logic                    := '0';
  signal usb_flagc  : std_logic                    := '0';
  signal usb_ifclk  : std_logic                    := '0';

  signal disp1_vgaR  : std_logic_vector(2 downto 0);
  signal disp1_vgaG  : std_logic_vector(2 downto 0);
  signal disp1_vgaB  : std_logic_vector(2 downto 1);
  signal disp1_hSync : std_logic;
  signal disp1_vSync : std_logic;

  signal mem_addr     : std_logic_vector(22 downto 0);
  signal nx_ram_clk   : std_logic;
  signal nx_ram_adv_n : std_logic;
  signal nx_ram_cre   : std_logic;
  signal nx_ram_ce_n  : std_logic;
  signal nx_ram_oe_n  : std_logic;
  signal nx_ram_we_n  : std_logic;
  signal nx_ram_lb_n  : std_logic;
  signal nx_ram_ub_n  : std_logic;
  signal nx_ram_wait  : std_logic;
  signal mem_dq       : std_logic_vector(15 downto 0);

  file stimulus : text open read_mode is "blah.txt";
  --file stimulus : text open read_mode is "ise_mc_debug.txt";

  component cellram
    port(
      clk    : in    std_logic;
      adv_n  : in    std_logic;
      cre    : in    std_logic;
      o_wait : out   std_logic;
      ce_n   : in    std_logic;
      oe_n   : in    std_logic;
      we_n   : in    std_logic;
      lb_n   : in    std_logic;
      ub_n   : in    std_logic;
      addr   : in    std_logic_vector(23 downto 1);
      dq     : inout std_logic_vector(15 downto 0));
  end component;

  
begin

  usbm : usb_master
    port map(clk           => clk, reset => reset,
             cmd_wr_strobe => cmd_wr_strobe, cmd_rd_strobe => cmd_rd_strobe, cb_address => cb_address, cb_data_to_write => cb_data_to_write, cb_read_data => cb_read_data, usb_busy => usb_busy,
             usb_fd        => usb_fd, usb_faddr => usb_faddr, usb_slrd => usb_slrd, usb_slwr => usb_slwr,
             usb_sloe      => usb_sloe, usb_slcs => usb_slcs, usb_int0 => usb_int0, usb_pktend => usb_pktend,
             usb_flaga     => usb_flaga, usb_flagb => usb_flagb, usb_flagc => usb_flagc, usb_ifclk => usb_ifclk);

  tvc_imp_1 : tvc_imp_usb
    port map (tvc_clock   => clk, tvc_reset_in => reset, tvc_reset_out => tvc_reset_out,
              disp1_vgaR  => disp1_vgaR, disp1_vgaG => disp1_vgaG, disp1_vgaB => disp1_vgaB, disp1_hSync => disp1_hSync, disp1_vSync => disp1_vSync,
              usb_fdata   => usb_fd, usb_faddr => usb_faddr, usb_slrd => usb_slrd, usb_slwr => usb_slwr, usb_sloe => usb_sloe,
              usb_slcs    => usb_slcs, usb_int0 => usb_int0, usb_pktend => usb_pktend, usb_flaga => usb_flaga,
              usb_flagb   => usb_flagb, usb_flagc => usb_flagc, usb_ifclk => usb_ifclk,
              mem_addr    => mem_addr, nx_ram_clk => nx_ram_clk, nx_ram_adv_n => nx_ram_adv_n, nx_ram_cre => nx_ram_cre,
              nx_ram_ce_n => nx_ram_ce_n, nx_ram_oe_n => nx_ram_oe_n, nx_ram_we_n => nx_ram_we_n, nx_ram_lb_n => nx_ram_lb_n,
              nx_ram_ub_n => nx_ram_ub_n, nx_ram_wait => nx_ram_wait, mem_dq => mem_dq);

  jram : cellram
    port map(
      clk    => nx_ram_clk,
      adv_n  => nx_ram_adv_n,
      cre    => nx_ram_cre,
      o_wait => nx_ram_wait,
      ce_n   => nx_ram_ce_n,
      oe_n   => nx_ram_oe_n,
      we_n   => nx_ram_we_n,
      lb_n   => nx_ram_lb_n,
      ub_n   => nx_ram_ub_n,
      addr   => mem_addr,
      dq     => mem_dq);

  system_clock : process                --system clock
  begin
    wait for 50 ns;
    clkloop : loop
      wait for 10 ns;                   --100MHz clock (1/2 period)
      clk <= not clk;
      if sigterm = '1' then
        exit;
      end if;
    end loop clkloop;
    wait for 1 us;
    wait;
  end process;

  system_reset : process
  begin
    wait for 5 ns;
    reset <= '1';
    wait for 400 ns;
    reset <= '0';
    wait;
  end process;

  test_process : process
    variable cmd_line : line;
    variable out_line : line;
    variable cmd      : std_logic_vector(31 downto 0);
    variable cmd_addr : std_logic_vector(31 downto 0);
    variable cmd_data : std_logic_vector(31 downto 0);
  begin
    cmd_wr_strobe <= '0';
    cmd_rd_strobe <= '0';

    wait for 5 us;
    wait until tvc_reset_out = '0';
    wait for 1 us;

    wait for 200 us;

    loop
      write(output, "jtb: loop1 start\n");
      readline(stimulus, cmd_line);
      --next when cmd_line.all(0) = '#'; --how do I do this?
      --write(output, "reading cmd");
      hread(cmd_line, cmd);
      --write(output, "reading addr");
      hread(cmd_line, cmd_addr);
      --write(output, "reading data");
      hread(cmd_line, cmd_data);
      --write(out_line, string'("read_command: "));
      hwrite(out_line, cmd);
      write(out_line, string'(" "));
      hwrite(out_line, cmd_addr);
      write(out_line, string'(" "));
      hwrite(out_line, cmd_data);
      --writeline(output, out_line);
      --write(output, string'("before wait"));
      wait until usb_busy = '0' for 50 ns;
      --write(output, string'("after wait"));

      cb_address       <= cmd_addr(7 downto 0);
      cb_data_to_write <= cmd_data;

      --loop exists only for command 3 where we must re-read
      --a command bus address until it equals a certain value
      loop
        write(output, "jtb: loop2 start\n");

        case cmd(3 downto 0) is
          when X"1" =>
            cmd_wr_strobe <= '1';
          when X"2" =>
            cmd_rd_strobe <= '1';
          when X"3" =>
            cmd_rd_strobe <= '1';
          when others =>
            assert false report "jtb: invalid command from file" severity failure;
        end case;

        wait for 20 ns;
        wait until usb_busy = '1' for 20 ns;
        cmd_wr_strobe <= '0';
        cmd_rd_strobe <= '0';
        wait until usb_busy = '0';
        wait for 20 ns;
        exit when cmd = X"00000001";  --don't wait for read data if we just wrote.

        --we know we are in command 2 or 3
        wait for 500 ns;  --?enough time for a possible read to complete?

        --write(output, string'("oh hey, usb_busy must be zero?"));
        assert usb_busy = '0' report "jtb: spi_busy not zero" severity failure;
        exit when cmd /= X"00000003";
        --we know cmd == 3
        exit when cb_read_data(15 downto 0) = cmd_data(15 downto 0);
        deallocate(out_line);
        write(out_line, string'("jtb: cmd3 result: "));
        hwrite(out_line, cb_read_data);
        writeline(output, out_line);
      end loop;
      deallocate(out_line);
      write(out_line, string'("jtb: after loop 2! cmd: "));
      hwrite(out_line, cmd);
      writeline(output, out_line);

      case cmd(3 downto 0) is
        when X"2" =>
          deallocate(out_line);
          write(out_line, string'("jtb: read_value: "));
          hwrite(out_line, cb_read_data);
          writeline(output, out_line);
        when X"3" =>
          write(out_line, string'("jtb: cmd 3 completed!"));
        when others =>
          null;
      end case;
    end loop;

  end process;

end behaviour;
