--file created 2014
--modified
--released to the Public Domain by the author

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

package tvc_pack is

  component tvc
    port (
      tvc_clock        : in    std_logic;
      ram_clock        : in    std_logic;
      ram_clock_adv    : in    std_logic;
      tvc_reset_in     : in    std_logic;
      tvc_reset_out    : out   std_logic;
      disp_raw_clk     : in    std_logic;
      disp_raw_reset   : in    std_logic;
      disp1_vgaR       : out   std_logic_vector(2 downto 0);
      disp1_vgaG       : out   std_logic_vector(2 downto 0);
      disp1_vgaB       : out   std_logic_vector(2 downto 1);
      disp1_hSync      : out   std_logic;
      disp1_vSync      : out   std_logic;
      cmd_bus_data_in  : in    std_logic_vector(31 downto 0);
      cmd_bus_addr     : in    std_logic_vector(7 downto 0);
      cmd_bus_strobe   : in    std_logic;
      cmd_bus_wr       : in    std_logic;
      cmd_bus_data_out : out   std_logic_vector(31 downto 0);
      busy_bits        : out   std_logic_vector(31 downto 0);
      mem_addr         : out   std_logic_vector(22 downto 0);
      nx_ram_clk       : out   std_logic;
      nx_ram_adv_n     : out   std_logic;
      nx_ram_cre       : out   std_logic;
      nx_ram_ce_n      : out   std_logic;
      nx_ram_oe_n      : out   std_logic;
      nx_ram_we_n      : out   std_logic;
      nx_ram_lb_n      : out   std_logic;
      nx_ram_ub_n      : out   std_logic;
      nx_ram_wait      : in    std_logic;
      mem_dq           : inout std_logic_vector(15 downto 0)
      );
  end component;

  component nexus2_ram_interface is
    --here to speed simulation. if not added takes too long to simulate
    generic (
      startup_delay_time : std_logic_vector(15 downto 0) := X"FFFF";
      start_clear_addr   : std_logic_vector(22 downto 0) := "000" & X"0_0000");
    port (
      busy          : out std_logic;
      startup_delay : out std_logic;

      blk_rd_req         : in  std_logic;
      blk_wr_req         : in  std_logic;
      blk_xfer_addr      : in  std_logic_vector(23 downto 2);
      blk_xfer_des_count : in  std_logic_vector(11 downto 0);
      blk_dout           : out std_logic_vector(31 downto 0);
      blk_din            : in  std_logic_vector(31 downto 0);
      --external connection should either consume or produce new data
      blk_strobe         : out std_logic;

      --dword read/write interface; will also be used for generic reads and writes.
      --bottom bit is ignored in sync mode
      dword_strobe   : in  std_logic;
      dword_addr     : in  std_logic_vector(23 downto 2);
      dword_data_out : out std_logic_vector(31 downto 0);
      dword_data_in  : in  std_logic_vector(31 downto 0);
      dword_byte_we  : in  std_logic_vector(3 downto 0);

      --interface to the ram chip
      nx_ram_addr  : out std_logic_vector(22 downto 0);
      nx_ram_clk   : out std_logic;
      nx_ram_adv_n : out std_logic;     --address valid active low
      nx_ram_cre   : out std_logic;     --control register enable
      nx_ram_ce_n  : out std_logic;     --chip enable active low
      nx_ram_oe_n  : out std_logic;     --output enable active low
      nx_ram_we_n  : out std_logic;     --write enable active low
      nx_ram_lb_n  : out std_logic;     --lower byte enable active low
      nx_ram_ub_n  : out std_logic;     --upper byte enable active low
      nx_ram_wait  : in  std_logic;     --wait input

      nx_ram_dq : inout std_logic_vector(15 downto 0);

      clock_ram     : in std_logic;
      clock_ram_adv : in std_logic;
      reset         : in std_logic);
  end component;

  component reset_sync is
    generic (init_state : std_logic := '1');
    port (clock : in  std_logic;
          di    : in  std_logic;
          do    : out std_logic);
  end component;

  component std_sync is
    generic (init_state : std_logic := '0');
    port (
      sig_in  : in  std_logic;
      sig_out : out std_logic;
      clock   : in  std_logic;
      reset   : in  std_logic
      );
  end component;

  component scanline_cache is
    port(
      mu_bus_address     : in    std_logic_vector(1 downto 0);
      mu_bus_data        : inout std_logic_vector(31 downto 0);
      mu_bus_write       : in    std_logic;
      mu_bus_strobe      : in    std_logic;
      sc_irq             : out   std_logic;
      sc_select          : in    std_logic;
      tvc_clock          : in    std_logic;
      tvc_reset          : in    std_logic;
      sc_start_fill      : in    std_logic;
      sc_start_fill_addr : in    std_logic_vector(23 downto 0);
      pixel_clock        : in    std_logic;
      next_pix_addr      : in    std_logic_vector(11 downto 0);
      next_pix_values    : out   std_logic_vector(31 downto 0)
      );
  end component;

  component disp_driver is
    port (
      tvc_clock         : in  std_logic;
      tvc_reset         : in  std_logic;
      odd_fill_req      : out std_logic;
      even_fill_req     : out std_logic;
      reset_cntrs_req   : out std_logic;
      pixel_clk         : in  std_logic;
      pixel_reset       : in  std_logic;
      next_pix_addr     : out std_logic_vector(11 downto 0);
      odd_next_pix_val  : in  std_logic_vector(31 downto 0);
      even_next_pix_val : in  std_logic_vector(31 downto 0);
      disp_pix_value    : out std_logic_vector(7 downto 0);
      disp_horiz_sync   : out std_logic;
      disp_vert_sync    : out std_logic
      );
  end component;

  component pix_pusher
    port(
      mu_bus_address  : in    std_logic_vector(1 downto 0);
      mu_bus_data     : inout std_logic_vector(31 downto 0);
      mu_bus_write    : in    std_logic;
      mu_bus_strobe   : in    std_logic;
      odd_sc_irq      : out   std_logic;
      odd_sc_select   : in    std_logic;
      even_sc_irq     : out   std_logic;
      even_sc_select  : in    std_logic;
      tvc_clock       : in    std_logic;
      tvc_reset       : in    std_logic;
      disp_raw_clk    : in    std_logic;
      disp_raw_reset  : in    std_logic;
      disp_pix_value  : out   std_logic_vector(7 downto 0);
      disp_horiz_sync : out   std_logic;
      disp_vert_sync  : out   std_logic
      );
  end component;

  component mcu_struct is
    port (
      nx_ram_addr          : out   std_logic_vector(22 downto 0);
      nx_ram_clk           : out   std_logic;
      nx_ram_adv_n         : out   std_logic;
      nx_ram_cre           : out   std_logic;
      nx_ram_ce_n          : out   std_logic;
      nx_ram_oe_n          : out   std_logic;
      nx_ram_we_n          : out   std_logic;
      nx_ram_lb_n          : out   std_logic;
      nx_ram_ub_n          : out   std_logic;
      nx_ram_wait          : in    std_logic;
      nx_ram_dq            : inout std_logic_vector(15 downto 0);
      tvc_clock            : in    std_logic;
      tvc_reset            : in    std_logic;
      ram_clock            : in    std_logic;
      ram_clock_adv        : in    std_logic;
      ram_start_delay      : out   std_logic;
      mu_bus_address       : out   std_logic_vector(1 downto 0);
      mu_bus_data          : inout std_logic_vector(31 downto 0);
      mu_bus_write         : out   std_logic;  --if high the MCU is going to write to a mu
      mu_bus_strobe        : out   std_logic;
      odd_sc_irq           : in    std_logic;
      odd_sc_select        : out   std_logic;
      even_sc_irq          : in    std_logic;
      even_sc_select       : out   std_logic;
      block_reader_irq     : in    std_logic;
      block_reader_select  : out   std_logic;
      block_writer_irq     : in    std_logic;
      block_writer_select  : out   std_logic;
      rp_writer_irq        : in    std_logic;
      rp_writer_select     : out   std_logic;
      line_drawer_irq      : in    std_logic;
      line_drawer_select   : out   std_logic;
      poly_scanline_irq    : in    std_logic;
      poly_scanline_select : out   std_logic
      );
  end component;

  component mcu_sched is
    port (
      tvc_clock            : in  std_logic;
      tvc_reset            : in  std_logic;
      mcu_start            : out std_logic;
      mcu_busy             : in  std_logic;
      odd_sc_irq           : in  std_logic;
      odd_sc_select        : out std_logic;
      even_sc_irq          : in  std_logic;
      even_sc_select       : out std_logic;
      block_reader_irq     : in  std_logic;
      block_reader_select  : out std_logic;
      block_writer_irq     : in  std_logic;
      block_writer_select  : out std_logic;
      rp_writer_irq        : in  std_logic;
      rp_writer_select     : out std_logic;
      line_drawer_irq      : in  std_logic;
      line_drawer_select   : out std_logic;
      poly_scanline_irq    : in  std_logic;
      poly_scanline_select : out std_logic
      );
  end component;

  component mcu_hl is
    port (
      mu_bus_address         : out   std_logic_vector(1 downto 0);
      mu_bus_data            : inout std_logic_vector(31 downto 0);
      mu_bus_write           : out   std_logic;
      mu_bus_strobe          : out   std_logic;
      tvc_clock              : in    std_logic;
      tvc_reset              : in    std_logic;
      mcu_hl_start_strobe    : in    std_logic;
      mcu_hl_busy            : out   std_logic;
      mcu_ll_data_out        : out   std_logic_vector(31 downto 0);
      mcu_ll_data_in         : in    std_logic_vector(31 downto 0);
      mcu_ll_address         : out   std_logic_vector(31 downto 2);
      mcu_ll_xfer_write      : out   std_logic;
      mcu_ll_xfer_count      : out   std_logic_vector(11 downto 0);
      mcu_ll_data_valid_done : in    std_logic;  --either the mcu is done reading our data or has just written new data to us. high only a single cycle
      mcu_ll_start_strobe    : out   std_logic;
      mcu_ll_idle            : in    std_logic  --goes high on cycle after start strobe. goes low if transfer complete or an error occurs.
      );
  end component;

  component block_reader is
    port(
      mu_bus_address      : in    std_logic_vector(1 downto 0);
      mu_bus_data         : inout std_logic_vector(31 downto 0);
      mu_bus_write        : in    std_logic;  --if high the MCU is going to write to a mu
      mu_bus_strobe       : in    std_logic;
      block_reader_irq    : out   std_logic;
      block_reader_select : in    std_logic;
      tvc_clock           : in    std_logic;
      tvc_reset           : in    std_logic;
      cmd_bus_data        : in    std_logic_vector(31 downto 0);
      cmd_bus_addr        : in    std_logic_vector(7 downto 0);
      cmd_bus_wr          : in    std_logic;
      cmd_bus_strobe      : in    std_logic;
      cmd_bus_data_out    : out   std_logic_vector(31 downto 0);
      idle                : out   std_logic
      );
  end component;

  component block_writer is
    port(
      mu_bus_address      : in    std_logic_vector(1 downto 0);
      mu_bus_data         : inout std_logic_vector(31 downto 0);
      mu_bus_write        : in    std_logic;
      mu_bus_strobe       : in    std_logic;
      block_writer_irq    : out   std_logic;
      block_writer_select : in    std_logic;
      tvc_clock           : in    std_logic;
      tvc_reset           : in    std_logic;
      cmd_bus_data        : in    std_logic_vector(31 downto 0);
      cmd_bus_addr        : in    std_logic_vector(7 downto 0);
      cmd_bus_wr          : in    std_logic;
      cmd_bus_strobe      : in    std_logic;
      idle                : out   std_logic
      );
  end component;

  component block_set is
    port(
      mu_bus_address   : in    std_logic_vector(1 downto 0);
      mu_bus_data      : inout std_logic_vector(31 downto 0);
      mu_bus_write     : in    std_logic;
      mu_bus_strobe    : in    std_logic;
      block_set_irq    : out   std_logic;
      block_set_select : in    std_logic;
      tvc_clock        : in    std_logic;
      tvc_reset        : in    std_logic;
      cmd_bus_data     : in    std_logic_vector(31 downto 0);
      cmd_bus_addr     : in    std_logic_vector(7 downto 0);
      cmd_bus_wr       : in    std_logic;
      cmd_bus_strobe   : in    std_logic;
      idle             : out   std_logic
      );
  end component;

  component poly_scanline is
    port (
      mu_bus_address       : in    std_logic_vector(1 downto 0);
      mu_bus_data          : inout std_logic_vector(31 downto 0);
      mu_bus_write         : in    std_logic;
      mu_bus_strobe        : in    std_logic;
      poly_scanline_irq    : out   std_logic;
      poly_scanline_select : in    std_logic;
      tvc_clock            : in    std_logic;
      tvc_reset            : in    std_logic;
      cmd_bus_data         : in    std_logic_vector(31 downto 0);
      cmd_bus_addr         : in    std_logic_vector(7 downto 0);
      cmd_bus_wr           : in    std_logic;
      cmd_bus_strobe       : in    std_logic;
      idle                 : out   std_logic
      );
  end component;

  component usb_cb_master is
    port(
      clock            : in  std_logic;
      reset            : in  std_logic;
      busy_bits        : in  std_logic_vector(31 downto 0);
      cmd_bus_data_in  : in  std_logic_vector(31 downto 0);
      cmd_bus_data_out : out std_logic_vector(31 downto 0);
      cmd_bus_addr     : out std_logic_vector(7 downto 0);
      cmd_bus_wr       : out std_logic;
      cmd_bus_strobe   : out std_logic;
      usb_data_in      : in  std_logic_vector(7 downto 0);
      usb_strobe_in    : in  std_logic;
      usb_in_busy      : out std_logic;
      usb_data_out     : out std_logic_vector(8 downto 0);
      usb_strobe_out   : out std_logic;
      usb_out_busy     : in  std_logic
      );
  end component;

  component usb_io_async is
    port(
      clock         : in    std_logic;
      reset         : in    std_logic;
      usb_ifclk     : in    std_logic;
      usb_fd        : inout std_logic_vector(7 downto 0);
      usb_fifo_ad   : out   std_logic_vector(1 downto 0);
      usb_flag_a    : in    std_logic;
      usb_flag_b    : in    std_logic;
      usb_flag_c    : in    std_logic;
      usb_int0_n    : out   std_logic;
      usb_slcs_n    : out   std_logic;
      usb_slrd      : out   std_logic;
      usb_slwr      : out   std_logic;
      usb_sloe      : out   std_logic;
      usb_pktend    : out   std_logic;
      fifo2_inbusy  : out   std_logic;  --writes are ignored if high
      fifo2_pktend  : in    std_logic;
      fifo2_inbyte  : in    std_logic_vector(7 downto 0);
      fifo2_inflag  : in    std_logic;
      fifo0_outbyte : out   std_logic_vector(7 downto 0);
      fifo0_outflag : out   std_logic;  --strobed only once.
      fifo0_busy    : in    std_logic  --I won't read more usb data if this is high.
      );
  end component;

end package;

